// Copyright (C) 1999-2010
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#ifndef __base_h__
#define __base_h__

#include "widget.h"
#include "vector3d.h"

#include "callback.h"
#include "fvcontour.h"
#include "frscale.h"
#include "grid.h"
#include "hist.h"
#include "inversescale.h"
#include "list.h"
#include "tag.h"
#include "util.h"

extern int DebugAST;
extern int DebugMosaic;
extern int DebugPerf;
extern int DebugWCS;
extern int DebugBin;
extern int DebugCompress;
extern int DebugGZ;
extern int DebugRGB;

#define DEFAULTLOW 0
#define DEFAULTHIGH 100
#define HISTEQUSIZE 16384
#define MULTWCS 28
#define SCALESIZE 16384

class Attribute;
class BaseBox;
class BaseEllipse;
class BaseMarker;
class Composite;
class Context;
class Contour;
class Filter;
class FitsImage;
class FitsMask;
class FVContour;
class Grid;
class Marker;

class Annulus;
class Box;
class BoxAnnulus;
class Bpanda;
class Circle;
class Compass;
class Contour;
class Cpanda;
class Ellipse;
class EllipseAnnulus;
class Epanda;
class Line;
class Point;
class CirclePoint;
class BoxPoint;
class DiamondPoint;
class CrossPoint;
class ExPoint;
class ArrowPoint;
class BoxCirclePoint;
class Projection;
class Circle3d;
class Polygon;
class Ruler;
class Text;
class Vect;

class Base : public Widget {
  friend class Attribute;
  friend class BaseBox;
  friend class BaseEllipse;
  friend class BaseMarker;
  friend class Composite;
  friend class Context;
  friend class Contour;
  friend class FitsImage;
  friend class FitsMask;
  friend class FVContour;
  friend class Grid;
  friend class Marker;

  friend class Annulus;
  friend class Box;
  friend class BoxAnnulus;
  friend class Bpanda;
  friend class Circle;
  friend class Compass;
  friend class Cpanda;
  friend class Ellipse;
  friend class EllipseAnnulus;
  friend class Epanda;
  friend class Line;
  friend class Point;
  friend class CirclePoint;
  friend class BoxPoint;
  friend class DiamondPoint;
  friend class CrossPoint;
  friend class ExPoint;
  friend class ArrowPoint;
  friend class BoxCirclePoint;
  friend class Projection;
  friend class Circle3d;
  friend class Polygon;
  friend class Ruler;
  friend class Text;
  friend class Vect;

public:
  enum CompressType {NOCOMPRESS, GZ};
  enum FileNameType {ROOTBASE, FULLBASE, ROOT, FULL};
  enum MarkerFormat {DS9, XML, CIAO, SAOTNG, SAOIMAGE, PROS, RAWXY};
  enum MarkerLayer {USER, TEMPLATE, CATALOG};
  enum RenderMethod {MIP,AIP,THRESHOLD,BLEND};
  enum ShmType {SHMID,KEY};
  enum UndoMarkerType {NONE, MOVE, EDIT, DELETE};
  enum UpdateType {MATRIX, BASE, PIXMAP, NOUPDATE};

  enum MemType {ALLOC, ALLOCGZ, CHANNEL, MMAP, SMMAP, MMAPINCR, 
		SHARE, SSHARE, SOCKET, SOCKETGZ, VAR, HIST, COMPRESS};
  enum MosaicType {NOMOSAIC, IRAF, WCSMOSAIC, WFPC2};
  enum LayerType {IMG, MASK};
  enum LoadMethod {LOADALL, INCR};

 private:
  InverseScale* inverseScale;

 protected:
  List<Marker>* markers;     // pointer to current marker stack
  List<Marker>* undoMarkers; // pointer to current marker stack
  List<Marker>* pasteMarkers;// pointer to current marker stack

  List<Marker> userMarkers;      // foreground marker stack
  List<Marker> undoUserMarkers;  // remember marker to undo
  List<Marker> pasteUserMarkers; // cut/copy/paste

  List<Marker> catalogMarkers;      // background marker stack
  List<Marker> undoCatalogMarkers;  // remember marker to undo
  List<Marker> pasteCatalogMarkers; // cut/copy/paste

  UndoMarkerType undoMarkerType; // type

  Vector markerBegin;        // marker move begin in image coords
  Marker* editMarker;        // pointer to marker currently being edited
  Marker* rotateMarker;      // pointer to marker currently being rotated
  Vector regionBegin;        // select region begin in canvas coords
  Vector regionEnd;          // select region end in canvas coords
  CoordSystem markerDefaultSystem; // default marker sys (RAWXY)
  SkyFrame markerDefaultSky; // default marker sky (RAWXY)

  Composite* compositeMarker; // pointer to current composite marker

  Vector cursor; // current cursor position in USER coords
  Vector crosshair; // crosshair location in REF coords

  XImage* baseXImage; // ximage of image
  Pixmap basePixmap; // pixmap of image
  UpdateType needsUpdate; // updateFlags

  Context* currentContext;
  Context* keyContext;
  int keyContextSet;

  Orientation orientation;   // current image orientation
  Matrix orientationMatrix;  // current image orientation matrix
  double rotation;           // current image rotation angle

  double rotateRotation;     // tmp cursor for interactive rotation

  int preservePan;           // flag to indicate preserve between loads

  Vector zoom_; // current image zoom value

  Pixmap panPM;              // tmp pixmap for pan motion
  Vector panCursor;          // tmp cursor for interactive panning
  GC panGCXOR;               // GC for interactive panning
  Pixmap pannerPixmap;       // pixmap for panner
  XImage* pannerXImage;      // ximage for panner
  int pannerWidth;           // panner width
  int pannerHeight;          // panner height
  char pannerName[32];       // panner widget name
  CoordSystem pannerSystem;  // panner wcs coord system
  SkyFrame pannerSky;        // panner wcs sky frame
  int usePanner;             // panner status

  Pixmap magnifierPixmap;    // pixmap for magnifier
  XImage* magnifierXImage;   // ximage for magnifier
  int magnifierWidth;        // magnifier width
  int magnifierHeight;       // magnifier height
  double magnifierZoom_;     // magnifier zoom factor
  char magnifierName[32];    // magnifer widget name
  int useMagnifier;          // magnifer status
  int useMagnifierGraphics;  // magnifer marker status
  int useMagnifierCursor;    // magnifier cursor status
  GC magnifierGC;            // GC for magnifier
  Vector magnifierCursor;    // we need to save the last cursor used
  char* magnifierColorName;

  int wcsAlignment; // flag to indicate wcs alignment
  CoordSystem wcsSystem; // wcs system
  SkyFrame wcsSky; // wcs sky frame
  Orientation wcsOrientation; // wcs image orientation
  Matrix wcsOrientationMatrix; // wcs image orientation matrix
  double wcsRotation; // wcs image rotation angle
  Orientation irafOrientation; // orientation of first iraf segment

  char* maskColorName;
  float maskAlpha;
  int maskMark;

  int invert;

  FitsHist::Function binFunction_;
  Vector binFactor_;
  int binBufferSize_;
  int binDepth_;

  int doSmooth_;
  int smoothFunction_;
  int smoothRadius_;

  GC highliteGC;
  int useHighlite;

  XRectangle rectWidget[1];
  XRectangle rectWindow[1];

  int useCrosshair;
  GC crosshairGC;
  GC crosshairGCXOR;

  int showMarkers;
  int showMarkersText;
  int centroidAuto;
  int centroidIteration;
  float centroidRadius;
  int preserveMarkers;       // flag to indicate preserve between loads

  GC markerGC; // marker gc
  GC markerGCXOR; // marker xor gc
  GC selectGCXOR; // select gc

  Grid* grid;
  GC gridGC;

  GC contourGC;

  char* bgColorName;
  XColor* bgColor;
  char bgTrueColor_[4];   // color encoded

  char* nanColorName;
  XColor* nanColor;
  char nanTrueColor_[4];  // color encoded

  Matrix canvasToRef;
  Matrix canvasToUser;
  Matrix canvasToWidget;
  Matrix canvasToWindow;
  Matrix dataToImage;
  Matrix pannerToRef;
  Matrix pannerToUser;
  Matrix pannerToWidget;
  Matrix refToCanvas;
  Matrix refToPanner;
  Matrix refToUser;
  Matrix refToWidget;
  Matrix refToWindow;
  Matrix imageToData;
  Matrix userToCanvas;
  Matrix userToMagnifier;
  Matrix userToPanner;
  Matrix userToRef;
  Matrix userToWidget;
  Matrix widgetToCanvas;
  Matrix widgetToPanner;
  Matrix widgetToRef;
  Matrix widgetToUser;
  Matrix widgetToWindow;
  Matrix windowToCanvas;
  Matrix windowToRef;
  Matrix windowToWidget;

  Matrix3d dataToImage3d;
  Matrix3d imageToData3d;

 private:
  void bltHist(char*, char*); // frblt.C

  void doubleToTclArray(double, Precision, const char*, const char*, 
			const char*);
  void invalidPixmap();

  int updatePixmap(const BBox&);

 protected:
  virtual void alignWCS() =0;
  virtual void alignWCS(FitsImage*) =0;
  void analysisFits();

  virtual void buildXImage(XImage*, double* (FitsImage::*)()) =0;

  virtual double calcZoom(Vector,Vector) =0;
  virtual void centerImage() =0;

  void coordToTclArray(FitsImage*, const Vector&, InternalSystem, CoordSystem, 
		       Precision, const char*, const char*);

  virtual int doRender() =0;

  virtual void encodeTrueColor(int, int) =0;
  virtual void encodeTrueColor(XColor*, char*) =0;
  virtual void encodeTrueColor(unsigned char*, XImage*) =0;

  virtual unsigned char* fillImage(int, int, double* (FitsImage::*)()) =0;

  Matrix& getCanvasToWidget() {return canvasToWidget;}
  virtual void getInfo(const Vector&, InternalSystem,
		       Matrix& (FitsImage::*)(), Matrix& (FitsImage::*)(),
		       SkyFrame, SkyFormat, char*);
  Orientation getOrientation() {return orientation;}
  Matrix& getRefToCanvas() {return refToCanvas;}
  Matrix& getRefToWidget() {return refToWidget;}
  Matrix& getRefToWindow() {return refToWindow;}
  double getRotation() {return rotation + wcsRotation;}
  Orientation getWCSOrientation(CoordSystem sys, SkyFrame sky);
  Orientation getWCSOrientation() {return wcsOrientation;}
  double getWCSRotation(CoordSystem sys, SkyFrame sky);
  Matrix& getWidgetToCanvas() {return widgetToCanvas;}

  int hasATMV();
  int hasContour();
  int hasContourAux();
  int hasDTMV();
  int hasLTMV();
  int hasWCS(CoordSystem);
  int hasWCSEqu(CoordSystem); 
  int hasWCSLinear(CoordSystem);
  virtual int isIIS() =0;
  void loadDone(int, LayerType);

  double mapDistFromRef(const Vector&, const Vector&, InternalSystem);
  Vector mapFromRef(const Vector&, InternalSystem);
  double mapLenFromRef(double, InternalSystem);
  Vector mapLenFromRef(const Vector&, InternalSystem);
  double mapLenToRef(double, InternalSystem);
  Vector mapLenToRef(const Vector&, InternalSystem);
  Vector mapToRef(const Vector&, InternalSystem);

  virtual void pannerPush() =0;
  FitsHead* parseWCS(istream&);
  void printCoordSystem(CoordSystem);
  void printDouble(double, Precision);
  void printInteger(int);
  void printSkyFormat(SkyFormat);
  void printSkyFrame(SkyFrame);
  void printVector(const Vector&, Precision);
  void ps(double* (FitsImage::*getMatrix)());
  void psImage(Filter&, PSColorSpace, int, int, float,
	       double* (FitsImage::*getMatrix)());

  void renderHighLite();
  virtual void renderMagnifierCursor(const Vector& v) =0;
  virtual void reset();
  void setClip();

  void setScanModeIncr(LoadMethod);

  virtual void unloadFits();
  virtual void unloadAllFits() =0;
  void update(UpdateType);
  void update(UpdateType,BBox);
  virtual void updateBase();
  virtual void updateBin(const Matrix&) =0;
  virtual void updateColorScale() =0;
  virtual void updateGCs();
  void updateMagnifier();
  void updateMagnifier(const Vector&);
  virtual void updateMagnifierMatrix() =0;
  virtual void updateMagnifierZoom(const Vector& v) =0;
  virtual void updateMatrices();
  void updateNow(UpdateType); 
  void updateNow(UpdateType,BBox);
  virtual void updatePanner() =0;
  virtual void updatePM(const BBox&);
  virtual void updatePS(float, int, int) =0;

  char* varcat(char* buf, char* var, char r);
  virtual int validColorScale() =0;

  void ximageToPixmap(Pixmap pmap, XImage* xmap, 
		      double* (FitsImage::*getMatrix)());

  Vector zoom() {return zoom_;}

#ifdef _MACOSX
  void macosx(double* (FitsImage::*getMatrix)());
  void macosxImage(float, int, int, const Vector&, const Vector&,
		   double* (FitsImage::*getMatrix)());
#endif
#ifdef _WIN32
  void win32(double* (FitsImage::*getMatrix)());
  void win32Image(float, int, int, const Vector&, const Vector&,
		  double* (FitsImage::*getMatrix)());
#endif

 public:
  Base(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item);
  virtual ~Base();

  void calcAlignWCS(FitsImage*, CoordSystem, SkyFrame,
		    Orientation*, Matrix*, double*);
  void calcAlignWCS(FitsImage*, FitsImage*, CoordSystem, SkyFrame,
		    Orientation*, Matrix*, double*, Vector*);
  Matrix calcAlignWCS(FitsImage*, FitsImage*, CoordSystem, SkyFrame);

  FitsImage* findFits();
  FitsImage* findFits(CoordSystem, const Vector&);
  FitsImage* findFits(const Vector&);
  FitsImage* findFits(int);
  int findFits(FitsImage*);
  FitsImage* findAllFits(int);

  virtual Matrix getCanvasToRef() {return Matrix();}
  virtual SkyFrame getMarkerDefaultSky() {return FK5;}
  virtual CoordSystem getMarkerDefaultSystem() {return WCS;}

  Orientation IRAFOrientation(Orientation);
  int isMosaic();
  int isCube();

  double mapAngleFromRef(double, CoordSystem, SkyFrame =FK5);
  double mapAngleToRef(double, CoordSystem, SkyFrame =FK5);

  int parse(istringstream&);

  virtual void resetCompositeMarker() {}

  // Background Commands
  void bgColorCmd(const char*);
  void getBgColorCmd();

  // Bin Commands
  void binCmd(const Vector&, const char*, const char*, const char*);
  void binCmd(const Vector&, const Vector&, const char*, const char*, 
	      const char*);
  void binCmd(const Vector&, int, const Vector&, 
	      const char*, const char*, const char*, const char*);
  void binCmd(const Vector&, int, const Vector&, const Vector&, 
	      const char*, const char*, const char*, const char*);
  void binAboutCmd();
  void binAboutCmd(const Vector&);
  void binColsCmd(const char*, const char*, const char*);
  void binDepthCmd(int);
  void binFactorCmd(const Vector&);
  void binFactorAboutCmd(const Vector&, const Vector&);
  void binFactorToCmd(const Vector&);
  void binFactorToAboutCmd(const Vector&, const Vector&);
  void binFunctionCmd(FitsHist::Function);
  void binBufferSizeCmd(int);
  void binFilterCmd(const char*);
  void binToFitCmd();
  void getBinDepthCmd();
  void getBinFactorCmd();
  void getBinFunctionCmd();
  void getBinBufferSizeCmd();
  void getBinCursorCmd();
  void getBinFilterCmd();
  void getBinColsCmd();
  void getBinColsMinMaxCmd(const char*);
  void getBinColsDimCmd(const char*);
  void getBinListCmd();

  // Clip Commands
  void clipMinMaxModeCmd(FrScale::MinMaxMode);
  void clipMinMaxSampleCmd(int);
  void clipModeCmd(float);
  void clipModeCmd(FrScale::ClipMode);
  void clipPreserveCmd(int);
  void clipScopeCmd(FrScale::ClipScope);
  void clipUserCmd(double, double);
  void clipUserLowCmd(double);
  void clipUserHighCmd(double);
  void clipZScaleContrastCmd(float);
  void clipZScaleSampleCmd(int);
  void clipZScaleLineCmd(int);
  void getClipCmd();
  void getClipCmd(float);
  void getClipCmd(FrScale::ClipMode);
  void getClipMinMaxModeCmd();
  void getClipMinMaxSampleCmd();
  void getClipModeCmd();
  void getClipPreserveCmd();
  void getClipScopeCmd();
  void getClipUserCmd();
  void getClipZScaleContrastCmd();
  void getClipZScaleSampleCmd();
  void getClipZScaleLineCmd();
  void getHistogramCmd(char*, char*);
  void getMinMaxCmd();
  void hasDATAMINCmd();
  void hasIRAFMINCmd();

  // Colormap Commands
  virtual void colormapCmd(int, float, float, int, unsigned short*, 
			   unsigned char*, int) {}
  virtual void colormapCmd(float, float, float, float, float, float, int,
			   unsigned char*, int) {}
  virtual void colormapBeginCmd() {}
  virtual void colormapEndCmd() {}
  virtual void colormapMotionCmd(int, float, float, int, unsigned short*,
				 unsigned char*, int) {}
  virtual void colormapMotionCmd(float, float, float, float, float, float, int,
				 unsigned char*, int) {}
  virtual void getColorbarCmd() =0;

  // Contour Commands
  virtual void contourAuxHeadCmd() {}
  virtual void contourAuxNextCmd() {}
  virtual void contourAuxSaveCmd(const char*, CoordSystem, SkyFrame) {}
  virtual void contourCopyCmd(CoordSystem, SkyFrame) {}
  virtual void contourCreateCmd(const char*, int, int, FVContour::Method, int, int, FrScale::ColorScaleType, float, float, Vector, const char*) {}
  virtual void contourDeleteCmd() {}
  virtual void contourDeleteAllCmd() {}
  virtual void contourLevelCmd(const char*) {}
  virtual void contourLoadCmd(const char*, int, int, const char*, CoordSystem, SkyFrame) {}
  virtual void contourPasteCmd(const char*, int, int, void*, CoordSystem, SkyFrame) {}
  virtual void contourSaveCmd(const char*, CoordSystem, SkyFrame) {}
  virtual void contourSetColorCmd(const char*) {}
  virtual void contourSetDashCmd(int) {}
  virtual void contourSetLineWidthCmd(int) {}
  void getContourAuxColorNameCmd();
  void getContourAuxDashCmd();
  void getContourAuxLineWidthCmd();
  void getContourCmd(CoordSystem sys, SkyFrame sky);
  void getContourClipCmd();
  void getContourClipModeCmd();
  void getContourColorNameCmd();
  void getContourDashCmd();
  void getContourLevelCmd();
  void getContourLineWidthCmd();
  void getContourMethodCmd();
  void getContourNumLevelCmd();
  void getContourScaleCmd();
  void getContourScaleLogCmd();
  void getContourSmoothCmd();
  void hasContourCmd();
  void hasContourAuxCmd();

  // Coordinate Commands
  void getWCSNameCmd(CoordSystem);
  virtual void getWCSAlignCmd() =0;
  virtual void getCoordCmd(const Vector&, InternalSystem, CoordSystem, 
		   SkyFrame, SkyFormat) {}
  virtual void getCoordFromRefCmd(double, CoordSystem, int) {}
  virtual void getCoordToRefCmd(double, CoordSystem, int) {}
  virtual void getCursorCmd(InternalSystem) =0;
  virtual void getCursorCmd(CoordSystem, SkyFrame, SkyFormat, Precision) =0;
  void hasSystemCmd(CoordSystem);
  void hasAmplifierCmd();
  void hasDetectorCmd();
  void hasPhysicalCmd();
  void hasWCSCmd(CoordSystem);
  void hasWCSEquCmd(CoordSystem);
  void hasWCSLinearCmd(CoordSystem);

  // CrossHair Commands
  virtual void crosshairCmd(int) {}
  virtual void crosshairCmd(const Vector&, InternalSystem) {}
  virtual void crosshairCmd(const Vector&, CoordSystem, SkyFrame) {}
  virtual void crosshairBeginCmd(const Vector&, InternalSystem) {}
  virtual void crosshairBeginCmd(const Vector&, CoordSystem) {}
  virtual void crosshairMotionCmd(const Vector&, InternalSystem) {}
  virtual void crosshairMotionCmd(const Vector&, CoordSystem) {}
  virtual void crosshairWarpCmd(const Vector&) {}
  virtual void getCrosshairCmd(InternalSystem) {}
  virtual void getCrosshairCmd(CoordSystem, SkyFrame, SkyFormat, Precision) {}
  virtual void getCrosshairStatusCmd() {}
  virtual void getHorzCutCmd(char*, char*, const Vector&, InternalSystem) {}
  virtual void getVertCutCmd(char*, char*, const Vector&, InternalSystem) {}

  // Data Values
  virtual void getDataValuesCmd(int, const Vector&, CoordSystem, SkyFrame, 
			const Vector&, char*) {}
  virtual void getDataValuesCmd(const Vector&, InternalSystem, const Vector&) {}

  // DATASEC Commands
  void DATASECCmd(int);
  void getDATASECCmd();
  void hasDATASECCmd();

  // Fits Commands
  void getBitpixCmd();
  void getFitsCountCmd();
  virtual void getFitsCenterCmd(CoordSystem, SkyFrame, SkyFormat, Precision) {}
  void getFitsDepthCmd(int);
  virtual void getFitsExtCmd(const Vector&, InternalSystem) {}
  void getFitsFileNameCmd(FileNameType);
  void getFitsFileNameCmd(int,FileNameType);
  virtual void getFitsFileNameCmd(const Vector&, InternalSystem, FileNameType) {}
  void getFitsHeaderCmd(int);
  void getFitsHeaderWCSCmd(int);
  void getFitsHeaderKeywordCmd(int,const char*);
  void getFitsObjectNameCmd();
  void getFitsSizeCmd();
  virtual void getFitsSizeCmd(CoordSystem, SkyFrame, SkyFormat, Precision) {}
  void getFitsSliceCmd(int);
  void hasFitsCmd();
  void hasFitsBinCmd();
  void hasFitsCubeCmd();
  void hasFitsMosaicCmd();

  void loadIncrDataCmd(int, int, int, int, int);
  void loadIncrMinMaxCmd(int, int, int, int, int);
  void loadIncrEndCmd();

  void loadFitsAllocCmd(const char*, const char*, LayerType);
  void loadFitsAllocGZCmd(const char*, const char*, LayerType);
  void loadFitsChannelCmd(const char*, const char*, LayerType);
  void loadFitsMMapCmd(const char* fn, LoadMethod, LayerType);
  void loadFitsSMMapCmd(const char* hdr, const char* fn, LoadMethod,LayerType);
  void loadFitsMMapIncrCmd(const char* fn, LoadMethod, LayerType);
  void loadFitsShareCmd(ShmType, int, const char*, LoadMethod, LayerType);
  void loadFitsSShareCmd(ShmType, int, int, const char*, LoadMethod,LayerType);
  void loadFitsSocketCmd(int, const char* fn, LayerType);
  void loadFitsSocketGZCmd(int, const char* fn, LayerType);
  void loadFitsVarCmd(const char*, const char*, LoadMethod, LayerType);

  void loadArrAllocCmd(const char*, const char*, LayerType);
  void loadArrAllocGZCmd(const char*, const char*, LayerType);
  void loadArrChannelCmd(const char*, const char*, LayerType);
  void loadArrMMapCmd(const char* fn, LayerType);
  void loadArrMMapIncrCmd(const char* fn, LayerType);
  void loadArrShareCmd(ShmType, int, const char*, LayerType);
  void loadArrSocketCmd(int, const char* fn, LayerType);
  void loadArrSocketGZCmd(int, const char* fn, LayerType);
  void loadArrVarCmd(const char*, const char*, LayerType);

  void loadDataCubeAllocCmd(const char*, const char*, LayerType);
  void loadDataCubeAllocGZCmd(const char*, const char*, LayerType);
  void loadDataCubeChannelCmd(const char*, const char*, LayerType);
  void loadDataCubeMMapCmd(const char* fn, LoadMethod, LayerType);
  void loadDataCubeMMapIncrCmd(const char* fn, LoadMethod, LayerType);
  void loadDataCubeShareCmd(ShmType, int, const char*, LoadMethod, LayerType);
  void loadDataCubeSocketCmd(int, const char* fn, LayerType);
  void loadDataCubeSocketGZCmd(int, const char* fn, LayerType);
  void loadDataCubeVarCmd(const char*, const char*, LoadMethod, LayerType);

  void loadMosaicImageAllocCmd(MosaicType, CoordSystem,
			       const char*, const char*, LayerType);

  void loadMosaicImageAllocGZCmd(MosaicType, CoordSystem,
				 const char*, const char*, LayerType);
  void loadMosaicImageChannelCmd(MosaicType, CoordSystem,
				 const char*, const char*, LayerType);
  void loadMosaicImageMMapCmd(MosaicType, CoordSystem,
			      const char*, LoadMethod, LayerType);
  void loadMosaicImageMMapIncrCmd(MosaicType, CoordSystem,
				  const char*, LoadMethod, LayerType);
  void loadMosaicImageShareCmd(MosaicType, CoordSystem,
			       ShmType, int, const char*, LoadMethod, 
			       LayerType);
  void loadMosaicImageSocketCmd(MosaicType, CoordSystem,
				int, const char*, LayerType);
  void loadMosaicImageSocketGZCmd(MosaicType, CoordSystem, int,
				  const char*, LayerType);
  void loadMosaicImageVarCmd(MosaicType, CoordSystem,
			     const char*,const char*, LoadMethod, LayerType);

  void loadMosaicAllocCmd(MosaicType, CoordSystem,
			  const char*, const char*, LayerType);
  void loadMosaicAllocGZCmd(MosaicType, CoordSystem,
			    const char*, const char*, LayerType);
  void loadMosaicChannelCmd(MosaicType, CoordSystem,
			    const char*, const char*, LayerType);
  void loadMosaicMMapCmd(MosaicType, CoordSystem,
			 const char*, LoadMethod, LayerType);
  void loadMosaicSMMapCmd(MosaicType, CoordSystem, 
			  const char*, const char*, LoadMethod, LayerType);
  void loadMosaicMMapIncrCmd(MosaicType, CoordSystem,
			     const char*, LoadMethod, LayerType);
  void loadMosaicShareCmd(MosaicType, CoordSystem,
			  ShmType, int, const char*, LoadMethod, LayerType);
  void loadMosaicSShareCmd(MosaicType, CoordSystem,
			   ShmType, int, int, const char*, LoadMethod, 
			   LayerType);
  void loadMosaicSocketCmd(MosaicType, CoordSystem,
			   int, const char*, LayerType);
  void loadMosaicSocketGZCmd(MosaicType, CoordSystem,
			     int, const char*, LayerType);
  void loadMosaicVarCmd(MosaicType, CoordSystem,
			const char*, const char*, LoadMethod, LayerType);

  void loadMosaicImageWFPC2AllocCmd(const char*, const char*);
  void loadMosaicImageWFPC2AllocGZCmd(const char*, const char*);
  void loadMosaicImageWFPC2ChannelCmd(const char*, const char*);
  void loadMosaicImageWFPC2MMapCmd(const char* fn, LoadMethod);
  void loadMosaicImageWFPC2MMapIncrCmd(const char* fn, LoadMethod);
  void loadMosaicImageWFPC2ShareCmd(ShmType, int, const char*,LoadMethod);
  void loadMosaicImageWFPC2SocketCmd(int, const char* fn);
  void loadMosaicImageWFPC2SocketGZCmd(int, const char* fn);
  void loadMosaicImageWFPC2VarCmd(const char*, const char*, LoadMethod);

  // Fits RGB
  virtual void loadRGBCubeAllocCmd(const char*, const char*) {}
  virtual void loadRGBCubeAllocGZCmd(const char*, const char*) {}
  virtual void loadRGBCubeChannelCmd(const char*, const char*) {}
  virtual void loadRGBCubeMMapCmd(const char*, LoadMethod) {}
  virtual void loadRGBCubeSMMapCmd(const char*, const char*, LoadMethod) {}
  virtual void loadRGBCubeMMapIncrCmd(const char*, LoadMethod) {}
  virtual void loadRGBCubeShareCmd(ShmType, int, const char*, LoadMethod) {}
  virtual void loadRGBCubeSShareCmd(ShmType,int,int,const char*,LoadMethod) {}
  virtual void loadRGBCubeSocketCmd(int, const char*) {}
  virtual void loadRGBCubeSocketGZCmd(int, const char*) {}
  virtual void loadRGBCubeVarCmd(const char*, const char*, LoadMethod) {}

  virtual void loadRGBImageAllocCmd(const char*, const char*) {}
  virtual void loadRGBImageAllocGZCmd(const char*, const char*) {}
  virtual void loadRGBImageChannelCmd(const char*, const char*) {}
  virtual void loadRGBImageMMapCmd(const char*, LoadMethod) {}
  virtual void loadRGBImageMMapIncrCmd(const char*, LoadMethod) {}
  virtual void loadRGBImageShareCmd(ShmType, int, const char*, LoadMethod) {}
  virtual void loadRGBImageSocketCmd(int, const char*) {}
  virtual void loadRGBImageSocketGZCmd(int, const char*) {}
  virtual void loadRGBImageVarCmd(const char*, const char*, LoadMethod) {}

  virtual void loadArrRGBCubeAllocCmd(const char*, const char*) {}
  virtual void loadArrRGBCubeAllocGZCmd(const char*, const char*) {}
  virtual void loadArrRGBCubeChannelCmd(const char*, const char*) {}
  virtual void loadArrRGBCubeMMapCmd(const char*) {}
  virtual void loadArrRGBCubeMMapIncrCmd(const char*) {}
  virtual void loadArrRGBCubeShareCmd(ShmType, int, const char*) {}
  virtual void loadArrRGBCubeSocketCmd(int, const char*) {}
  virtual void loadArrRGBCubeSocketGZCmd(int, const char*) {}
  virtual void loadArrRGBCubeVarCmd(const char*, const char*) {}

  void saveFitsImageFileCmd(const char*, int);
  void saveFitsImageChannelCmd(const char*, int);
  void saveFitsImageSocketCmd(int, int);
  virtual void saveFitsResampleFileCmd(const char*, int) {}
  virtual void saveFitsResampleChannelCmd(const char*, int) {}
  virtual void saveFitsResampleSocketCmd(int, int) {}
  void saveFitsTableFileCmd(const char*, int);
  void saveFitsTableChannelCmd(const char*, int);
  void saveFitsTableSocketCmd(int, int);
  void saveArrayFileCmd(const char*, FitsFile::ArchType);
  void saveArrayChannelCmd(const char*, FitsFile::ArchType);
  void saveArraySocketCmd(int, FitsFile::ArchType);
  virtual void sliceCmd(int, int) =0;

  void updateFitsCmd(int);
  virtual void updateFitsCmd(int, BBox, int) {}
  void unloadFitsCmd();

  // Fitsy Commands
  void fitsyHasExtCmd(const char*);

  // Frame Commands
  void clearCmd();
  void getHeightCmd();
  virtual void getTypeCmd() =0;
  void getWidthCmd();
  void highliteCmd(int);
  void warpCmd(const Vector&);
  void warpToCmd(const Vector&);

  // Grid Commands
  virtual void gridCmd(CoordSystem, SkyFrame, SkyFormat, Grid::GridType, const char*) {}
  virtual void gridDeleteCmd() {}
  virtual void getGridCmd() {}
  virtual void getGridOptionCmd() {}
  void hasGridCmd();

  // IIS Commands
  virtual void iisCmd(int, int) {}
  virtual void iisCursorModeCmd(int) {}
  virtual void iisEraseCmd() {}
  virtual void iisGetCmd(char*, int, int, int, int) {}
  virtual void iisGetCursorCmd() {}
  virtual void iisGetFileNameCmd() {}
  virtual void iisGetFileNameCmd(int) {}
  virtual void iisGetFileNameCmd(const Vector&) {}
  virtual void iisMessageCmd(const char*) {}
  virtual void iisSetCmd(const char*, int, int, int, int) {}
  virtual void iisSetCursorCmd(const Vector&, InternalSystem) {}
  virtual void iisSetCursorCmd(const Vector&, CoordSystem) {}
  virtual void iisSetFileNameCmd(const char*) {}
  virtual void iisSetFileNameCmd(const char*,int) {}
  virtual void iisUpdateCmd() {}
  virtual void iisWCSCmd(const Matrix&, const Vector&, int) {}
  void hasIISCmd();

  // Info Commands
  void getInfoCmd(char*);
  void getInfoCmd(const Vector&, InternalSystem, SkyFrame, SkyFormat, char*);
  void getInfoClipCmd();
  virtual void getValueCmd(const Vector&, InternalSystem) {}

  // Magnifier Commands
  void magnifierCmd(char*, int, int);
  void magnifierCmd(int);
  void magnifierGraphicsCmd(int);
  void magnifierCursorCmd(int);
  void magnifierZoomCmd(double);
  void magnifierColorCmd(const char*);
  void updateMagnifierCmd();
  void updateMagnifierCmd(const Vector& v);

  // Marker Commands
  virtual void createCircleCmd(const Vector&, 
			       double,
			       const char*, int*, int, const char*, 
			       const char*, unsigned short, const char*,
			       const List<Tag>&, const List<CallBack>& cb) {}
  virtual void createAnnulusCmd(const Vector&, 
				double, double, int,
				const char*, int*, int, const char*, 
				const char*, unsigned short, const char*,
				const List<Tag>&, const List<CallBack>& cb) {}
  virtual void createAnnulusCmd(const Vector&, 
				int, double*,
				const char*, int*, int, const char*, 
				const char*, unsigned short, const char*,
				const List<Tag>&, const List<CallBack>& cb) {}
  virtual void createCpandaCmd(const Vector&, 
			       double, double, int, 
			       double, double, int,
			       const char*, int*, int, const char*, 
			       const char*, unsigned short, const char*,
			       const List<Tag>&, const List<CallBack>& cb) {}
  virtual void createCpandaCmd(const Vector&, 
			       int, double*,
			       int, double*,
			       const char*, int*, int, const char*, 
			       const char*, unsigned short, const char*,
			       const List<Tag>&, const List<CallBack>& cb) {}


  virtual void createEllipseCmd(const Vector&, 
				const Vector&, 
				double, 
				const char*, int*, int, const char*, 
				const char*, unsigned short, const char*,
				const List<Tag>&, const List<CallBack>& cb) {}
  virtual void createEllipseAnnulusCmd(const Vector&, 
				       const Vector&, const Vector&, int, 
				       double,
				       const char*, int*, int, const char*, 
				       const char*, unsigned short, const char*,
				       const List<Tag>&, 
				       const List<CallBack>& cb) {}
  virtual void createEllipseAnnulusCmd(const Vector&, 
				       int, Vector*,
				       double, 
				       const char*, int*, int,  const char*, 
				       const char*, unsigned short, const char*,
				       const List<Tag>&, 
				       const List<CallBack>& cb) {}
  virtual void createEpandaCmd(const Vector&, 
			       double, double, int, 
			       const Vector&, const Vector&, int, 
			       double,
			       const char*, int*, int, const char*, 
			       const char*, unsigned short, const char*,
			       const List<Tag>&, const List<CallBack>& cb) {}
  virtual void createEpandaCmd(const Vector&, 
			       int, double*,
			       int, Vector*,
			       double,
			       const char*, int*, int, const char*, 
			       const char*, unsigned short, const char*,
			       const List<Tag>&, const List<CallBack>& cb) {}


  virtual void createBoxCmd(const Vector&, 
			    const Vector&, 
			    double, 
			    const char*, int*, int, const char*, 
			    const char*, unsigned short, const char*,
			    const List<Tag>&, const List<CallBack>& cb) {}
  virtual void createBoxAnnulusCmd(const Vector&, 
				   const Vector&, const Vector&, int, 
				   double,
				   const char*, int*, int, const char*, 
				   const char*, unsigned short, const char*,
				   const List<Tag>&, 
				   const List<CallBack>& cb) {}
  virtual void createBoxAnnulusCmd(const Vector&, 
				   int, Vector*,
				   double, 
				   const char*, int*, int, const char*, 
				   const char*, unsigned short, const char*,
				   const List<Tag>&, 
				   const List<CallBack>& cb) {}
  virtual void createBpandaCmd(const Vector&, 
			       double, double, int, 
			       const Vector&, const Vector&, int, 
			       double,
			       const char*, int*, int, const char*, 
			       const char*, unsigned short, const char*,
			       const List<Tag>&, const List<CallBack>& cb) {}
  virtual void createBpandaCmd(const Vector&, 
			       int, double*,
			       int, Vector*,
			       double,
			       const char*, int*, int, const char*, 
			       const char*, unsigned short, const char*,
			       const List<Tag>&, const List<CallBack>& cb) {}


  virtual void createLineCmd(const Vector&, 
			     const Vector&, 
			     int, int,
			     const char*, int*, int, const char*,
			     const char*, unsigned short, const char*,
			     const List<Tag>&, const List<CallBack>& cb) {}
  virtual void createVectCmd(const Vector&, 
			     const Vector&, 
			     int,
			     const char*, int*, int, const char*,
			     const char*, unsigned short, const char*,
			     const List<Tag>&, const List<CallBack>& cb) {}
  virtual void createVectCmd(const Vector&, 
			     double mag, double ang, 
			     int,
			     const char*, int*, int, const char*,
			     const char*, unsigned short, const char*,
			     const List<Tag>&, const List<CallBack>& cb) {}
  virtual void createProjectionCmd(const Vector&, 
				   const Vector&,
				   double, 
				   const char*, const char*,
				   const char*, int*, int, const char*,
				   const char*, unsigned short, const char*,
				   const List<Tag>&, 
				   const List<CallBack>& cb) {}
  virtual void createCircle3dCmd(const Vector&, 
				   double, 
				   const char*, const char*,
				   const char*, int*, int, const char*,
				   const char*, unsigned short, const char*,
				   const List<Tag>&, 
				   const List<CallBack>& cb) {}
  virtual void createRulerCmd(const Vector&, 
			      const Vector&, 
			      CoordSystem, SkyFrame, CoordSystem, SkyFormat,
			      const char*, int*, int, const char*,
			      const char*, unsigned short, const char*,
			      const List<Tag>&, const List<CallBack>& cb) {}

  virtual void createCirclePointCmd(const Vector&, int,
				    const char*, int*, int, const char*,
				    const char*, unsigned short, const char*,
				    const List<Tag>&, 
				    const List<CallBack>& cb) {}
  virtual void createBoxPointCmd(const Vector&, int,
				 const char*, int*, int, const char*,
				 const char*, unsigned short, const char*,
				 const List<Tag>&, const List<CallBack>& cb) {}
  virtual void createDiamondPointCmd(const Vector&, int,
				     const char*, int*, int, const char*, 
				     const char*, unsigned short, const char*,
				     const List<Tag>&, 
				     const List<CallBack>& cb) {}
  virtual void createCrossPointCmd(const Vector&, int,
				   const char*, int*, int, const char*,
				   const char*, unsigned short, const char*,
				   const List<Tag>&, 
				   const List<CallBack>& cb) {}
  virtual void createExPointCmd(const Vector&, int,
				const char*, int*, int, const char*,
				const char*, unsigned short, const char*,
				const List<Tag>&, const List<CallBack>& cb) {}
  virtual void createArrowPointCmd(const Vector&, int,
				   const char*, int*, int, const char*,
				   const char*, unsigned short, const char*,
				   const List<Tag>&, 
				   const List<CallBack>& cb) {}
  virtual void createBoxCirclePointCmd(const Vector&, int,
				       const char*, int*, int, const char*,
				       const char*, unsigned short, const char*,
				       const List<Tag>&, 
				       const List<CallBack>& cb) {}

  virtual void createTextCmd(const Vector&, 
			     double, int,
			     const char*, int*, int, const char*, 
			     const char*, unsigned short, const char*,
			     const List<Tag>&, const List<CallBack>& cb) {}

  virtual void createPolygonCmd(const Vector&, 
				const Vector&,
				const char*, int*, int, const char*, 
				const char*, unsigned short, const char*,
				const List<Tag>&, const List<CallBack>& cb) {}
  virtual void createPolygonCmd(const List<Vertex>&,
				const char*, int*, int, const char*, 
				const char*, unsigned short, const char*,
				const List<Tag>&, const List<CallBack>& cb) {}
  virtual void createContourPolygonCmd(const char*, int*, int, const char*, 
				       const char*, unsigned short, const char*,
				       const List<Tag>&, 
				       const List<CallBack>& cb) {}

  virtual void createCompassCmd(const Vector&, 
				double, 
				const char*, const char*,
				int, int,
				CoordSystem, SkyFrame,
				const char*, int*, int, const char*,
				const char*, unsigned short, const char*,
				const List<Tag>&, const List<CallBack>& cb) {}

  virtual void createCompositeCmd(const Vector&, double, int,
				  const char*, int*, int, const char*, 
				  const char*, unsigned short, const char*,
				  const List<Tag>&, const List<CallBack>& cb) {}
  virtual void createCompositeCmd(
				  const char*, int*, int, const char*, 
				  const char*, unsigned short, const char*,
				  const List<Tag>&, const List<CallBack>& cb) {}

  virtual void createTemplateCmd(const Vector&, const char*) {}
  virtual void createTemplateCmd(const Vector&, CoordSystem, SkyFrame, const char*) {}
  virtual void createTemplateVarCmd(const Vector&, const char* var) {}

  virtual void getMarkerAngleCmd(int) {}
  virtual void getMarkerAngleCmd(int, CoordSystem, SkyFrame) {}
  virtual void getMarkerAnnulusRadiusCmd(int, InternalSystem) {}
  virtual void getMarkerAnnulusRadiusCmd(int, CoordSystem, SkyFormat) {}
  virtual void getMarkerBoxAnnulusRadiusCmd(int, InternalSystem) {}
  virtual void getMarkerBoxAnnulusRadiusCmd(int, CoordSystem, SkyFormat) {}
  virtual void getMarkerBoxRadiusCmd(int, InternalSystem) {}
  virtual void getMarkerBoxRadiusCmd(int, CoordSystem, SkyFormat) {}
  virtual void getMarkerBpandaAnglesCmd(int) {}
  virtual void getMarkerBpandaAnglesCmd(int, CoordSystem, SkyFrame) {}
  virtual void getMarkerBpandaRadiusCmd(int, InternalSystem) {}
  virtual void getMarkerBpandaRadiusCmd(int, CoordSystem, SkyFormat) {}
  virtual void getMarkerCenterCmd(int, InternalSystem) {}
  virtual void getMarkerCenterCmd(int, CoordSystem, SkyFrame, SkyFormat) {}
  void getMarkerCentroidAutoCmd(); // frmarker.h
  void getMarkerCentroidRadiusCmd(); // frmarker.h
  void getMarkerCentroidIterationCmd(); // frmarker.h
  void getMarkerCentroidOptionCmd(); // frmarker.h
  virtual void getMarkerCircleRadiusCmd(int, InternalSystem) {}
  virtual void getMarkerCircleRadiusCmd(int, CoordSystem, SkyFormat) {}
  virtual void getMarkerColorCmd() {}
  virtual void getMarkerColorCmd(const char*) {}
  virtual void getMarkerColorCmd(int) {}
  virtual void getMarkerCompassArrowCmd(int) {}
  virtual void getMarkerCompassLabelCmd(int) {}
  virtual void getMarkerCompassRadiusCmd(int, InternalSystem) {}
  virtual void getMarkerCompassRadiusCmd(int, CoordSystem, SkyFormat) {}
  virtual void getMarkerCompassSystemCmd(int) {}
  virtual void getMarkerCompositeCmd(int) {}
  virtual void getMarkerCpandaAnglesCmd(int) {}
  virtual void getMarkerCpandaAnglesCmd(int, CoordSystem, SkyFrame) {}
  virtual void getMarkerCpandaRadiusCmd(int, InternalSystem) {}
  virtual void getMarkerCpandaRadiusCmd(int, CoordSystem, SkyFormat) {}
  virtual void getMarkerEllipseRadiusCmd(int, InternalSystem) {}
  virtual void getMarkerEllipseRadiusCmd(int, CoordSystem, SkyFormat) {}
  virtual void getMarkerEllipseAnnulusRadiusCmd(int, InternalSystem) {}
  virtual void getMarkerEllipseAnnulusRadiusCmd(int, CoordSystem, SkyFormat) {}
  virtual void getMarkerEpandaAnglesCmd(int) {}
  virtual void getMarkerEpandaAnglesCmd(int, CoordSystem, SkyFrame) {}
  virtual void getMarkerEpandaRadiusCmd(int, InternalSystem) {}
  virtual void getMarkerEpandaRadiusCmd(int, CoordSystem, SkyFormat) {}
  virtual void getMarkerFontCmd() {}
  virtual void getMarkerFontCmd(const char*) {}
  virtual void getMarkerFontCmd(int) {}
  virtual void getMarkerIdCmd(const char*) {}
  virtual void getMarkerIdCmd(const Vector&) {}
  virtual void getMarkerIdAllCmd() {}
  virtual void getMarkerLineCmd(int, InternalSystem) {}
  virtual void getMarkerLineCmd(int, CoordSystem, SkyFrame, SkyFormat) {}
  virtual void getMarkerLineArrowCmd(int) {}
  virtual void getMarkerLineLengthCmd(int, InternalSystem) {}
  virtual void getMarkerLineLengthCmd(int, CoordSystem, SkyFormat) {}
  virtual void getMarkerLineWidthCmd() {}
  virtual void getMarkerLineWidthCmd(int) {}
  virtual void getMarkerMapLenFromRefCmd(int, double, CoordSystem, SkyFormat) {}
  virtual void getMarkerNumberCmd() {}
  virtual void getMarkerPointSizeCmd(int) {}
  virtual void getMarkerPolygonSegmentCmd(const Vector&) {}
  void getMarkerPreserveCmd(); // frmarker.h
  virtual void getMarkerProjectionCmd(int, char*, char*, char*, char*) {}
  virtual void getMarkerProjectionPointsCmd(int, InternalSystem) {}
  virtual void getMarkerProjectionPointsCmd(int, CoordSystem, SkyFrame, SkyFormat) {}
  virtual void getMarkerProjectionLengthCmd(int, InternalSystem) {}
  virtual void getMarkerProjectionLengthCmd(int, CoordSystem, SkyFormat) {}
  virtual void getMarkerProjectionMethodCmd(int) {}
  virtual void getMarkerProjectionWidthCmd(int, InternalSystem) {}
  virtual void getMarkerProjectionWidthCmd(int, CoordSystem, SkyFormat) {}
  virtual void getMarkerProjectionSystemCmd(int) {}
  virtual void getMarkerCircle3dCmd(int, char*, char*, char*) {}
  virtual void getMarkerCircle3dMethodCmd(int) {}
  virtual void getMarkerCircle3dRadiusCmd(int, CoordSystem, SkyFormat) {}
  virtual void getMarkerCircle3dSystemCmd(int) {}
  virtual void getMarkerPropertyCmd(unsigned short) {}
  virtual void getMarkerPropertyCmd(const char*, unsigned short) {}
  virtual void getMarkerPropertyCmd(int, unsigned short) {}
  virtual void getMarkerRulerLengthCmd(int, InternalSystem) {}
  virtual void getMarkerRulerLengthCmd(int, CoordSystem, SkyFormat) {}
  virtual void getMarkerRulerPointCmd(int, InternalSystem) {}
  virtual void getMarkerRulerPointCmd(int, CoordSystem, SkyFrame, SkyFormat) {}
  virtual void getMarkerRulerSystemCmd(int) {}
  virtual void getMarkerSelectedCmd() {}
  virtual void getMarkerSelectedCmd(int) {}
  virtual void getMarkerSelectedCmd(const Vector&) {}
  virtual void getMarkerSelectedNumberCmd() {}
  void getMarkerShowCmd(); // frmarker.h
  void getMarkerShowTextCmd(); // frmarker.h
  virtual void getMarkerHandleCmd(const Vector&) {}
  virtual void getMarkerHighlitedCmd() {}
  virtual void getMarkerHighlitedCmd(int) {}
  virtual void getMarkerHighlitedCmd(const Vector&) {}
  virtual void getMarkerHighlitedNumberCmd() {}
  virtual void getMarkerTagCmd(const char*) {}
  virtual void getMarkerTagCmd(int) {}
  virtual void getMarkerTagCmd(int,int) {}
  virtual void getMarkerTagDefaultNameCmd() {}
  virtual void getMarkerTagNumberCmd(const char*) {}
  virtual void getMarkerTagsCmd() {}
  virtual void getMarkerTextCmd(int) {}
  virtual void getMarkerTextRotateCmd(int) {}
  virtual void getMarkerTypeCmd(int) {}
  virtual void getMarkerVectorCmd(int, InternalSystem) {}
  virtual void getMarkerVectorCmd(int, CoordSystem, SkyFrame, SkyFormat) {}
  virtual void getMarkerVectorArrowCmd(int) {}
  virtual void getMarkerVectorLengthCmd(int, InternalSystem) {}
  virtual void getMarkerVectorLengthCmd(int, CoordSystem, SkyFormat) {}

  virtual void hasMarkerHighlitedCmd() =0;
  virtual void hasMarkerPasteCmd() =0;
  virtual void hasMarkerSelectedCmd() =0;
  virtual void hasMarkerUndoCmd() =0;

  virtual void markerLayerCmd(MarkerLayer) {}

  virtual void markerAngleCmd(int, double) {}
  virtual void markerAngleCmd(int, double, CoordSystem, SkyFrame) {}
  virtual void markerAnnulusCreateRadiusCmd(int, const Vector&) {}
  virtual void markerAnnulusDeleteRadiusCmd(int, int) {}
  virtual void markerAnnulusRadiusCmd(int, double, double, int, InternalSystem) {}
  virtual void markerAnnulusRadiusCmd(int, double, double, int, CoordSystem,SkyFormat) {}
  virtual void markerAnnulusRadiusCmd(int, const char*, CoordSystem, SkyFormat) {}
  virtual void markerBackCmd() {}
  virtual void markerBackCmd(const char*) {}
  virtual void markerBackCmd(int) {}
  virtual void markerBoxAnnulusRadiusCmd(int, const Vector&, const Vector&, int, 
				 InternalSystem) {}
  virtual void markerBoxAnnulusRadiusCmd(int, const Vector&, const Vector&, int, 
				 CoordSystem, SkyFormat) {}
  virtual void markerBoxAnnulusRadiusCmd(int, const char*, CoordSystem, SkyFormat) {}
  virtual void markerBoxAnnulusCreateRadiusCmd(int, const Vector&) {}
  virtual void markerBoxAnnulusDeleteRadiusCmd(int, int) {}
  virtual void markerBoxRadiusCmd(int, const Vector&, InternalSystem) {}
  virtual void markerBoxRadiusCmd(int, const Vector&, CoordSystem, SkyFormat) {}
  virtual void markerBpandaCreateAnglesCmd(int, const Vector&) {}
  virtual void markerBpandaCreateRadiusCmd(int, const Vector&) {}
  virtual void markerBpandaDeleteCmd(int, int) {}
  virtual void markerBpandaEditCmd(int, double, double, int, 
			   const Vector&, const Vector&, int) {}
  virtual void markerBpandaEditCmd(int, double, double, int, 
			   const Vector&, const Vector&, int,
			   CoordSystem, SkyFrame) {}
  virtual void markerBpandaEditCmd(int, const char*, const char*, 
			   CoordSystem, SkyFrame,
			   CoordSystem, SkyFormat) {}
  virtual void markerCallBackCmd(int, CallBack::Type, const char*, const char*) {}
  virtual void markerCentroidCmd() {}
  virtual void markerCentroidCmd(const char*) {}
  virtual void markerCentroidCmd(int) {}
  virtual void markerCentroidAutoCmd(int) {}
  virtual void markerCentroidRadiusCmd(float) {}
  virtual void markerCentroidIterationCmd(int) {}
  virtual void markerCircleRadiusCmd(int, double, InternalSystem) {}
  virtual void markerCircleRadiusCmd(int, double, CoordSystem, SkyFormat) {}
  virtual void markerColorCmd(const char*) {}
  virtual void markerColorCmd(const char*, const char*) {}
  virtual void markerColorCmd(int, const char*) {}
  virtual void markerCompassArrowCmd(int, int, int) {}
  virtual void markerCompassLabelCmd(int, const char*, const char*) {}
  virtual void markerCompassRadiusCmd(int, double, InternalSystem) {}
  virtual void markerCompassRadiusCmd(int, double, CoordSystem, SkyFormat) {}
  virtual void markerCompassSystemCmd(int, CoordSystem, SkyFrame) {}
  virtual void markerCompositeCmd(int, int) {}
  virtual void markerCompositeDeleteCmd() {}
  virtual void markerCommandCmd(MarkerFormat, const char*) {}
  virtual void markerCommandVarCmd(MarkerFormat, const char*) {}
  virtual void markerCpandaCreateAnglesCmd(int, const Vector&) {}
  virtual void markerCpandaCreateRadiusCmd(int, const Vector&) {}
  virtual void markerCpandaDeleteCmd(int, int) {}
  virtual void markerCpandaEditCmd(int, double, double, int, double, double, int) {}
  virtual void markerCpandaEditCmd(int, double, double, int, double, double, int,
			  CoordSystem, SkyFrame) {}
  virtual void markerCpandaEditCmd(int, const char*, const char*, 
			   CoordSystem, SkyFrame,CoordSystem, SkyFormat) {}
  virtual void markerDeleteCmd() {}
  virtual void markerDeleteCmd(const char*) {}
  virtual void markerDeleteCmd(int) {}
  virtual void markerDeleteAllCmd() {}
  virtual void markerDeleteLastCmd() {}
  virtual void markerDeleteCallBackCmd(int, CallBack::Type, const char*) {}
  virtual void markerDeleteTagCmd(int) {}
  virtual void markerDeleteTagCmd(int, const char*) {}
  virtual void markerDeleteTagCmd(int, int) {}
  virtual void markerEditBeginCmd(int, int) {}
  virtual void markerEditBeginCmd(const Vector&, int) {}
  virtual void markerEditMotionCmd(const Vector&, int h) {}
  virtual void markerEditEndCmd() {}
  virtual void markerEllipseRadiusCmd(int, const Vector&, InternalSystem) {}
  virtual void markerEllipseRadiusCmd(int, const Vector&, CoordSystem, SkyFormat) {}
  virtual void markerEllipseAnnulusRadiusCmd(int, const Vector&, const Vector&, int, 
				     InternalSystem) {}
  virtual void markerEllipseAnnulusRadiusCmd(int, const Vector&, const Vector&, int, 
				     CoordSystem, SkyFormat) {}
  virtual void markerEllipseAnnulusRadiusCmd(int, const char*, 
				     CoordSystem, SkyFormat) {}
  virtual void markerEllipseAnnulusCreateRadiusCmd(int, const Vector&) {}
  virtual void markerEllipseAnnulusDeleteRadiusCmd(int, int) {}
  virtual void markerEpandaCreateAnglesCmd(int, const Vector&) {}
  virtual void markerEpandaCreateRadiusCmd(int, const Vector&) {}
  virtual void markerEpandaDeleteCmd(int, int) {}
  virtual void markerEpandaEditCmd(int, double, double, int, 
			   const Vector&, const Vector&, int) {}
  virtual void markerEpandaEditCmd(int, double, double, int, 
			   const Vector&, const Vector&, int,
			   CoordSystem, SkyFrame) {}
  virtual void markerEpandaEditCmd(int, const char*, const char*, 
			   CoordSystem, SkyFrame,
			   CoordSystem, SkyFormat) {}
  virtual void markerFontCmd(const char*) {}
  virtual void markerFontCmd(const char*, const char*) {}
  virtual void markerFontCmd(int, const char*) {}
  virtual void markerFrontCmd() {}
  virtual void markerFrontCmd(const char*) {}
  virtual void markerFrontCmd(int) {}
  virtual void markerKeyCmd() {}
  virtual void markerKeyCmd(const Vector&) {}
  virtual void markerLineCmd(int, const Vector&, const Vector&, InternalSystem) {}
  virtual void markerLineCmd(int, const Vector&, const Vector&, CoordSystem, SkyFrame) {}
  virtual void markerLineArrowCmd(int, int, int) {}
  virtual void markerLineWidthCmd(int) {}
  virtual void markerLineWidthCmd(int, int) {}
  virtual void markerListCmd(MarkerFormat, CoordSystem, SkyFrame, SkyFormat, 
			     int strip, int select,
			     unsigned short, unsigned short, List<Tag>&) {}
  virtual void markerLoadCmd(MarkerFormat,const char*) {}
  virtual void markerLoadCmd(MarkerFormat,const char*,CoordSystem,SkyFrame) {}
  virtual void markerLoadCmd(MarkerFormat,int) {}
  virtual void markerLoadCmd(MarkerFormat,int,CoordSystem,SkyFrame) {}
  virtual void markerLoadFitsCmd(const char*, const char*, int*, int, const char*) {}
  virtual void markerMoveCmd(const Vector&) {}
  virtual void markerMoveCmd(const char*, const Vector&) {}
  virtual void markerMoveCmd(int id, const Vector&) {}
  virtual void markerMoveBeginCmd(const Vector&) {}
  virtual void markerMoveBeginCmd(int, const Vector&) {}
  virtual void markerMoveMotionCmd(const Vector&) {}
  virtual void markerMoveMotionCmd(int, const Vector&) {}
  virtual void markerMoveEndCmd() {}
  virtual void markerMoveEndCmd(int) {}
  virtual void markerMoveToCmd(const Vector&, InternalSystem) {}
  virtual void markerMoveToCmd(const Vector&, CoordSystem, SkyFrame) {}
  virtual void markerMoveToCmd(const char*, const Vector&, InternalSystem) {}
  virtual void markerMoveToCmd(const char*, const Vector&, CoordSystem, SkyFrame) {}
  virtual void markerMoveToCmd(int, const Vector&, InternalSystem) {}
  virtual void markerMoveToCmd(int, const Vector&, CoordSystem, SkyFrame) {}
  virtual void markerPointSizeCmd(int,int) {}
  virtual void markerPolygonCreateVertexCmd(int, int, const Vector&) {}
  virtual void markerPolygonDeleteVertexCmd(int, int) {}
  virtual void markerPolygonResetCmd(int, const Vector&, InternalSystem) {}
  virtual void markerPolygonResetCmd(int, const Vector&, CoordSystem, SkyFormat) {}
  void markerPreserveCmd(int r) {preserveMarkers = r;}
  virtual void markerPropertyCmd(unsigned short, int) {}
  virtual void markerPropertyCmd(const char*, unsigned short, int) {}
  virtual void markerPropertyCmd(int, unsigned short, int) {}
  virtual void markerPropertyCmd(unsigned short, int, const Vector&) {}
  virtual void markerProjectionCmd(int, const Vector&, const Vector&, 
				   InternalSystem, double, int) {}
  virtual void markerProjectionCmd(int, const Vector&, const Vector&, 
				   CoordSystem, SkyFrame, 
				   double, int, CoordSystem, SkyFormat) {}
  virtual void markerProjectionSystemCmd(int, CoordSystem, SkyFrame) {}
  virtual void markerCircle3dCmd(int, double, InternalSystem, int) {}
  virtual void markerCircle3dCmd(int, double, CoordSystem, SkyFormat,int) {}
  virtual void markerCircle3dSystemCmd(int, CoordSystem) {}
  virtual void markerRotateBeginCmd(int) {}
  virtual void markerRotateBeginCmd(const Vector&) {}
  virtual void markerRotateMotionCmd(const Vector&, int h) {}
  virtual void markerRotateEndCmd() {}
  virtual void markerRulerPointCmd(int, const Vector&, const Vector&, 
			   CoordSystem, SkyFrame) {}
  virtual void markerRulerPointCmd(int, const Vector&, const Vector&, InternalSystem) {}
  virtual void markerRulerSystemCmd(int, CoordSystem, SkyFrame, 
			    CoordSystem, SkyFormat) {}
  virtual void markerSaveCmd(const char*, MarkerFormat, CoordSystem, SkyFrame,
			     SkyFormat, int strip) {}
  virtual void markerSaveTemplateCmd(const char*) {}

  virtual void markerVectorCmd(int, const Vector&, InternalSystem, double, double) {}
  virtual void markerVectorCmd(int, const Vector&, CoordSystem, SkyFrame, 
		       double, CoordSystem, SkyFormat, double) {}
  virtual void markerVectorArrowCmd(int, int) {}

  virtual void markerSelectAllCmd() {}
  virtual void markerSelectCmd(int) {} // not used
  virtual void markerSelectCmd(const char*) {}
  virtual void markerSelectToggleCmd() {}
  virtual void markerSelectToggleCmd(const Vector&) {}
  virtual void markerSelectOnlyCmd(int) {}
  virtual void markerSelectOnlyCmd(const char*) {}
  virtual void markerSelectOnlyCmd(const Vector&) {}
  virtual void markerUnselectAllCmd() {}
  virtual void markerUnselectCmd(int) {} // not used
  virtual void markerUnselectCmd(const char*) {} // not used

  virtual void markerHighliteAllCmd() {} // not used
  virtual void markerHighliteCmd(int) {} // not used
  virtual void markerHighliteCmd(const char*) {} // not used
  virtual void markerHighliteToggleCmd(const Vector&) {}
  virtual void markerHighliteOnlyCmd(int) {}
  virtual void markerHighliteOnlyCmd(const char*) {} // not used
  virtual void markerHighliteOnlyCmd(const Vector&) {}  // not used
  virtual void markerUnhighliteAllCmd() {}
  virtual void markerUnhighliteCmd(int) {} // not used
  virtual void markerUnhighliteCmd(const char*) {}

  virtual void markerShowCmd(int) {}
  virtual void markerShowTextCmd(int) {}
  virtual void markerTagCmd(const char*) {}
  virtual void markerTagCmd(int, const char*) {}
  virtual void markerTagEditCmd(const char*, const char*) {}
  virtual void markerTagDeleteCmd(const char*) {}
  virtual void markerTagDeleteAllCmd() {}
  virtual void markerTagUpdateCmd(const char*) {}
  virtual void markerTextCmd(int, const char*) {}
  virtual void markerTextRotateCmd(int, int) {}

  virtual void markerCopyCmd() {}
  virtual void markerCopyCmd(const char*) {}
  virtual void markerCutCmd() {}
  virtual void markerCutCmd(const char*) {}
  virtual void markerPasteCmd() {}
  virtual void markerPasteCmd(CoordSystem, SkyFrame) {}
  virtual void markerUndoCmd() {}

  virtual void regionHighliteBeginCmd(const Vector& v) {}
  virtual void regionHighliteMotionCmd(const Vector& v) {}
  virtual void regionHighliteEndCmd() {}
  virtual void regionHighliteShiftEndCmd() {}

  virtual void regionSelectBeginCmd(const Vector&) {}
  virtual void regionSelectMotionCmd(const Vector&) {}
  virtual void regionSelectEndCmd() {}
  virtual void regionSelectShiftEndCmd() {}

  // Mask Commands
  void getMaskColorCmd();
  void getMaskMarkCmd();
  void getMaskTransparencyCmd();

  void maskColorCmd(const char*);
  void maskClearCmd();
  void maskMarkCmd(int m) {maskMark=m;}
  void maskTransparencyCmd(float);

  // NaN Commands
  void getNANColorCmd();
  void nanColorCmd(const char*);

  // Pan Zoom Rotate Orient Commands
  void centerCmd();
  void getOrientCmd();
  void getRotateCmd(Precision);
  void getPanPreserveCmd();
  void getZoomCmd(Precision);
  void orientCmd(Orientation);
  virtual void panCmd(const Vector&, const Vector&) {}
  virtual void panCmd(const Vector&, InternalSystem) {}
  virtual void panCmd(const Vector&, CoordSystem, SkyFrame) {}
  virtual void panBBoxCmd(const Vector&) {}
  virtual void panBeginCmd(const Vector&) {}
  virtual void panMotionCmd(const Vector&) {}
  virtual void panEndCmd(const Vector&) {}
  void panPreserveCmd(int r) {preservePan = r;}
  virtual void panToCmd(const Vector&, InternalSystem) {}
  virtual void panToCmd(const Vector&, CoordSystem, SkyFrame) {}
  void rotateCmd(double);
  virtual void rotateBeginCmd() {}
  virtual void rotateMotionCmd(double) {}
  virtual void rotateEndCmd() {}
  void rotateToCmd(double);
  void wcsAlignCmd(int, CoordSystem, SkyFrame);
  void wcsAlignCmd(int, CoordSystem, SkyFrame, FitsImage*);
  void zoomCmd(const Vector&);
  virtual void zoomAboutCmd(const Vector&, const Vector&, InternalSystem) {}
  virtual void zoomAboutCmd(const Vector&, const Vector&, CoordSystem, SkyFrame) {}
  void zoomToCmd(const Vector&);
  virtual void zoomToAboutCmd(const Vector&, const Vector&, InternalSystem) {}
  virtual void zoomToAboutCmd(const Vector&, const Vector&, CoordSystem, SkyFrame) {}
  virtual void zoomToFitCmd(double =1) {}

  // Panner Commands
  void pannerCmd(int);
  void pannerCmd(char*, int, int);
  void pannerCmd(CoordSystem, SkyFrame);
  void updatePannerCmd();

  // Pixel Table Commands
  virtual void getPixelTableCmd(const Vector&, InternalSystem, int, char*) {}

  // Print Commands
#ifdef _MACOSX
  virtual void macosxPrintCmd() =0;
#endif
#ifdef _WIN32
  virtual void win32PrintCmd() =0;
#endif

  // Scale Commands
  void colorScaleCmd(FrScale::ColorScaleType);
  void colorScaleLogCmd(double);
  void getColorMapLevelCmd(int);
  void getColorMapLevelCmd(int, double, double, 
			   FrScale::ColorScaleType, float);
  virtual void getColorMapLevelCmd(int,const Vector&, InternalSystem) {}
  void getColorScaleCmd();
  void getColorScaleLevelCmd(int, double, double, 
			     FrScale::ColorScaleType, float);
  void getColorScaleLogCmd();

  // Smooth Commands
  void getSmoothFunctionCmd();
  void getSmoothRadiusCmd();
  void hasSmoothCmd();
  void smoothCmd(int,int);
  void smoothDeleteCmd();

  // WCS
  void wcsAppendCmd(int, int);
  void wcsAppendCmd(int, const char*);
  void wcsAppendTxtCmd(int, const char*);
  void wcsReplaceCmd(int, int);
  void wcsReplaceCmd(int, const char*);
  void wcsReplaceTxtCmd(int, const char*);
  void wcsResetCmd(int);

  // RGB Commands
  virtual void getRGBChannelCmd() =0;
  virtual void getRGBViewCmd() =0;
  virtual void getRGBSystemCmd() =0;
  virtual void setRGBChannelCmd(const char*) {}
  virtual void setRGBViewCmd(int, int, int) {}
  virtual void setRGBSystemCmd(CoordSystem) {}

  // 3d
  virtual void get3dViewCmd() =0;
  virtual void get3dRenderMethodCmd() =0;
  virtual void get3dThresholdCmd() =0;
  virtual void set3dViewCmd(float, float) {};
  virtual void set3dRenderMethodCmd(RenderMethod) {};
  virtual void set3dThresholdCmd(float) {};
};

#endif
