// Copyright (C) 1999-2010
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include <fstream>
#include "fdstream.hpp"

#include "tcl.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "base.h"
#include "context.h"
#include "fitsimage.h"
#include "mmap.h"

void Base::binCmd(const Vector& b, const Vector& v,
		       const char* x, const char* y, const char* filter)
{
  binFactor_ = b;
  binDepth_ = 1;

  if (currentContext->fits) {
    currentContext->fits->setBinToFactor(b);
    currentContext->fits->setBinDepth(1);
    currentContext->fits->setBinX(x);
    currentContext->fits->setBinY(y);
    currentContext->fits->setBinFilter(filter);
    updateBin(currentContext->fits->updateHist(v));
  }
}

void Base::binCmd(const Vector& b, const char* x, const char* y, 
		       const char* filter)
{
  binFactor_ = b;
  binDepth_ = 1;

  if (currentContext->fits) {
    currentContext->fits->setBinToFactor(b);
    currentContext->fits->setBinDepth(1);
    currentContext->fits->setBinX(x);
    currentContext->fits->setBinY(y);
    currentContext->fits->setBinFilter(filter);
    updateBin(currentContext->fits->updateHistCenter());
  }
}

void Base::binCmd(const Vector& b, int d, const Vector& lim, 
		       const Vector& v,
		       const char* x, const char* y, const char* z,
		       const char* filter)
{
  binFactor_ = b;
  binDepth_ = d;

  if (currentContext->fits) {
    currentContext->fits->setBinToFactor(b);
    currentContext->fits->setBinDepth(d);
    currentContext->fits->setBinX(x);
    currentContext->fits->setBinY(y);
    currentContext->fits->setBinZ(z);
    currentContext->fits->setBinFilter(filter);
    currentContext->fits->setBinColMinMax(z,lim);
    updateBin(currentContext->fits->updateHist(v));
  }
}

void Base::binCmd(const Vector& b, int d, const Vector& lim,
		       const char* x, const char* y, const char* z,
		       const char* filter)
{
  binFactor_ = b;
  binDepth_ = d;

  if (currentContext->fits) {
    currentContext->fits->setBinToFactor(b);
    currentContext->fits->setBinDepth(d);
    currentContext->fits->setBinX(x);
    currentContext->fits->setBinY(y);
    currentContext->fits->setBinZ(z);
    currentContext->fits->setBinFilter(filter);
    currentContext->fits->setBinColMinMax(z,lim);
    updateBin(currentContext->fits->updateHistCenter());
  }
}

void Base::binAboutCmd()
{
  if (currentContext->fits)
    updateBin(currentContext->fits->updateHistCenter());
}

void Base::binAboutCmd(const Vector& v)
{
  if (currentContext->fits)
    updateBin(currentContext->fits->updateHist(v));
}

void Base::binColsCmd(const char* x, const char* y, const char* z)
{
  if (currentContext->fits) {
    currentContext->fits->setBinX(x);
    currentContext->fits->setBinY(y);
    currentContext->fits->setBinZ(z);
    updateBin(currentContext->cfits->updateHistCursor());
  }
}

void Base::binDepthCmd(int d)
{
  binDepth_ = d;
  if (currentContext->fits) {
    currentContext->fits->setBinDepth(d);
    updateBin(currentContext->cfits->updateHistCursor());
  }
}

void Base::binFactorCmd(const Vector& b)
{
  Vector bb = b;
  binFactor_[0] *= bb[0];
  binFactor_[1] *= bb[1];
  if (currentContext->fits) {
    currentContext->fits->setBinFactor(b);
    updateBin(currentContext->cfits->updateHistCursor());
  }
}

void Base::binFactorAboutCmd(const Vector& b, const Vector& v)
{
  Vector bb = b;
  binFactor_[0] *= bb[0];
  binFactor_[1] *= bb[1];
  if (currentContext->fits) {
    currentContext->fits->setBinFactor(b);
    updateBin(currentContext->fits->updateHist(v));
  }
}

void Base::binFactorToCmd(const Vector& b)
{
  Vector bb = b;
  binFactor_[0] = bb[0];
  binFactor_[1] = bb[1];
  if (currentContext->fits) {
    currentContext->fits->setBinToFactor(b);
    updateBin(currentContext->cfits->updateHistCursor());
  }
}

void Base::binFactorToAboutCmd(const Vector& b, const Vector& v)
{
  Vector bb = b;
  binFactor_[0] = bb[0];
  binFactor_[1] = bb[1];
  if (currentContext->fits) {
    currentContext->fits->setBinToFactor(b);
    updateBin(currentContext->fits->updateHist(v));
  }
}

void Base::binFunctionCmd(FitsHist::Function func)
{
  binFunction_ = func;
  if (currentContext->fits) {
    currentContext->fits->setBinFunction(func);
    updateBin(currentContext->cfits->updateHistCursor());
  }
}

void Base::binBufferSizeCmd(int size)
{
  binBufferSize_ = size;
  if (currentContext->fits) {
    currentContext->fits->setBinBufferSize(size);
    updateBin(currentContext->cfits->updateHistCursor());
  }
}

void Base::binFilterCmd(const char* filter)
{
  if (currentContext->fits) {
    currentContext->fits->setBinFilter(filter);
    updateBin(currentContext->cfits->updateHistCursor());
  }
}

void Base::binToFitCmd()
{
  if (currentContext->fits) {
    double bf = 1/calcZoom(Vector(options->width,options->height),
			   currentContext->fits->getHistDim());

    // round up to next power of 2
    if (bf < 1) {
      currentContext->fits->setBinToFactor(Vector(1,1));
      binFactor_ = Vector(1,1);
    }
    else {
      int p=1;
      while (p<bf)
	p*=2;
      currentContext->fits->setBinToFactor(Vector(p,p));
      binFactor_ = Vector(p,p);
    }
  
    updateBin(currentContext->cfits->updateHistCursor());
  }
}

void Base::bgColorCmd(const char* color)
{
  if (bgColorName)
    delete [] bgColorName;
  bgColorName = dupstr(color);
  bgColor = getXColor(bgColorName);
  update(BASE);
}

void Base::centerCmd()
{
  centerImage();
  update(MATRIX);
}

void Base::clearCmd()
{
  unloadAllFits();
  reset();
  update(MATRIX);
}

void Base::clipMinMaxModeCmd(FrScale::MinMaxMode m)
{
  if (currentContext->frScale.mmMode() != m) {
    currentContext->frScale.setMMMode(m);
    currentContext->updateClip();
    updateColorScale();
    update(BASE);
  }
}

void Base::clipMinMaxSampleCmd(int i)
{
  if (currentContext->frScale.mmIncr() != i) {
    currentContext->frScale.setMMIncr(i);
    currentContext->updateClip();
    updateColorScale();
    update(BASE);
  }
}

void Base::clipModeCmd(float per)
{
  if (per == 100) {
    if (currentContext->frScale.clipMode() != FrScale::MINMAX) {
      currentContext->frScale.setClipMode(FrScale::MINMAX);
      currentContext->updateClip();
      updateColorScale();
      update(BASE);
    }
  }
  else {
    if (currentContext->frScale.clipMode() != FrScale::AUTOCUT || 
	currentContext->frScale.autoCutPer() != per) {
      currentContext->frScale.setAutoCutPer(per);
      currentContext->frScale.setClipMode(FrScale::AUTOCUT);
      currentContext->updateClip();
      updateColorScale();
      update(BASE);
    }
  }
}

void Base::clipModeCmd(FrScale::ClipMode m)
{
  if (currentContext->frScale.clipMode() != m) {
    currentContext->frScale.setClipMode(m);
    currentContext->updateClip();
    updateColorScale();
    update(BASE);
  }
}

void Base::clipPreserveCmd(int r)
{
  if (currentContext->frScale.preserve() != r) {
    currentContext->frScale.setPreserve(r);
    currentContext->updateClip();
    updateColorScale();
    update(BASE);
  }
}

void Base::clipScopeCmd(FrScale::ClipScope s)
{
  if (currentContext->frScale.clipScope() != s) {
    currentContext->frScale.setClipScope(s);
    currentContext->updateClip();
    updateColorScale();
    update(BASE);
  }
}

void Base::clipUserCmd(double l, double h)
{
  if (currentContext->frScale.uLow() != l || currentContext->frScale.uHigh()) {
    currentContext->frScale.setULow(l);
    currentContext->frScale.setUHigh(h);
    currentContext->updateClip();
    updateColorScale();
    update(BASE);
  }
}

void Base::clipUserLowCmd(double l)
{
  if (currentContext->frScale.uLow() != l) {
    currentContext->frScale.setULow(l);
    currentContext->updateClip();
    updateColorScale();
    update(BASE);
  }
}

void Base::clipUserHighCmd(double h)
{
  if (currentContext->frScale.uHigh() != h) {
    currentContext->frScale.setUHigh(h);
    currentContext->updateClip();
    updateColorScale();
    update(BASE);
  }
}

void Base::clipZScaleContrastCmd(float c)
{
  if (currentContext->frScale.zContrast() != c) {
    currentContext->frScale.setZContrast(c);
    currentContext->updateClip();
    updateColorScale();
    update(BASE);
  }
}

void Base::clipZScaleSampleCmd(int s)
{
  if (currentContext->frScale.zSample() != s) {
    currentContext->frScale.setZSample(s);
    currentContext->updateClip();
    updateColorScale();
    update(BASE);
  }
}

void Base::clipZScaleLineCmd(int l)
{
  if (currentContext->frScale.zLine() != l) {
    currentContext->frScale.setZLine(l);
    currentContext->updateClip();
    updateColorScale();
    update(BASE);
  }
}

void Base::colorScaleCmd(FrScale::ColorScaleType s)
{
  if (currentContext->frScale.colorScaleType() != s) {
    currentContext->frScale.setColorScaleType(s);
    updateColorScale();
    update(BASE);
  }
}

void Base::colorScaleLogCmd(double exp)
{
  if (currentContext->frScale.expo() != exp) {
    currentContext->frScale.setExpo(exp);
    if (currentContext->frScale.colorScaleType() == FrScale::LOGSCALE) {
      updateColorScale();
      update(BASE);
    }
  }
}

void Base::DATASECCmd(int which)
{
  switch (currentContext->frScale.scanMode()) {
  case FrScale::NODATASEC:
  case FrScale::DATASEC:
  case FrScale::RESET:
    currentContext->
      frScale.setScanMode(which ? FrScale::DATASEC : FrScale::NODATASEC);
    break;
  case FrScale::UNODATASEC:
  case FrScale::UDATASEC:
    currentContext->
      frScale.setScanMode(which ? FrScale::UDATASEC : FrScale::UNODATASEC);
    break;
  }

  currentContext->updateClip();
  updateColorScale();
  update(MATRIX);
}

void Base::fitsyHasExtCmd(const char* fn)
{
  // verify that we have an ext specified
  if (fn && (fn[strlen(fn)-1] != ']')) {
    Tcl_AppendResult(interp, "0", NULL);
    return;
  }

  FitsFile* ext = new FitsFitsMMap(fn, FitsFile::EXACT);
  if (ext->isValid())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);

  delete ext;
  return;
}

void Base::getBinCursorCmd()
{
  if (currentContext->fits)
    printVector(currentContext->fits->getHistCursor(), DEFAULT);
}

void Base::getBinColsCmd()
{
  if (currentContext->fits && currentContext->fits->isHist())
    if (currentContext->fits->binDepth()>1)
      Tcl_AppendResult(interp, currentContext->fits->getHistX(), " ", 
		       currentContext->fits->getHistY(), " ",
		       currentContext->fits->getHistZ(), NULL);
    else
      Tcl_AppendResult(interp, currentContext->fits->getHistX(), " ", 
		       currentContext->fits->getHistY(), NULL);

  else
    Tcl_AppendResult(interp, "", NULL);
}

void Base::getBinColsMinMaxCmd(const char* col)
{
  if (currentContext->fits)
    printVector(currentContext->fits->getHistColMinMax(col), DEFAULT);
}

void Base::getBinColsDimCmd(const char* col)
{
  if (currentContext->fits)
    printVector(currentContext->fits->getHistColDim(col), DEFAULT);
}

void Base::getBinDepthCmd()
{
  if (currentContext->fits)
    binDepth_ = currentContext->fits->binDepth();
 
  printDouble(binDepth_, DEFAULT);
}

void Base::getBinFactorCmd()
{
  if (currentContext->fits)
    binFactor_ = currentContext->fits->binFactor();

  printVector(binFactor_, DEFAULT);
}

void Base::getBinFunctionCmd()
{
  if (currentContext->fits)
    binFunction_ = currentContext->fits->binFunction();

  switch (binFunction_) {
  case FitsHist::AVERAGE:
    Tcl_AppendResult(interp, "average", NULL);
    return;
  case FitsHist::SUM:
    Tcl_AppendResult(interp, "sum", NULL);
    return;
  }
}

void Base::getBinBufferSizeCmd()
{
  if (currentContext->fits)
    binBufferSize_ = currentContext->fits->binBufferSize();

  printInteger(binBufferSize_);
}

void Base::getBinFilterCmd()
{
  if (currentContext->fits && currentContext->fits->isHist())
    Tcl_AppendResult(interp, currentContext->fits->getHistFilter(), NULL);
  else
    Tcl_AppendResult(interp, "", NULL);
}

void Base::getBinListCmd()
{
  if (currentContext->fits && currentContext->fits->isHist()) {
    char* cols = currentContext->fits->getHistList();
    Tcl_AppendResult(interp, cols, NULL);
    delete [] cols;
  }
  else
    Tcl_AppendResult(interp, "", NULL);
}

void Base::getBitpixCmd()
{
  if (currentContext->cfits)
    printInteger(currentContext->cfits->bitpix());
}

void Base::getBgColorCmd()
{
  Tcl_AppendResult(interp, bgColorName, NULL);
}

void Base::getClipCmd()
{
  printVector(currentContext->getClip(), DEFAULT);
}

void Base::getClipCmd(float per)
{
  FrScale::ClipMode cm = (per == 100) ? FrScale::MINMAX : FrScale::AUTOCUT;
  float ac = per;

  ostringstream str;
  str << currentContext->getClip(cm, ac) << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Base::getClipCmd(FrScale::ClipMode cm)
{
  ostringstream str;
  str << currentContext->getClip(cm, currentContext->frScale.autoCutPer()) << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Base::getClipMinMaxModeCmd()
{
  switch (currentContext->frScale.mmMode()) {
  case FrScale::AUTOSCAN:
    Tcl_AppendResult(interp, "auto", NULL);
    return;
  case FrScale::SCAN:
    Tcl_AppendResult(interp, "scan", NULL);
    return;
  case FrScale::SAMPLE:
    Tcl_AppendResult(interp, "sample", NULL);
    return;
  case FrScale::DATAMIN:
    Tcl_AppendResult(interp, "datamin", NULL);
    return;
  case FrScale::IRAFMIN:
    Tcl_AppendResult(interp, "irafmin", NULL);
    return;
  }
}

void Base::getClipMinMaxSampleCmd()
{
  printInteger(currentContext->frScale.mmIncr());
}

void Base::getClipModeCmd()
{
  switch (currentContext->frScale.clipMode()) {
  case FrScale::MINMAX:
    Tcl_AppendResult(interp, "minmax", NULL);
    break;
  case FrScale::ZSCALE:
    Tcl_AppendResult(interp, "zscale", NULL);
    break;
  case FrScale::ZMAX:
    Tcl_AppendResult(interp, "zmax", NULL);
    break;
  case FrScale::AUTOCUT:
    printDouble(currentContext->frScale.autoCutPer(), DEFAULT);
    break;
  case FrScale::USERCLIP:
    Tcl_AppendResult(interp, "user", NULL);
    break;
  }
}

void Base::getClipPreserveCmd()
{
  if (currentContext->frScale.preserve())
      Tcl_AppendResult(interp, "1", NULL);
  else
      Tcl_AppendResult(interp, "0", NULL);
}

void Base::getClipScopeCmd()
{
  switch (currentContext->frScale.clipScope()) {
  case FrScale::GLOBAL:
    Tcl_AppendResult(interp, "global", NULL);
    break;
  case FrScale::LOCAL:
    Tcl_AppendResult(interp, "local", NULL);
    break;
  }
}

void Base::getClipUserCmd()
{
  printVector(Vector(currentContext->frScale.uLow(), 
		     currentContext->frScale.uHigh()), DEFAULT);
}

void Base::getClipZScaleContrastCmd()
{
  ostringstream str;
  str << currentContext->frScale.zContrast() << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Base::getClipZScaleSampleCmd()
{
  ostringstream str;
  str << currentContext->frScale.zSample() << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Base::getClipZScaleLineCmd()
{
  ostringstream str;
  str << currentContext->frScale.zLine() << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Base::getColorMapLevelCmd(int count)
{
  if (currentContext->cfits)
    getColorMapLevelCmd(count, currentContext->cfits->getLowDouble(),
			currentContext->cfits->getHighDouble(),
			currentContext->frScale.colorScaleType(),
			currentContext->frScale.expo());		
  else
    getColorMapLevelCmd(count, currentContext->frScale.low(), 
			currentContext->frScale.high(),
			currentContext->frScale.colorScaleType(),
			currentContext->frScale.expo());		
}

void Base::getColorMapLevelCmd(int count, double ll, double hh,
			       FrScale::ColorScaleType scaleType,
			       float expo)
{
  if (inverseScale)
    delete inverseScale;
  inverseScale = NULL;

  switch (scaleType) {
  case FrScale::LINEARSCALE:
  case FrScale::IISSCALE:
    inverseScale = new LinearInverseScale(count, ll, hh);
    break;
  case FrScale::LOGSCALE:
    inverseScale = new LogInverseScale(count, ll, hh, expo);
    break;
  case FrScale::POWSCALE:
    inverseScale = new PowInverseScale(count, ll, hh, expo);
    break;
  case FrScale::SQRTSCALE:
    inverseScale = new SqrtInverseScale(count, ll, hh);
    break;
  case FrScale::SQUAREDSCALE:
    inverseScale = new SquaredInverseScale(count, ll, hh);
    break;
  case FrScale::HISTEQUSCALE:
    inverseScale = new HistEquInverseScale(count, ll, hh, 
				    currentContext->histequ(),
				    HISTEQUSIZE);
    break;
  }

  if (inverseScale) {
    ostringstream str;
    str << inverseScale->size() << ' ' << inverseScale->level() << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }
}

void Base::getColorScaleCmd()
{
  switch (currentContext->frScale.colorScaleType()) {
  case FrScale::LINEARSCALE:
    Tcl_AppendResult(interp, "linear", NULL);
    break;
  case FrScale::LOGSCALE:
    Tcl_AppendResult(interp, "log", NULL);
    break;
  case FrScale::POWSCALE:
    Tcl_AppendResult(interp, "pow", NULL);
    break;
  case FrScale::SQRTSCALE:
    Tcl_AppendResult(interp, "sqrt", NULL);
    break;
  case FrScale::SQUAREDSCALE:
    Tcl_AppendResult(interp, "squared", NULL);
    break;
  case FrScale::IISSCALE:
    Tcl_AppendResult(interp, "linear", NULL);
    break;
  case FrScale::HISTEQUSCALE:
    Tcl_AppendResult(interp, "histequ", NULL);
    break;
  }
}

void Base::getColorScaleLevelCmd(int count, double ll, double hh,
				 FrScale::ColorScaleType scaleType, 
				 float expo)
{
  InverseScale* scale;
  switch (scaleType) {
  case FrScale::LINEARSCALE:
  case FrScale::IISSCALE:
    scale = new LinearInverseScale(count, ll, hh);
    break;
  case FrScale::LOGSCALE:
    scale = new LogInverseScale(count, ll, hh, expo);
    break;
  case FrScale::POWSCALE:
    scale = new PowInverseScale(count, ll, hh, expo);
    break;
  case FrScale::SQRTSCALE:
    scale = new SqrtInverseScale(count, ll, hh);
    break;
  case FrScale::SQUAREDSCALE:
    scale = new SquaredInverseScale(count, ll, hh);
    break;
  case FrScale::HISTEQUSCALE:
    scale = new HistEquInverseScale(count, ll, hh, 
				    currentContext->histequ(),
				    HISTEQUSIZE);
    break;
  }

  ostringstream str;
  str << *scale << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
 
  delete scale;
}

void Base::getColorScaleLogCmd()
{
  ostringstream str;
  str << currentContext->frScale.expo() << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Base::getContourCmd(CoordSystem sys, SkyFrame sky)
{
  if (currentContext->cfits && hasContour()) {
    List<Vertex>& v = (List<Vertex>&)(currentContext->contour->contours());
    if (v.head())
      do {
	Vector w = v.current()->vector;
	if (w[0] != DBL_MAX)
	  printVector(currentContext->cfits->mapFromRef(w, sys, sky),DEFAULT);
	Tcl_AppendResult(interp, "\n", NULL);
      } while (v.next());
  }
}

void Base::getContourMethodCmd()
{
  if (hasContour())
    Tcl_AppendResult(interp, currentContext->contour->methodName(), NULL);
}

void Base::getContourClipCmd()
{
  if (hasContour())
    printVector(currentContext->contour->limits(),DEFAULT);
}

void Base::getContourClipModeCmd()
{
  if (hasContour())
    if (currentContext->contour->clipMode() == FrScale::MINMAX)
      Tcl_AppendResult(interp, "minmax", NULL);
    else if (currentContext->contour->clipMode() == FrScale::ZSCALE)
      Tcl_AppendResult(interp, "zscale", NULL);
    else if (currentContext->contour->clipMode() == FrScale::ZMAX)
      Tcl_AppendResult(interp, "zmax", NULL);
    else if (currentContext->contour->clipMode() == FrScale::USERCLIP)
      Tcl_AppendResult(interp, "user", NULL);
    else {
      ostringstream str;
      str << currentContext->contour->clipMode() << ends;
      Tcl_AppendResult(interp, str.str().c_str(), NULL);
    }
}

void Base::getContourColorNameCmd()
{
  if (hasContour())
    Tcl_AppendResult(interp, currentContext->contour->getColorName(), NULL);
}

void Base::getContourAuxColorNameCmd()
{
  if (hasContourAux())
    Tcl_AppendResult(interp, currentContext->auxcontours.current()->getColorName(), NULL);
}

void Base::getContourDashCmd()
{
  if (hasContour() && currentContext->contour->getDash())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::getContourAuxDashCmd()
{
  if (hasContourAux() && currentContext->auxcontours.current()->getDash())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::getContourLevelCmd()
{
  if (hasContour())
    Tcl_AppendResult(interp, currentContext->contour->level(), NULL);
}

void Base::getContourLineWidthCmd()
{
  if (hasContour()) {
    ostringstream str;
    str << currentContext->contour->getLineWidth() << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }
}

void Base::getContourAuxLineWidthCmd()
{
  if (hasContourAux()) {
    ostringstream str;
    str << currentContext->auxcontours.current()->getLineWidth() << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }
}

void Base::getContourNumLevelCmd()
{
  if (hasContour()) {
    ostringstream str;
    str << currentContext->contour->numLevel() << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }
}

void Base::getContourSmoothCmd()
{
  if (hasContour()) {
    ostringstream str;
    str << currentContext->contour->smooth() << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }
}

void Base::getContourScaleCmd()
{
  if (hasContour())
    switch (currentContext->contour->colorScaleType()) {
    case FrScale::LINEARSCALE:
      Tcl_AppendResult(interp, "linear", NULL);
      break;
    case FrScale::LOGSCALE:
      Tcl_AppendResult(interp, "log", NULL);
      break;
    case FrScale::POWSCALE:
      Tcl_AppendResult(interp, "pow", NULL);
      break;
    case FrScale::SQRTSCALE:
      Tcl_AppendResult(interp, "sqrt", NULL);
      break;
    case FrScale::SQUAREDSCALE:
      Tcl_AppendResult(interp, "squared", NULL);
      break;
    case FrScale::IISSCALE:
      Tcl_AppendResult(interp, "linear", NULL);
      break;
    case FrScale::HISTEQUSCALE:
      Tcl_AppendResult(interp, "histequ", NULL);
      break;
    }
}

void Base::getContourScaleLogCmd()
{
  if (hasContour()) {
    ostringstream str;
    str << currentContext->contour->expo() << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }
}

void Base::getDATASECCmd()
{
  if (currentContext->frScale.scanMode())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::getFitsCountCmd()
{
  printInteger(currentContext->fitsCount());
}

void Base::getFitsDepthCmd(int ii)
{
  int dd = currentContext->naxis(ii);
  if (dd>1)
    printInteger(dd);
  else
    printInteger(1);
}

void Base::getFitsHeaderCmd(int which)
{
  int prim = which < 0 ? 1:0;
  which = abs(which);

  // modified for medatacube
  FitsImage* rr = findAllFits(which);
  if (rr) {
    char* hh = !prim ? rr->displayHeader() : rr->displayPrimary();
    Tcl_AppendResult(interp, hh, NULL);
    delete [] hh;
  }
  else
    result = TCL_ERROR;
}

void Base::getFitsHeaderWCSCmd(int which)
{
  // no primary
  FitsImage* rr = findAllFits(which);
  if (rr) {
    char* hh = rr->displayWCS();
    Tcl_AppendResult(interp, hh, NULL);
    delete [] hh;
  }
  else
    result = TCL_ERROR;
}

void Base::getFitsHeaderKeywordCmd(int which, const char* key)
{
  which = abs(which);

  FitsImage* rr = findAllFits(which);
  if (rr) {
    char* value = rr->getKeyword(key);
    if (value) {
      Tcl_AppendResult(interp, value, NULL);
      delete [] value;
      return;
    }
  }
  else
    result = TCL_ERROR;
}

void Base::getFitsFileNameCmd(FileNameType type)
{
  if (currentContext->cfits) {
    switch (type) {
    case ROOTBASE:
      Tcl_AppendResult(interp, currentContext->cfits->getRootBaseFileName(), NULL);
      break;
    case FULLBASE:
      Tcl_AppendResult(interp, currentContext->cfits->getFullBaseFileName(), NULL);
      break;
    case ROOT:
      Tcl_AppendResult(interp, currentContext->cfits->getRootFileName(), NULL);
      break;
    case FULL:
      Tcl_AppendResult(interp, currentContext->cfits->getFullFileName(), NULL);
      break;
    }
  }
}

void Base::getFitsFileNameCmd(int which, FileNameType type)
{
  FitsImage* rr = findAllFits(which);
  if (rr) {
    switch (type) {
    case ROOTBASE:
      Tcl_AppendResult(interp, rr->getRootBaseFileName(), NULL);
      break;
    case FULLBASE:
      Tcl_AppendResult(interp, rr->getFullBaseFileName(), NULL);
      break;
    case ROOT:
      Tcl_AppendResult(interp, rr->getRootFileName(), NULL);
      break;
    case FULL:
      Tcl_AppendResult(interp, rr->getFullFileName(), NULL);
      break;
    }
  }
  else
    result = TCL_ERROR;
}

void Base::getFitsObjectNameCmd()
{
  if (currentContext->cfits)
    Tcl_AppendResult(interp, currentContext->cfits->getObjectName(), NULL);
}

void Base::getFitsSizeCmd()
{
  if (keyContext->fits)
    printVector(Vector(keyContext->fits->width(), keyContext->fits->height()),
		DEFAULT);
  else
    printVector(Vector(), DEFAULT);
}

void Base::getFitsSliceCmd(int ii)
{
  int ss = currentContext->slice(ii);
  if (ss>1)
    printInteger(ss);
  else
    printInteger(1);
}

void Base::getHeightCmd()
{
  if (currentContext->cfits)
    printInteger(currentContext->cfits->height());
}

void Base::getHistogramCmd(char* xName, char* yName)
{
  currentContext->bltHist(interp,xName,yName);
  printDouble((currentContext->frScale.max()-currentContext->frScale.min())/256., DEFAULT);
}

void Base::getInfoCmd(char* var)
{
  if (currentContext->cfits) {
    Tcl_SetVar2(interp,var,"filename",
		(char*)currentContext->cfits->getRootBaseFileName(),0);
    Tcl_SetVar2(interp,var,"objectname",(char*)currentContext->cfits->getObjectName(),0);
    Tcl_SetVar2(interp,var,"min",(char*)currentContext->cfits->getMin(),0);
    Tcl_SetVar2(interp,var,"max",(char*)currentContext->cfits->getMax(),0);
    Tcl_SetVar2(interp,var,"low",(char*)currentContext->cfits->getLow(),0);
    Tcl_SetVar2(interp,var,"high",(char*)currentContext->cfits->getHigh(),0);
  }
  else {
    Tcl_SetVar2(interp,var,"objectname","",0);
    Tcl_SetVar2(interp,var,"filename","",0);
    Tcl_SetVar2(interp,var,"min","",0);
    Tcl_SetVar2(interp,var,"max","",0);
    Tcl_SetVar2(interp,var,"low","",0);
    Tcl_SetVar2(interp,var,"high","",0);
  }

  Tcl_SetVar2(interp,var,"value","",0);
  Tcl_SetVar2(interp,var,"xImage","",0);
  Tcl_SetVar2(interp,var,"yImage","",0);
  Tcl_SetVar2(interp,var,"xPhysical","",0);
  Tcl_SetVar2(interp,var,"yPhysical","",0);
  Tcl_SetVar2(interp,var,"xAmplifier","",0);
  Tcl_SetVar2(interp,var,"yAmplifier","",0);
  Tcl_SetVar2(interp,var,"xDetector","",0);
  Tcl_SetVar2(interp,var,"yDetector","",0);

  Tcl_SetVar2(interp,var,"xWCS,","",0);
  Tcl_SetVar2(interp,var,"yWCS,","",0);
  Tcl_SetVar2(interp,var,"sysWCS,","",0);
  Tcl_SetVar2(interp,var,"xUnitWCS,","",0);
  Tcl_SetVar2(interp,var,"yUnitWCS,","",0);
}

void Base::getInfoCmd(const Vector& v, InternalSystem ref,
		      SkyFrame sky, SkyFormat format, char* var)
{
  switch (ref) {
  case CANVAS:
    getInfo(v, ref, &FitsImage::getCanvasToData, &FitsImage::getCanvasToImage,
	    sky, format, var);
    break;
  case PANNER:
    getInfo(v, ref, &FitsImage::getPannerToData, &FitsImage::getPannerToImage,
	    sky, format, var);
    break;
  }
}

void Base::getInfoClipCmd()
{
  if (currentContext->cfits) {
    Tcl_AppendElement(interp, (char*)currentContext->cfits->getLow());
    Tcl_AppendElement(interp, (char*)currentContext->cfits->getHigh());
  }
  else {
    Tcl_AppendElement(interp, "0");
    Tcl_AppendElement(interp, "0");
  }
}

void Base::getMaskColorCmd()
{
  Tcl_AppendResult(interp, maskColorName, NULL);
}

void Base::getMaskMarkCmd()
{
  if (maskMark)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::getMaskTransparencyCmd()
{
  printDouble((1-maskAlpha)*100., DEFAULT);
}

void Base::getMinMaxCmd()
{
  printVector(Vector(currentContext->frScale.min(), currentContext->frScale.max()), DEFAULT);
}

void Base::getNANColorCmd()
{
  Tcl_AppendResult(interp, nanColorName, NULL);
}

void Base::getOrientCmd()
{
  switch (orientation) {
  case NORMAL:
    Tcl_AppendResult(interp, "none", NULL);
    return;
  case XX:
    Tcl_AppendResult(interp, "x", NULL);
    return;
  case YY:
    Tcl_AppendResult(interp, "y", NULL);
    return;
  case XY:
    Tcl_AppendResult(interp, "xy", NULL);
    return;
  }
}

void Base::getPanPreserveCmd()
{
  if (preservePan)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::getRotateCmd(Precision p)
{
  printDouble(radToDeg(rotation), p);
}

void Base::getSmoothFunctionCmd()
{
  if (currentContext->fits) 
    smoothFunction_ = currentContext->fits->smoothFunction();

  switch (smoothFunction_) {
  case FitsImage::BOXCAR:
    Tcl_AppendResult(interp, "boxcar", NULL);
    return;
  case FitsImage::TOPHAT:
    Tcl_AppendResult(interp, "tophat", NULL);
    return;
  case FitsImage::GAUSSIAN:
    Tcl_AppendResult(interp, "gaussian", NULL);
    return;
  }
}

void Base::getSmoothRadiusCmd()
{
  if (currentContext->fits)
    smoothRadius_ = currentContext->fits->smoothRadius();

  printInteger(smoothRadius_);
}

void Base::getWCSNameCmd(CoordSystem sys)
{
  if (currentContext->cfits && currentContext->cfits->hasWCS(sys)) {
    char* wcsname = (char*)currentContext->cfits->getWCSName(sys);
    if (wcsname) {
      Tcl_AppendResult(interp, wcsname, NULL);
      return;
    }
  }
  Tcl_AppendResult(interp, NULL);
}

void Base::getWidthCmd()
{
  if (currentContext->cfits)
    printInteger(currentContext->cfits->width());
}

void Base::getZoomCmd(Precision p)
{
  printVector(zoom_, p);
}

void Base::hasAmplifierCmd()
{
  if (hasATMV())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasContourCmd()
{
  if (hasContour())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasContourAuxCmd()
{
  if (hasContourAux())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasDATAMINCmd()
{
  if (currentContext->cfits && currentContext->cfits->hasDATAMIN())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasDATASECCmd()
{
  if (currentContext->cfits && currentContext->cfits->hasDATASEC())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasDetectorCmd()
{
  if (hasDTMV())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasFitsCmd()
{
  if (currentContext->fits)
      Tcl_AppendResult(interp, "1", NULL);
  else
      Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasFitsBinCmd()
{
  if (currentContext->fits && currentContext->fits->isHist())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasFitsCubeCmd()
{
  if (isCube())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasFitsMosaicCmd()
{
  if (isMosaic())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasGridCmd()
{
  if (grid)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasIISCmd()
{
  if (isIIS())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasIRAFMINCmd()
{
  if (currentContext->cfits && currentContext->cfits->hasIRAFMIN())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasPhysicalCmd()
{
  if (hasLTMV())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasSmoothCmd()
{
  if (doSmooth_)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasSystemCmd(CoordSystem sys)
{
  switch (sys) {
  case IMAGE:
    Tcl_AppendResult(interp, "1", NULL);
    return;
  case PHYSICAL:
    hasPhysicalCmd();
    return;
  case AMPLIFIER:
    hasAmplifierCmd();
    return;
  case DETECTOR:
    hasDetectorCmd();
    return;
  default:
    hasWCSCmd(sys);
    return;
  }
}

void Base::hasWCSCmd(CoordSystem sys)
{
  if (hasWCS(sys))
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasWCSEquCmd(CoordSystem sys)
{
  if (hasWCSEqu(sys))
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasWCSLinearCmd(CoordSystem sys)
{
  if (hasWCSLinear(sys))
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::highliteCmd(int which)
{
  useHighlite = which ? 1 : 0;
  update(PIXMAP);
}

void Base::loadIncrDataCmd(int which, int x0, int y0, int x1, int y1)
{
  FitsImage* ptr = currentContext->fits;
  if (which > 0) {
    for (int i=0; i<(which-1); i++) {
      if (ptr)
	ptr = ptr->nextMosaic();
    }
  }

  if (ptr)
    ptr->setDataParams(x0,y0,x1,y1);
}

void Base::loadIncrMinMaxCmd(int which, int x0, int y0, int x1, int y1)
{
  FitsImage* ptr = currentContext->fits;
  if (which > 0) {
    for (int i=0; i<(which-1); i++) {
      if (ptr)
	ptr = ptr->nextMosaic();
    }
  }

  if (ptr)
    ptr->setMinMaxParams(x0,y0,x1,y1);

  currentContext->updateClip();
  updateColorScale();
}

void Base::loadIncrEndCmd()
{
  switch (currentContext->frScale.scanMode()) {
  case FrScale::NODATASEC:
  case FrScale::UNODATASEC:
  case FrScale::RESET:
    currentContext->frScale.setScanMode(FrScale::NODATASEC);
    break;
  case FrScale::DATASEC:
  case FrScale::UDATASEC:
    currentContext->frScale.setScanMode(FrScale::DATASEC);
    break;
  }

  currentContext->updateClip();
  updateColorScale();
  update(MATRIX);
}

void Base::magnifierCmd(int s)
{
  useMagnifier = s;
}

void Base::magnifierCmd(char* n, int w, int h)
{
  strcpy(magnifierName,n);
  magnifierWidth = w;
  magnifierHeight = h;

  if (magnifierPixmap)
    Tk_FreePixmap(display, magnifierPixmap);
  magnifierPixmap = 0;

  if (magnifierXImage)
    XDestroyImage(magnifierXImage);
  magnifierXImage = NULL;

  if (magnifierWidth > 0 && magnifierHeight > 0) {
    magnifierPixmap = Tk_GetPixmap(display, Tk_WindowId(tkwin),
			     magnifierWidth, magnifierHeight, depth);
    if (!magnifierPixmap) {
      internalError("Unable to Create Magnifier Pixmap");
      return;
    }

    if (!magnifierXImage) {
      if (!(magnifierXImage = XGetImage(display, magnifierPixmap, 0, 0, 
					magnifierWidth, magnifierHeight,
					AllPlanes, ZPixmap))){
	internalError("Unable to Create Magnifier XImage");
	return;
      }
    }
  }

  if (!magnifierGC)
    magnifierGC = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);

  updateMagnifierMatrix();
}

void Base::magnifierCursorCmd(int which)
{
  useMagnifierCursor = which;
  updateMagnifier();
}

void Base::magnifierColorCmd(const char* color)
{
  if (magnifierColorName)
    delete [] magnifierColorName;
  magnifierColorName = dupstr(color);
  updateMagnifier();
}

void Base::magnifierGraphicsCmd(int which)
{
  useMagnifierGraphics = which;
  updateMagnifier();
}

void Base::magnifierZoomCmd(double z)
{
  magnifierZoom_ = fabs(z);
  updateMagnifierMatrix();
  updateMagnifier();
}

void Base::maskClearCmd()
{
  currentContext->mask.deleteAll();
  update(BASE);
}

void Base::maskColorCmd(const char* color)
{
  if (maskColorName)
    delete [] maskColorName;

  maskColorName = dupstr(color);
}

void Base::maskTransparencyCmd(float t)
{
  maskAlpha = 1-(t/100.);
  update(BASE);
}

void Base::nanColorCmd(const char* color)
{
  if (nanColorName)
    delete [] nanColorName;
  nanColorName = dupstr(color);
  nanColor = getXColor(nanColorName);
  update(BASE);
}

void Base::orientCmd(Orientation which)
{
  orientation = which;

  switch (orientation) {
  case NORMAL:
    orientationMatrix.identity();
    break;
  case XX:
    orientationMatrix = FlipX();
    break;
  case YY:
    orientationMatrix = FlipY();
    break;
  case XY:
    orientationMatrix = FlipXY();
    break;
  }

  update(MATRIX);
}

void Base::pannerCmd(int s)
{
  usePanner = s;
  updatePanner();
}

void Base::pannerCmd(char* n, int w, int h)
{
  strcpy(pannerName,n);
  pannerWidth = w;
  pannerHeight = h;

  if (pannerPixmap)
    Tk_FreePixmap(display, pannerPixmap);
  pannerPixmap = 0;

  if (pannerXImage)
    XDestroyImage(pannerXImage);
  pannerXImage = NULL;

  if (pannerWidth > 0 && pannerHeight > 0) {
    if (!(pannerPixmap = Tk_GetPixmap(display, Tk_WindowId(tkwin), 
				      pannerWidth, pannerHeight, depth))) {
      internalError("Unable to Create Panner Pixmap");
      return;
    }

    if (!(pannerXImage = XGetImage(display, pannerPixmap, 0, 0,
				   pannerWidth, pannerHeight,
				   AllPlanes, ZPixmap))){
      internalError("Unable to Create Panner XImage");
      return;
    }
  }

  // update panner matrices
  update(MATRIX);
}

void Base::pannerCmd(CoordSystem sys, SkyFrame sky)
{
  pannerSystem = sys;
  pannerSky = sky;
  updatePanner();
}

void Base::rotateCmd(double r)
{
  rotation += r;
  update(MATRIX);
}

void Base::rotateToCmd(double r)
{
  rotation = r;
  update(MATRIX);
}

void Base::saveFitsImageFileCmd(const char* fn, int compress)
{
  if (currentContext->cfits) {
    if (!currentContext->cfits->saveFitsImageFile(fn, compress)) {
      Tcl_AppendResult(interp, "Unable to save Fits", NULL);
      result = TCL_ERROR;
    }
  }
}

void Base::saveFitsImageChannelCmd(const char* ch, int compress)
{
  if (currentContext->cfits) {
    if (!currentContext->cfits->saveFitsImageChannel(interp, ch, compress)) {
      Tcl_AppendResult(interp, "Unable to save fits to channel ", ch, NULL);
      result = TCL_ERROR;
    }
  }
}

void Base::saveFitsImageSocketCmd(int s, int compress)
{
  if (currentContext->cfits) {
    if (!currentContext->cfits->saveFitsImageSocket(s,compress)) {
      Tcl_AppendResult(interp, "Unable to save Fits", NULL);
      result = TCL_ERROR;
    }
  }
}

void Base::saveFitsTableFileCmd(const char* fn, int compress)
{
  if (currentContext->cfits) {
    if (!currentContext->cfits->saveFitsTableFile(fn, compress)) {
      Tcl_AppendResult(interp, "Unable to save Fits", NULL);
      result = TCL_ERROR;
    }
  }
}

void Base::saveFitsTableChannelCmd(const char* ch, int compress)
{
  if (currentContext->cfits) {
    if (!currentContext->cfits->saveFitsTableChannel(interp, ch, compress)) {
      Tcl_AppendResult(interp, "Unable to save fits to channel ", ch, NULL);
      result = TCL_ERROR;
    }
  }
}

void Base::saveFitsTableSocketCmd(int s, int compress)
{
  if (currentContext->cfits) {
    if (!currentContext->cfits->saveFitsTableSocket(s, compress)) {
      Tcl_AppendResult(interp, "Unable to save Fits", NULL);
      result = TCL_ERROR;
    }
  }
}

void Base::saveArrayFileCmd(const char* fn, FitsFile::ArchType endian)
{
  if (currentContext->cfits) {
    if (!currentContext->cfits->saveArrayFile(fn, endian)) {
      Tcl_AppendResult(interp, "Unable to save Array", NULL);
      result = TCL_ERROR;
    }
  }
}

void Base::saveArrayChannelCmd(const char* ch, FitsFile::ArchType endian)
{
  if (currentContext->cfits) {
    if (!currentContext->cfits->saveArrayChannel(interp, ch, endian)) {
      Tcl_AppendResult(interp, "Unable to save Array to channel ", ch, NULL);
      result = TCL_ERROR;
    }
  }
}

void Base::saveArraySocketCmd(int s, FitsFile::ArchType endian)
{
  if (currentContext->cfits) {
    if (!currentContext->cfits->saveArraySocket(s, endian)) {
      Tcl_AppendResult(interp, "Unable to save Array", NULL);
      result = TCL_ERROR;
    }
  }
}

void Base::smoothCmd(int f, int r)
{
  doSmooth_ = 1;
  smoothFunction_ = f;
  smoothRadius_ = r;

  if (currentContext->fits) {
    currentContext->fits->setDoSmooth(doSmooth_);
    currentContext->fits->setSmoothFunction((FitsImage::SmoothFunction)smoothFunction_);
    currentContext->fits->setSmoothRadius(smoothRadius_);
  }

  analysisFits();
  update(BASE);
}

void Base::smoothDeleteCmd()
{
  doSmooth_ = 0;

  if (currentContext->fits)
    currentContext->fits->setDoSmooth(doSmooth_);

  analysisFits();
  update(BASE);
}

// not used
// don't know if this is used by anyone else
void Base::unloadFitsCmd()
{
  unloadAllFits();
  update(MATRIX);
}

void Base::updateFitsCmd(int now)
{
  if (now)
    updateNow(BASE);
  else
    update(BASE);
}

void Base::updateMagnifierCmd()
{
  updateMagnifier();
}

void Base::updateMagnifierCmd(const Vector& v)
{
  updateMagnifier(v);
}

void Base::updatePannerCmd()
{
  updatePanner();
}

void Base::warpCmd(const Vector& v)
{
  Vector a=v; // can't use a const here
  XWarpPointer(display, None, None, 0, 0, 0, 0, (int)a[0], (int)a[1]);
}

void Base::warpToCmd(const Vector& v)
{
  Vector a =v*canvasToWindow;
  XWarpPointer(display, None, Tk_WindowId(tkwin), 0,0,0,0,(int)a[0],(int)a[1]);
}

void Base::wcsAlignCmd(int which, CoordSystem sys, SkyFrame sky)
{
  wcsAlignment = which;
  wcsSystem = sys;
  wcsSky = sky;

  alignWCS();
  update(MATRIX);
}

void Base::wcsAlignCmd(int which, CoordSystem sys, SkyFrame sky,
		       FitsImage* ptr)
{
  wcsAlignment = which;
  wcsSystem = sys;
  wcsSky = sky;

  alignWCS(ptr);
  update(MATRIX);
}

void Base::wcsAppendCmd(int which, int fd)
{
  if (!currentContext->cfits)
    return;

  boost::fdistream str(fd);
  if (!str) {
    Tcl_AppendResult(interp, " unable to read wcs infomation", NULL);
    result = TCL_ERROR;
    return;
  }

  FitsImage* rr = findAllFits(which);
  if (rr) {
    FitsHead* hh = parseWCS(str);
    rr->appendWCS(hh);
    delete hh;
  }
  else
    result = TCL_ERROR;
}

void Base::wcsAppendCmd(int which, const char* fn)
{
  if (!currentContext->cfits)
    return;

  ifstream str(fn);
  if (!str) {
    Tcl_AppendResult(interp, " unable to load wcs file ", fn, NULL);
    result = TCL_ERROR;
    return;
  }

  FitsImage* rr = findAllFits(which);
  if (rr) {
    FitsHead* hh = parseWCS(str);
    rr->appendWCS(hh);
    delete hh;
  }
  else
    result = TCL_ERROR;
}

void Base::wcsAppendTxtCmd(int which, const char* txt)
{
  if (!currentContext->cfits)
    return;

  istringstream str(txt);
  if (!str) {
    Tcl_AppendResult(interp, " unable to process text", NULL);
    result = TCL_ERROR;
    return;
  }

  FitsImage* rr = findAllFits(which);
  if (rr) {
    FitsHead* hh = parseWCS(str);
    rr->appendWCS(hh);
    delete hh;
  }
  else
    result = TCL_ERROR;
}

void Base::wcsResetCmd(int which)
{
  if (!currentContext->cfits) 
    return;

  FitsImage* rr = findAllFits(which);
  if (rr)
    rr->resetWCS();
  else
    result = TCL_ERROR;
}

void Base::wcsReplaceCmd(int which, int fd)
{
  if (!currentContext->cfits)
    return;

  boost::fdistream str(fd);
  if (!str) {
    Tcl_AppendResult(interp, " unable to read wcs infomation", NULL);
    result = TCL_ERROR;
    return;
  }

  FitsImage* rr = findAllFits(which);
  if (rr) {
    FitsHead* hh = parseWCS(str);
    rr->replaceWCS(hh);
    delete hh;
  }
  else
    result = TCL_ERROR;
}

void Base::wcsReplaceCmd(int which, const char* fn)
{
  if (!currentContext->cfits)
    return;

  ifstream str(fn);
  if (!str) {
    Tcl_AppendResult(interp, " unable to load wcs file ", fn, NULL);
    result = TCL_ERROR;
    return;
  }

  FitsImage* rr = findAllFits(which);
  if (rr) {
    FitsHead* hh = parseWCS(str);
    rr->replaceWCS(hh);
    delete hh;
  }
  else
    result = TCL_ERROR;
}

void Base::wcsReplaceTxtCmd(int which, const char* txt)
{
  if (!currentContext->cfits)
    return;

  istringstream str(txt);
  if (!str) {
    Tcl_AppendResult(interp, " unable to process text", NULL);
    result = TCL_ERROR;
    return;
  }

  FitsImage* rr = findAllFits(which);
  if (rr) {
    FitsHead* hh = parseWCS(str);
    rr->replaceWCS(hh);
    delete hh;
  }
  else
    result = TCL_ERROR;
}

void Base::zoomCmd(const Vector& z)
{
  Vector az = ((Vector&)z).abs();
  zoom_[0] *= az[0];
  zoom_[1] *= az[1];

  update(MATRIX);
}

void Base::zoomToCmd(const Vector& z)
{
  zoom_ = ((Vector&)z).abs();
  update(MATRIX);
}
