// Copyright (C) 1999-2010
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "composite.h"
#include "framebase.h"
#include "fitsimage.h"

Composite::Composite(const Composite& a) : Marker(a) 
{
  members = a.members;
  global = a.global;
}

Composite::Composite(FrameBase* p, const Vector& ctr, 
		     double ang, int gl,
		     const char* clr, int* dsh, 
		     int wth, const char* fnt, const char* txt,
		     unsigned short prop, const char* cmt,
		     const List<Tag>& tg, const List<CallBack>& cb)
  : Marker(p, ctr, ang, clr, dsh, wth, fnt, txt, prop, cmt, tg, cb)
{
  strcpy(type, "composite");

  global = gl;

  handle = new Vector[4];
  numHandle = 4;

  updateBBox();
}

void Composite::draw(Drawable drawable, int tt)
{
  renderXHandles(drawable);
  if (tt)
    renderXText(drawable, parent->refToWidget, SRC);

  Marker* mk=members.head();
  while (mk) {
    Marker* m = mk->dup();
    m->setComposite(fwdRefMatrix(), angle);
    if (global)
      m->setComposite(properties | INCLUDE, colorName, lineWidth, highlited);
    m->draw(drawable, tt);
    delete m;
    mk=mk->next();
  }

  renderX(drawable, parent->refToWidget, SRC);
}

void Composite::drawXOR()
{
  renderXText(Tk_WindowId(parent->tkwin), parent->refToWindow, XOR);

  Marker* mk=members.head();
  while (mk) {
    Marker* m = mk->dup();
    m->setComposite(fwdRefMatrix(), angle);
    if (global)
      m->setComposite(properties | INCLUDE, colorName, lineWidth, highlited);
    m->drawXOR();
    delete m;
    mk=mk->next();
  }

  renderX(Tk_WindowId(parent->tkwin), parent->refToWindow, XOR);
}

void Composite::drawMagnifier(const Matrix& mx, int tt)
{
  if (tt)
    renderXText(parent->magnifierPixmap, mx, SRC);

  Marker* mk=members.head();
  while (mk) {
    Marker* m = mk->dup();
    m->setComposite(fwdRefMatrix(), angle);
    if (global)
      m->setComposite(properties | INCLUDE, colorName, lineWidth, highlited);
    m->drawMagnifier(mx, tt);
    delete m;
    mk=mk->next();
  }

  renderX(parent->magnifierPixmap, mx, SRC);
}

void Composite::renderX(Drawable drawable, const Matrix& mx, RenderMode mode)
{
  // INCLUDE/EXCLUDE
  GC lgc = renderXGC(mode);
  if (!(properties & INCLUDE)) {
    Vector ll = (handle[0] * parent->getCanvasToRef() * mx).round();
    Vector ur = (handle[2] * parent->getCanvasToRef() * mx).round();

    if (mode==SRC)
      XSetForeground(display, gc, parent->getColor("red"));

    XDrawLine(display, drawable, lgc, ll[0], ll[1], ur[0], ur[1]);    
  }
}

void Composite::ps(int mode, int tt)
{
  if (tt)
    renderPSText(mode);

  Marker* mk=members.head();
  while (mk) {
    Marker* m = mk->dup();
    m->setComposite(fwdRefMatrix(), angle);
    if (global)
      m->setComposite(properties | INCLUDE, colorName, lineWidth, highlited);
    m->ps(mode,tt);
    delete m;
    mk=mk->next();
  }

  renderPS(mode);
}

void Composite::renderPS(int mode)
{
  // INCLUDE/EXCLUDE
  renderPSGC(mode);
  if (!(properties & INCLUDE)) {
    renderPSColor(mode, parent->getXColor("red"));
    ostringstream str;
    str << "newpath " 
	<< handle[0].TkCanvasPs(parent->canvas) << "moveto"
	<< handle[2].TkCanvasPs(parent->canvas) << "lineto"
	<< " stroke" << endl << ends;
    Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);
  }
}

#ifdef _MACOSX
void Composite::macosx(int tt)
{
  if (tt)
    renderMACOSXText();

  Marker* mk=members.head();
  while (mk) {
    Marker* m = mk->dup();
    m->setComposite(fwdRefMatrix(), angle);
    if (global)
      m->setComposite(properties | INCLUDE, colorName, lineWidth, highlited);
    m->macosx(tt);
    delete m;
    mk=mk->next();
  }

  renderMACOSX();
}

void Composite::renderMACOSX()
{
  // INCLUDE/EXCLUDE
  renderMACOSXGC();
  if (!(properties & INCLUDE)) {
    macosxColor(parent->getXColor("red"));
    macosxDrawLine(handle[0], handle[2]);
  }
}
#endif

#ifdef _WIN32
void Composite::win32(int tt)
{
  if (tt)
    renderWIN32Text();

  Marker* mk=members.head();
  while (mk) {
    Marker* m = mk->dup();
    m->setComposite(fwdRefMatrix(), angle);
    if (global)
      m->setComposite(properties | INCLUDE, colorName, lineWidth, highlited);
    m->win32(tt);
    delete m;
    mk=mk->next();
  }

  renderWIN32();
}

void Composite::renderWIN32()
{
  // INCLUDE/EXCLUDE
  renderWIN32GC();
  if (!(properties & INCLUDE)) {
    win32Color(parent->getXColor("red"));
    win32DrawLine(handle[0], handle[2]);
  }
}
#endif

void Composite::updateBBox()
{
  // Handles
  BBox bb(center * bckCanvasMatrix());
  Marker* mk=members.head();
  while (mk) {
    Marker* m = mk->dup();
    m->setComposite(fwdRefMatrix(), angle);
    for(int i=0; i<m->getNumHandle(); i++)
      bb.bound(m->getHandle(i) * bckRefMatrix());
    delete m;
    mk=mk->next();
  }
  bb.expand(3); // a little more room around the edges

  Matrix mm = fwdCanvasMatrix();
  handle[0] = bb.ll   * mm;
  handle[1] = bb.lr() * mm;
  handle[2] = bb.ur   * mm;
  handle[3] = bb.ul() * mm;

  // bbox
  bbox = BBox(center * parent->refToCanvas);
  for (int i=0; i<numHandle; i++)
    bbox.bound(handle[i]);
  bbox.expand(3); // give us room for control handles

  // calculate overall bbox
  calcAllBBox();
}

void Composite::updateCoords(const Matrix& mx)
{
  Marker* mk=members.head();
  while (mk) {
    Vector cc = center;
    mk->setComposite(fwdRefMatrix(), angle);
    mk->updateCoords(mx);
    center = cc*mx;
    mk->setComposite(bckCanvasMatrix(), -angle);
    center = cc;
    mk=mk->next();
  }

  Marker::updateCoords(mx);
}

int Composite::isIn(const Vector& v)
{
  if (!bbox.isIn(v))
    return 0;

  Marker* mk=members.head();
  while (mk) {
    Marker* m = mk->dup();
    m->setComposite(fwdRefMatrix(), angle);
    if (m->isIn(v)) {
      delete m;
      return 1;
    }
    delete m;

    mk=mk->next();
  }

  return 0;
}

void Composite::append(Marker* m)
{
  m->setComposite(bckCanvasMatrix(), -angle);

  members.append(m);
  updateBBox();
}

Marker* Composite::extract()
{
  Marker* mk=members.head();
  if (mk) {
    members.extractNext(mk);
    mk->setComposite(fwdRefMatrix(), angle);
    updateBBox();
  }
  return mk;
}

// list

void Composite::list(ostream& str, CoordSystem sys, SkyFrame sky, 
		 SkyFormat format, int conj, int strip)
{
  if (!strip) {
    FitsImage* ptr = parent->findFits(sys,center);
    listPre(str, sys, sky, ptr, strip, 1);

    switch (sys) {
    case IMAGE:
    case PHYSICAL:
    case DETECTOR:
    case AMPLIFIER:
      {
	Vector v = ptr->mapFromRef(center,sys);
	str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	    << radToDeg(parent->mapAngleFromRef(angle,sys)) << ')';
      }
      break;
    default:
      if (ptr->hasWCS(sys)) {
	if (ptr->hasWCSEqu(sys)) {
	  switch (format) {
	  case DEGREES:
	    {
	      Vector v = ptr->mapFromRef(center,sys,sky);
	      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] 
		  << ','
		  << radToDeg(parent->mapAngleFromRef(angle,sys,sky)) << ')';
	    }
	    break;
	  case SEXAGESIMAL:
	    {
	      char buf[64];
	      ptr->mapFromRef(center,sys,sky,format,buf,64);
	      char ra[16];
	      char dec[16];
	      string x(buf);
	      istringstream wcs(x);
	      wcs >> ra >> dec;
	      str << type << '(' << ra << ',' << dec << ',' 
		  << radToDeg(parent->mapAngleFromRef(angle,sys,sky)) << ')';
	    }
	    break;
	  }
	}
	else {
	  Vector v = ptr->mapFromRef(center,sys);
	  str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ','
	      << radToDeg(parent->mapAngleFromRef(angle,sys)) << ')';
	}
      }
    }
      
    str << " ||";
    str << " composite=" << global;
    listProperties(str, 0);
  }

  Marker* mk=members.head();
  while (mk) {
    Marker* m = mk->dup();
    mk=mk->next();

    m->setComposite(fwdRefMatrix(), angle);
    m->list(str, sys, sky, format, (mk?1:0), strip);
    delete m;
  }
}

void Composite::listCiao(ostream& str, CoordSystem sys, int strip)
{
  Marker* mk=members.head();
  while (mk) {
    Marker* m = mk->dup();
    mk=mk->next();

    m->setComposite(fwdRefMatrix(), angle);
    m->listCiao(str, sys, strip);
    delete m;
  }
}

void Composite::listPros(ostream& str, CoordSystem sys, SkyFrame sky,
			 SkyFormat format, int strip)
{
  Marker* mk=members.head();
  while (mk) {
    Marker* m = mk->dup();
    m->setComposite(fwdRefMatrix(), angle);
    m->listPros(str, sys, sky, format, strip);
    delete m;

    mk=mk->next();
  }
}

void Composite::listSAOtng(ostream& str, CoordSystem sys, SkyFrame sky,
			   SkyFormat format, int strip)
{
  Marker* mk=members.head();
  while (mk) {
    Marker* m = mk->dup();
    m->setComposite(fwdRefMatrix(), angle);
    m->listSAOtng(str, sys, sky, format, strip);
    delete m;

    mk=mk->next();
  }
}

void Composite::listSAOimage(ostream& str, int strip)
{
  Marker* mk=members.head();
  while (mk) {
    Marker* m = mk->dup();
    m->setComposite(fwdRefMatrix(), angle);
    m->listSAOimage(str, strip);
    delete m;

    mk=mk->next();
  }
}

void Composite::listXY(ostream& str, CoordSystem sys, SkyFrame sky,
		       SkyFormat format, int strip)
{
  Marker* mk=members.head();
  while (mk) {
    Marker* m = mk->dup();
    m->setComposite(fwdRefMatrix(), angle);
    m->listXY(str, sys, sky, format, strip);
    delete m;

    mk=mk->next();
  }
}

