// Copyright (C) 1999-2010
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#ifndef __fitsimage_h__
#define __fitsimage_h__

#include "vector.h"
#include "vector3d.h"
#include "fitsdata.h"
#include "coord.h"
#include "file.h"
#include "wcs.h"

#define WCSXMAX 7

class Base;
class FrameBase;

class FitsImage {
 public:
  enum CompressType {GZIP, RICE, PLIO, HCOMPRESS};
  enum SmoothFunction {BOXCAR, TOPHAT, GAUSSIAN};

 protected:
  Base* parent;            // pointer to parent

  char* objectName;         // object name
  char* rootBaseFileName;   // base root filename
  char* fullBaseFileName;   // base full filename
  char* rootFileName;       // root filename
  char* fullFileName;       // full filename
  char* iisFileName;        // alt file name for iis

  // native fits file
  FitsFile* fits_;

  // base
  FitsFile* compress_;      // fits compressed
  FitsFile* hist_;          // fits bin

  FitsFile* base_;
  FitsData* basedata_;

  // analysis
  FitsFile* smooth_;
  FitsData* smoothdata_;

  // final
  FitsFile* image_;
  FitsData* data_;

  int naxis_[FTY_MAXAXES];
  int bitpix_;

  Vector histCursor;        // start point to bin about (physical coords)
  Vector actualHistCursor;  // actual bin center (physical coords)

  int doSmooth_;
  SmoothFunction smoothFunction_;
  int smoothRadius_;

  FitsHist::Function binFunction_; // binning operation
  Vector binFactor_;        // binning factor
  int binBufferSize_;       // size of bin buffer
  int binDepth_;            // binning depth

  char buf[32];             // tmp storage for returning strings

  // Mosaic items
  int keyLTMV;
  int keyATMV;
  int keyDTMV;
  int keyDATASEC;

  BBox datasec;             // DATASEC keyword
  
  Matrix wcsToRef;          // iraf/wcs matrix
  Matrix irafToRef;

  Matrix amplifierToCanvas;
  Matrix amplifierToPhysical;
  Matrix amplifierToImage;
  Matrix amplifierToRef;
  Matrix amplifierToPanner;
  Matrix amplifierToUser;
  Matrix canvasToAmplifier;
  Matrix canvasToPhysical;
  Matrix canvasToData;
  Matrix canvasToDetector;
  Matrix canvasToImage;
  Matrix canvasToRef;
  Matrix physicalToAmplifier;
  Matrix physicalToCanvas;
  Matrix physicalToData;
  Matrix physicalToDetector;
  Matrix physicalToImage;
  Matrix physicalToRef;
  Matrix physicalToUser;
  Matrix physicalToPanner;
  Matrix dataToCanvas;
  Matrix dataToPhysical;
  Matrix dataToImage;
  Matrix dataToMagnifier;
  Matrix dataToRef;
  Matrix dataToPanner;
  Matrix dataToPS;
  Matrix dataToUser;
  Matrix dataToWidget;
  Matrix detectorToCanvas;
  Matrix detectorToPhysical;
  Matrix detectorToImage;
  Matrix detectorToPanner;
  Matrix detectorToRef;
  Matrix detectorToUser;
  Matrix imageToCanvas;
  Matrix imageToPhysical;
  Matrix imageToAmplifier;
  Matrix imageToDetector;
  Matrix imageToData;
  Matrix imageToRef;
  Matrix imageToPanner;
  Matrix imageToUser;
  Matrix imageToWidget;
  Matrix magnifierToData;
  Matrix pannerToAmplifier;
  Matrix pannerToPhysical;
  Matrix pannerToData;
  Matrix pannerToDetector;
  Matrix pannerToImage;
  Matrix refToPhysical;
  Matrix refToData;
  Matrix refToImage;
  Matrix refToAmplifier;
  Matrix refToDetector;
  Matrix refToCanvas;
  Matrix userToAmplifier;
  Matrix userToPhysical;
  Matrix userToData;
  Matrix userToDetector;
  Matrix userToImage;
  Matrix widgetToData;
  Matrix widgetToImage;
  Matrix PSToData;

  // this is for optimization of frame
  double* mWidgetToData;
  double* mPannerToData;
  double* mMagnifierToData;
  double* mPSToData;

  Matrix3d dataToRef3d;
  Matrix3d dataToPanner3d;
  Matrix3d dataToMagnifier3d;
  Matrix3d dataToPS3d;
  Matrix3d dataToWidget3d;
  Matrix3d refToData3d;
  Matrix3d pannerToData3d;
  Matrix3d magnifierToData3d;
  Matrix3d widgetToData3d;
  Matrix3d PSToData3d;

  double* mWidgetToData3d;
  double* mPannerToData3d;
  double* mMagnifierToData3d;
  double* mPSToData3d;

  FitsBound iparams; // image clip/render bbox
  FitsBound dparams; // DATASEC clip/render bbox
  FitsBound imparams; // user specified minmax scan bbox
  FitsBound irparams; // user specified valid data bbox
  FitsBound dmparams; // DATASEC user specified minmax scan bbox
  FitsBound drparams; // DATASEC user specified valid data bbox

  int iisMode;              // flag to indicate if iis
  Vector iisz;              // z1 and z2
  int iiszt;                // 0-Uniary, 1-Linear, 2-Log

  FitsImage* nextMosaic_;   // next mosaic
  FitsImage* nextSlice_;    // next slice

  WorldCoor** wcs;          // wcs list
  FitsHead* wcsHeader; // alt wcs header

  int wcsx[WCSXMAX*MULTWCS]; // xth Axis WCS
  double crpixx[WCSXMAX*MULTWCS];
  double crvalx[WCSXMAX*MULTWCS];
  double cdx[WCSXMAX*MULTWCS];

  void reset();

  void process(const char*);
  void initCompress();
  void initHist();
  Vector getHistCenter();

  char* root(const char*);
  char* strip(const char*);
  int parseSection(char*, Vector*, Vector*);

  int smooth();

  void internalError(const char*);

public:
  FitsImage(Base*);
  virtual ~FitsImage();

  void load();
  void analysis();

  FitsImage* nextMosaic() {return nextMosaic_;}
  FitsImage* nextSlice() {return nextSlice_;}
  void setNextMosaic(FitsImage* n) {nextMosaic_ = n;}
  void setNextSlice(FitsImage* n) {nextSlice_ = n;}

  FitsFile* fitsFile() {return fits_;}
  FitsFile* baseFile() {return base_;}
  FitsFile* imageFile() {return image_;}

  int isValid() {return image_ ? 1 : 0;}
  int isCompress() {return compress_ ? 1 : 0;}
  int isHist() {return hist_ ? 1 : 0;}

  int isImage() {return fits_ ? fits_->isImage() : 0;}
  int isBinTable() {return fits_ ? fits_->isBinTable() : 0;}

  void close() {if (fits_) fits_->done();}

  Matrix updateHistCenter();
  Matrix updateHistCursor();
  Matrix updateHist(const Vector&);
  Matrix nextHist(const Vector&);
  void calcHistSize();
  Vector getHistDim();
  Vector getHistColMinMax(const char*);
  Vector getHistColDim(const char*);
  Vector getHistCursor() {return actualHistCursor;}
  const char* getHistFilter();
  const char* getHistX();
  const char* getHistY();
  const char* getHistZ();
  char* getHistList();

  int doSmooth() {return doSmooth_;}
  SmoothFunction smoothFunction() {return smoothFunction_;}
  int smoothRadius() {return smoothRadius_;}

  void setDoSmooth(int d) {doSmooth_ = d;}
  void setSmoothFunction(SmoothFunction f) {smoothFunction_ = f;}
  void setSmoothRadius(int r) {smoothRadius_ = r;}

  FitsHist::Function binFunction() {return binFunction_;}
  Vector binFactor() {return binFactor_;}
  int binBufferSize() {return binBufferSize_;}
  int binDepth() {return binDepth_;}

  void setBinCursor(const Vector&);
  void setBinFunction(FitsHist::Function f) {binFunction_ = f;}
  void setBinFactor(const Vector&);
  void setBinToFactor(const Vector&);
  void setBinBufferSize(int s) {binBufferSize_ = s;}
  void setBinDepth(int d) {binDepth_ = d < 1 ? 1 : d;}
  void setBinX(const char* str) {if (fits_) fits_->setpBinX(str);}
  void setBinY(const char* str) {if (fits_) fits_->setpBinY(str);}
  void setBinZ(const char* str) {if (fits_) fits_->setpBinZ(str);}
  void setBinFilter(const char* fil) {if (fits_) fits_->setpFilter(fil);}
  void setBinSliceFilter(const char* fil) 
    {if (fits_) fits_->setpSliceFilter(fil);}
  void setBinColMinMax(const char* str, const Vector& lim)
    {if (fits_) fits_->setColMinMax(str,lim);}

  const char* getObjectName() {return objectName;}
  void setObjectName(const char*);
  const char* getRootBaseFileName() {return rootBaseFileName;}
  const char* getFullBaseFileName() {return fullBaseFileName;}
  const char* getRootFileName() {return rootFileName;}
  const char* getFullFileName() {return fullFileName;}
  const char* iisGetFileName() {return iisFileName;}
  void iisSetFileName(const char*);
  void setFileName(const char*);
  void updateFileName();

  long naxis(int ii) {return naxis_[ii];}
  long width() {return naxis_[0];}
  long height() {return naxis_[1];}
  long depth() {return naxis_[2]>0?naxis_[2]:1;}
  int naxes();

  int bitpix() {return bitpix_;}
  int ext() {return fits_->ext();}

  FitsBound* getDataParams(FrScale::ScanMode);   // return bbox in IMAGE
  FitsBound* getMinMaxParams(FrScale::ScanMode);
  void setDataParams(int,int,int,int);
  void setMinMaxParams(int,int,int,int);

  int hasLTMV() {return keyLTMV;}
  int hasATMV() {return keyATMV;}
  int hasDTMV() {return keyDTMV;}
  int isIIS() {return iisMode;}

  char* pix2wcs(Vector, CoordSystem, SkyFrame, SkyFormat, char*, int);
  Vector pix2wcs(Vector, CoordSystem);
  Vector pix2wcs(Vector, CoordSystem, SkyFrame);
  Vector wcs2pix(Vector, CoordSystem);
  Vector wcs2pix(Vector, CoordSystem, SkyFrame);
  Vector pix2wcsDist(Vector, CoordSystem);
  Vector wcs2pixDist(Vector, CoordSystem);
  double wcsdist(Vector, Vector, CoordSystem);

  double pix2wcsx(double, CoordSystem, int);
  double wcs2pixx(double, CoordSystem, int);

  void initWCS(FitsHead*);
  void listWCS(ostream&, CoordSystem);
  void resetWCS();
  void replaceWCS(FitsHead*);
  void appendWCS(FitsHead*);
  void initWCS0(const Vector&);
  void resetWCS0();

  void processKeywords();
  int processKeywordsIRAF(FitsImage*);
  int processKeywordsWCS(FitsImage*, CoordSystem);

  WorldCoor* getWCS(CoordSystem sys) 
    {return (wcs && wcs[sys-WCS]) ? wcs[sys-WCS] : NULL;}
  const char* getWCSName(CoordSystem sys) 
    {return (wcs && wcs[sys-WCS]) ? wcs[sys-WCS]->wcsname : NULL;}
  Orientation getWCSOrientation(CoordSystem, SkyFrame);
  double getWCSRotation(CoordSystem, SkyFrame);
  Vector getWCScrval(CoordSystem);
  Vector getWCScrpix(CoordSystem);
  Vector getWCScdelt(CoordSystem);

  int hasWCS(CoordSystem);
  int hasWCSx(CoordSystem, int);
  int hasWCSEqu(CoordSystem);
  int hasWCSLinear(CoordSystem);

  void updateMatrices(Matrix&, Matrix&, Matrix&, Matrix&, Matrix&);
  void updateMatrices(Matrix3d&, Matrix3d&, Matrix3d&, Matrix3d&);

  void updateMagnifierMatrices(Matrix&);
  void updateMagnifierMatrices(Matrix3d&);

  void updatePS(Matrix);
  void updatePS(Matrix3d);

  double* getmWidgetToData() {return mWidgetToData;}
  double* getmPannerToData() {return mPannerToData;}
  double* getmMagnifierToData() {return mMagnifierToData;}
  double* getmPSToData() {return mPSToData;}

  Matrix& getAmplifierToPhysical() {return amplifierToPhysical;}
  Matrix& getAmplifierToImage() {return amplifierToImage;}
  Matrix& getAmplifierToRef() {return amplifierToRef;}
  Matrix& getAmplifierToUser() {return amplifierToUser;}
  Matrix& getCanvasToAmplifier() {return canvasToAmplifier;}
  Matrix& getCanvasToPhysical() {return canvasToPhysical;}
  Matrix& getCanvasToData() {return canvasToData;}
  Matrix& getCanvasToDetector() {return canvasToDetector;}
  Matrix& getCanvasToImage() {return canvasToImage;}
  Matrix& getCanvasToRef() {return canvasToRef;}
  Matrix& getPhysicalToAmplifier() {return physicalToAmplifier;}
  Matrix& getPhysicalToDetector() {return physicalToDetector;}
  Matrix& getPhysicalToImage() {return physicalToImage;}
  Matrix& getPhysicalToRef() {return physicalToRef;}
  Matrix& getPhysicalToUser() {return physicalToUser;}
  Matrix& getDataToCanvas() {return dataToCanvas;}
  Matrix& getDataToPhysical() {return dataToPhysical;}
  Matrix& getDataToImage() {return dataToImage;}
  Matrix& getDataToPanner() {return dataToPanner;}
  Matrix& getDataToRef() {return dataToRef;}
  Matrix& getDataToUser() {return dataToUser;}
  Matrix& getDataToWidget() {return dataToWidget;}
  Matrix& getDetectorToPhysical() {return detectorToPhysical;}
  Matrix& getDetectorToImage() {return detectorToImage;}
  Matrix& getDetectorToRef() {return detectorToRef;}
  Matrix& getDetectorToUser() {return detectorToUser;}
  Matrix& getImageToCanvas() {return imageToCanvas;}
  Matrix& getImageToPhysical() {return imageToPhysical;}
  Matrix& getImageToAmplifier() {return imageToAmplifier;}
  Matrix& getImageToDetector() {return imageToDetector;}
  Matrix& getImageToData() {return imageToData;}
  Matrix& getImageToRef() {return imageToRef;}
  Matrix& getImageToUser() {return imageToUser;}
  Matrix& getImageToWidget() {return imageToWidget;}
  Matrix& getMagnifierToData() {return magnifierToData;}
  Matrix& getRefToAmplifier() {return refToAmplifier;}
  Matrix& getRefToPhysical() {return refToPhysical;}
  Matrix& getRefToData() {return refToData;}
  Matrix& getRefToDetector() {return refToDetector;}
  Matrix& getRefToImage() {return refToImage;}
  Matrix& getRefToCanvas() {return refToCanvas;}
  Matrix& getPannerToAmplifier() {return pannerToAmplifier;}
  Matrix& getPannerToPhysical() {return pannerToPhysical;}
  Matrix& getPannerToData() {return pannerToData;}
  Matrix& getPannerToDetector() {return pannerToDetector;}
  Matrix& getPannerToImage() {return pannerToImage;}
  Matrix& getPhysicalToData() {return physicalToData;}
  Matrix& getUserToAmplifier() {return userToAmplifier;}
  Matrix& getUserToPhysical() {return userToPhysical;}
  Matrix& getUserToData() {return userToData;}
  Matrix& getUserToDetector() {return userToDetector;}
  Matrix& getUserToImage() {return userToImage;}
  Matrix& getWidgetToData() {return widgetToData;}
  Matrix& getWidgetToImage() {return widgetToImage;}

  Matrix& getWCSToRef() {return wcsToRef;}

  // 3d
  double* getmMagnifierToData3d() {return mMagnifierToData3d;}
  double* getmPannerToData3d() {return mPannerToData3d;}
  double* getmWidgetToData3d() {return mWidgetToData3d;}
  double* getmPSToData3d() {return mPSToData3d;}

  Matrix3d& getDataToMagnifier3d() {return dataToMagnifier3d;}
  Matrix3d& getDataToPanner3d() {return dataToPanner3d;}
  Matrix3d& getDataToRef3d() {return dataToRef3d;}
  Matrix3d& getDataToWidget3d() {return dataToWidget3d;}
  Matrix3d& getMagnifierToData3d() {return magnifierToData3d;}
  Matrix3d& getPannerToData3d() {return pannerToData3d;}
  Matrix3d& getRefToData3d() {return refToData3d;}
  Matrix3d& getWidgetToData3d() {return widgetToData3d;}

  Vector mapFromRef(const Vector&, CoordSystem, SkyFrame =FK5);
  void mapFromRef(const Vector&, CoordSystem, SkyFrame, SkyFormat, char*,int);
  Vector mapToRef(const Vector&, CoordSystem, SkyFrame =FK5);
  double mapLenFromRef(double, CoordSystem, SkyFormat =DEGREES);
  Vector mapLenFromRef(const Vector&, CoordSystem, SkyFormat =DEGREES);
  double mapLenToRef(double, CoordSystem, SkyFormat =DEGREES);
  Vector mapLenToRef(const Vector&, CoordSystem, SkyFormat =DEGREES);
  double mapDistFromRef(const Vector&, const Vector&, CoordSystem,
		       SkyFormat =DEGREES);
  double mapFromImagex(double, CoordSystem, int);
  double mapToImagex(double, CoordSystem, int);

// we assume that 'data' is valid

  const char* getValue(const Vector& v);
  float getValueFloat(const Vector& v) {return data_->getValueFloat(v);}
  double getValueDouble(const Vector& v) {return data_->getValueDouble(v);}
  int getValueMask(const Vector& v) {return data_->getValueMask(v);}
  int getValueMask(double x, double y) {return data_->getValueMask(x,y);}

  float getValueFloat(long i) {return data_->getValueFloat(i);}
  double getValueDouble(long i) {return data_->getValueDouble(i);}
  int getValueMask(long i) {return data_->getValueMask(i);}

  void setClip(double l, double h) {data_->setClip(l,h);}

  const char* getMin() {return data_ ? data_->getMin() : NULL;}
  const char* getMax() {return data_ ? data_->getMax() : NULL;}

  double getMinDouble() {return data_ ? data_->getMinDouble() : 0;}
  double getMaxDouble() {return data_ ? data_->getMaxDouble() : 0;}

  const char* getLow() {return data_ ? data_->getLow() : NULL;}
  const char* getHigh() {return data_ ? data_->getHigh() : NULL;}
  double getLowDouble() {return data_ ? data_->getLowDouble() : 0;}
  double getHighDouble() {return data_ ? data_->getHighDouble() : 0;}

  void updateClip(FrScale* frs)
    {if (data_) data_->updateClip(frs,getMinMaxParams(frs->scanMode()));}

  int hasDATAMIN() {return data_ ? data_->hasDATAMIN() : 0;}
  int hasDATASEC() {return keyDATASEC;}
  int hasIRAFMIN() {return data_ ? data_->hasIRAFMIN() : 0;}

  void bin(double* b, int l, double mn, double mx, FrScale::ScanMode ds)
    {data_->bin(b,l,mn,mx,getDataParams(ds));}

  char* display(FitsHead*);
  char* displayHeader();
  char* displayPrimary();
  char* displayWCS();
  FitsHead* getHead();
  char* getCards();
  int getNCards();
  char* getKeyword(const char*);
  int findKeyword(const char*);

  int saveFitsImageFile(const char* fn, int compress)
    {return image_ ? image_->saveFitsImageFile(fn, compress) : 0;}

  int saveFitsImageChannel(Tcl_Interp* interp, const char* ch, 
				      int compress)
    {return image_ ? image_->saveFitsImageChannel(interp, ch, compress) : 0;}

  int saveFitsImageSocket(int s, int compress)
    {return image_ ? image_->saveFitsImageSocket(s, compress) : 0;}

  int saveFitsTableFile(const char* fn, int compress)
    {return fits_ ? fits_->saveFitsTableFile(fn, compress) : 0;}

  int saveFitsTableChannel(Tcl_Interp* interp, const char* ch, 
				      int compress)
    {return fits_ ? fits_->saveFitsTableChannel(interp, ch, compress): 0;}

  int saveFitsTableSocket(int s, int compress)
    {return fits_ ? fits_->saveFitsTableSocket(s, compress) : 0;}

  int saveArrayFile(const char* fn, FitsFile::ArchType endian)
  {return image_ ? image_->saveArrayFile(fn, endian) : 0;}

  int saveArrayChannel(Tcl_Interp* interp, const char* ch,
		       FitsFile::ArchType endian)
  {return image_ ? image_->saveArrayChannel(interp, ch, endian) : 0;}

  int saveArraySocket(int s, FitsFile::ArchType endian)
  {return image_ ? image_->saveArraySocket(s, endian) : 0;}
};

// Fits

class FitsImageFitsAlloc : public FitsImage {
public:
  FitsImageFitsAlloc(Base*, const char*, const char*, 
		     FitsFile::FlushMode);
};

class FitsImageFitsAllocGZ : public FitsImage {
public:
  FitsImageFitsAllocGZ(Base*, const char*, const char*, 
		       FitsFile::FlushMode);
};

class FitsImageFitsChannel : public FitsImage {
public:
  FitsImageFitsChannel(Base*, Tcl_Interp*, const char*, const char*,
		       FitsFile::FlushMode);
};

class FitsImageFitsMMap : public FitsImage {
public:
  FitsImageFitsMMap(Base*, const char*);
};

class FitsImageFitsSMMap : public FitsImage {
public:
  FitsImageFitsSMMap(Base*, const char*, const char*);
};

class FitsImageFitsMMapIncr : public FitsImage {
public:
  FitsImageFitsMMapIncr(Base*, const char*);
};

class FitsImageFitsShare : public FitsImage {
public:
  FitsImageFitsShare(Base*, Base::ShmType, int, const char*);
};

class FitsImageFitsSShare : public FitsImage {
public:
  FitsImageFitsSShare(Base*, Base::ShmType, int, int, const char*);
};

class FitsImageFitsSocket : public FitsImage {
public:
  FitsImageFitsSocket(Base*, int, const char*, FitsFile::FlushMode);
};

class FitsImageFitsSocketGZ : public FitsImage {
public:
  FitsImageFitsSocketGZ(Base*, int, const char*, FitsFile::FlushMode);
};

class FitsImageFitsVar : public FitsImage {
public:
  FitsImageFitsVar(Base*, Tcl_Interp*, const char*, const char*);
};

// Fits Next

class FitsImageFitsNextAlloc : public FitsImage {
public:
  FitsImageFitsNextAlloc(Base*, const char*, FitsFile*);
};

class FitsImageFitsNextAllocGZ : public FitsImage {
public:
  FitsImageFitsNextAllocGZ(Base*, const char*, FitsFile*);
};

class FitsImageFitsNextChannel : public FitsImage {
public:
  FitsImageFitsNextChannel(Base*, const char*, FitsFile*);
};

class FitsImageFitsNextMMap : public FitsImage {
public:
  FitsImageFitsNextMMap(Base*, const char*, FitsFile*);
};

class FitsImageFitsNextSMMap : public FitsImage {
public:
  FitsImageFitsNextSMMap(Base*, const char*, FitsFile*);
};

class FitsImageFitsNextMMapIncr : public FitsImage {
public:
  FitsImageFitsNextMMapIncr(Base*, const char*, FitsFile*);
};

class FitsImageFitsNextShare : public FitsImage {
public:
  FitsImageFitsNextShare(Base*, const char*, FitsFile*);
};

class FitsImageFitsNextSShare : public FitsImage {
public:
  FitsImageFitsNextSShare(Base*, const char*, FitsFile*);
};

class FitsImageFitsNextSocket : public FitsImage {
public:
  FitsImageFitsNextSocket(Base*, const char*, FitsFile*);
};

class FitsImageFitsNextSocketGZ : public FitsImage {
public:
  FitsImageFitsNextSocketGZ(Base*, const char*, FitsFile*);
};

class FitsImageFitsNextVar : public FitsImage {
public:
  FitsImageFitsNextVar(Base*, const char*, FitsFile*);
};

class FitsImageFitsNextHist : public FitsImage {
public:
  FitsImageFitsNextHist(Base*, FitsImage*, FitsFile*);
};

class FitsImageFitsNextCompress : public FitsImage {
public:
  FitsImageFitsNextCompress(Base*, FitsImage*, FitsFile*);
};

// Array

class FitsImageArrAlloc : public FitsImage {
public:
  FitsImageArrAlloc(Base*, const char*, const char*, 
		    FitsFile::FlushMode);
};

class FitsImageArrAllocGZ : public FitsImage {
public:
  FitsImageArrAllocGZ(Base*, const char*, const char*, 
		      FitsFile::FlushMode);
};

class FitsImageArrChannel : public FitsImage {
public:
  FitsImageArrChannel(Base*, Tcl_Interp*, const char*, const char*, 
		      FitsFile::FlushMode);
};

class FitsImageArrMMap : public FitsImage {
public:
  FitsImageArrMMap(Base*, const char*);
};

class FitsImageArrMMapIncr : public FitsImage {
public:
  FitsImageArrMMapIncr(Base*, const char*);
};

class FitsImageArrShare : public FitsImage {
public:
  FitsImageArrShare(Base*, Base::ShmType, int, const char*);
};

class FitsImageArrSocket : public FitsImage {
public:
  FitsImageArrSocket(Base*, int, const char*, FitsFile::FlushMode);
};

class FitsImageArrSocketGZ : public FitsImage {
public:
  FitsImageArrSocketGZ(Base*, int, const char*, FitsFile::FlushMode);
};

class FitsImageArrVar : public FitsImage {
public:
  FitsImageArrVar(Base*, Tcl_Interp*, const char*, const char*);
};

// Mosaic

class FitsImageMosaicAlloc : public FitsImage {
public:
  FitsImageMosaicAlloc(Base*, const char*, const char*, 
		       FitsFile::FlushMode);
};

class FitsImageMosaicAllocGZ : public FitsImage {
public:
  FitsImageMosaicAllocGZ(Base*, const char*, const char*, 
			 FitsFile::FlushMode);
};

class FitsImageMosaicChannel : public FitsImage {
public:
  FitsImageMosaicChannel(Base*, Tcl_Interp*, const char*, const char*,
			 FitsFile::FlushMode);
};

class FitsImageMosaicMMap : public FitsImage {
public:
  FitsImageMosaicMMap(Base*, const char*);
};

class FitsImageMosaicMMapIncr : public FitsImage {
public:
  FitsImageMosaicMMapIncr(Base*, const char*);
};

class FitsImageMosaicShare : public FitsImage {
public:
  FitsImageMosaicShare(Base*, Base::ShmType, int, const char*);
};

class FitsImageMosaicSocket : public FitsImage {
public:
  FitsImageMosaicSocket(Base*, int, const char*, FitsFile::FlushMode);
};

class FitsImageMosaicSocketGZ : public FitsImage {
public:
  FitsImageMosaicSocketGZ(Base*, int, const char*, FitsFile::FlushMode);
};

class FitsImageMosaicVar : public FitsImage {
public:
  FitsImageMosaicVar(Base*, Tcl_Interp*, const char*, const char*);
};

// Mosaic Next

class FitsImageMosaicNextAlloc : public FitsImage {
public:
  FitsImageMosaicNextAlloc(Base*, const char*, FitsFile*, 
			   FitsFile::FlushMode);
};

class FitsImageMosaicNextAllocGZ : public FitsImage {
public:
  FitsImageMosaicNextAllocGZ(Base*, const char*, FitsFile*, 
			     FitsFile::FlushMode);
};

class FitsImageMosaicNextChannel : public FitsImage {
public:
  FitsImageMosaicNextChannel(Base*, const char*, FitsFile*, 
			     FitsFile::FlushMode);
};

class FitsImageMosaicNextMMap : public FitsImage {
public:
  FitsImageMosaicNextMMap(Base*, const char*, FitsFile*);
};

class FitsImageMosaicNextMMapIncr : public FitsImage {
public:
  FitsImageMosaicNextMMapIncr(Base*, const char*, FitsFile*);
};

class FitsImageMosaicNextShare : public FitsImage {
public:
  FitsImageMosaicNextShare(Base*, const char*, FitsFile*);
};

class FitsImageMosaicNextSocket : public FitsImage {
public:
  FitsImageMosaicNextSocket(Base*, const char*, FitsFile*, 
			    FitsFile::FlushMode);
};

class FitsImageMosaicNextSocketGZ : public FitsImage {
public:
  FitsImageMosaicNextSocketGZ(Base*, const char*, FitsFile*, 
			      FitsFile::FlushMode);
};

class FitsImageMosaicNextVar : public FitsImage {
public:
  FitsImageMosaicNextVar(Base*, const char*, FitsFile*);
};

// IIS

class FitsImageIIS : public FitsImage {
public:
  FitsImageIIS(Base*, int, int);
  void iisErase();
  const char* iisGet(int x, int y);
  void iisSet(int x, int y, const char* d, int l);
  void iisWCS(const Matrix&, const Vector&, int);
};

#endif
