// Copyright (C) 1999-2010
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "frame.h"
#include "fitsimage.h"
#include "ps.h"
#include "NaN.h"

#include "sigbus.h"

// Frame Member Functions

Frame::Frame(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item) 
  : FrameBase(i,c,item)
{
  currentContext = &context;
  keyContext = &context;
  keyContextSet =1;
  
  cmapID = 1;
  bias = 0.5;
  contrast = 1.0;

  colorCount = 0;
  colorScale = NULL;
  colorCells = NULL;
  indexCells = NULL;

  iisSaveRoot = 0;
  iisInteractive = 0;
}

Frame::~Frame()
{
  if (colorScale)
    delete colorScale;

  if (colorCells)
    delete [] colorCells;

  if (indexCells)
    delete [] indexCells;
}

void Frame::alignWCS()
{
  if (!wcsAlignment || !context.cfits || !hasWCS(wcsSystem)) {
    wcsOrientation = NORMAL;
    wcsOrientationMatrix.identity();
    wcsRotation = 0;
    return;
  }

  calcAlignWCS(context.cfits, wcsSystem, wcsSky,
	       &wcsOrientation, &wcsOrientationMatrix, &wcsRotation);
}

void Frame::alignWCS(FitsImage* ptr)
{
  if (!wcsAlignment || !ptr || !context.cfits || !hasWCS(wcsSystem)) {
    wcsOrientation = NORMAL;
    wcsOrientationMatrix.identity();
    wcsRotation = 0;
    return;
  }

  calcAlignWCS(ptr, context.cfits, wcsSystem, wcsSky,
	       &wcsOrientation, &wcsOrientationMatrix, &wcsRotation, &zoom_);
}

unsigned char* Frame::fillImage(int width, int height,
				double* (FitsImage::*getMatrix)())
{
  // img
  unsigned char* img = new unsigned char[width*height*3];
  {
    unsigned char* ptr = img;
    for (int jj=0; jj<height; jj++)
      for (int ii=0; ii<width; ii++) {
	*ptr++ = (unsigned char)bgColor->red;
	*ptr++ = (unsigned char)bgColor->green;
	*ptr++ = (unsigned char)bgColor->blue;
      }	
  }

  if (!context.cfits)
    return img;

  // basics
  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  FitsImage* sptr = context.cfits;
  int mosaic = isMosaic();

  // variable
  double* mm = (sptr->*getMatrix)();
  FitsBound* params = sptr->getDataParams(context.frScale.scanMode());
  int srcw = sptr->width();

  double ll = sptr->getLowDouble();
  double hh = sptr->getHighDouble();
  double diff = hh - ll;

  // main loop
  unsigned char* dest = img;

  SETSIGBUS
  for (long jj=0; jj<height; jj++) {
    for (long ii=0; ii<width; ii++, dest+=3) {
      if (mosaic) {
	sptr = context.cfits;

	mm = (sptr->*getMatrix)();
	params = sptr->getDataParams(context.frScale.scanMode());
	srcw = sptr->width();

	ll = sptr->getLowDouble();
	hh = sptr->getHighDouble();
	diff = hh - ll;
      }

      do {
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	if (xx>=params->xmin && xx<params->xmax && 
	    yy>=params->ymin && yy<params->ymax) {
	  double value = sptr->getValueDouble(long(yy)*srcw + long(xx));

	  if (!isnand(value)) {
	    if (value <= ll) {
	      *(dest+2) = table[0];
	      *(dest+1) = table[1];
	      *dest = table[2];
	    }
	    else if (value >= hh) {
	      *(dest+2) = table[length*3];
	      *(dest+1) = table[length*3+1];
	      *dest = table[length*3+2];
	    }
	    else {
	      int l = (int)(((value - ll)/diff * length) + .5);
	      *(dest+2) = table[l*3];
	      *(dest+1) = table[l*3+1];
	      *dest = table[l*3+2];
	    }
	  }
	  else {
	    *(dest+2) = nanColor->blue;
	    *(dest+1) = nanColor->green;
	    *dest = nanColor->red;
	  }

	  break;
	}
	else {
	  if (mosaic) {
	    sptr = sptr->nextMosaic();

	    if (sptr) {
	      mm = (sptr->*getMatrix)();
	      params = sptr->getDataParams(context.frScale.scanMode());
	      srcw = sptr->width();

	      ll = sptr->getLowDouble();
	      hh = sptr->getHighDouble();
	      diff = hh - ll;
	    }
	  }
	}
      }
      while (mosaic && sptr);
    }
  }
  CLEARSIGBUS

  return img;
}

int Frame::isIIS() 
{
  return context.cfits && context.cfits->isIIS();
}

BBox Frame::imageBBox()
{
  // returns imageBBox in IMAGE coords
  //   and extends edge to edge

  BBox rr;
  FitsImage* ptr = context.fits;
  while (ptr) {
    FitsBound* params = ptr->getDataParams(context.frScale.scanMode());
    Matrix mm = ptr->getWCSToRef() * dataToImage;

    Vector aa =  Vector(params->xmin,params->ymin) * mm;
    if (ptr == context.fits)
      rr = BBox(aa,aa);
    else
      rr.bound(aa);

    rr.bound(Vector(params->xmax,params->ymin) * mm);
    rr.bound(Vector(params->xmax,params->ymax) * mm);
    rr.bound(Vector(params->xmin,params->ymax) * mm);

    ptr = ptr->nextMosaic();
  }

  return rr;
}

void Frame::reset()
{
  cmapID = 1;
  bias = 0.5;
  contrast = 1.0;

  FrameBase::reset();
}

void Frame::setBinCursor(const Vector& v)
{
  if (context.cfits)
    context.cfits->setBinCursor(cursor);
}

void Frame::updateBin(const Matrix& mx)
{
  // Note: cursor is in USER coords, imageCenter() in IMAGE coords
  cursor = imageCenter();
  crosshair *= mx;
  updateMarkerCoords(&userMarkers, mx);
  updateMarkerCoords(&catalogMarkers, mx);

  context.updateBin(this,mx);

  alignWCS();
  updateColorScale();
  updateNow(MATRIX); // because we have changed zoom

  // update markers call backs
  // wait til matrices have been updated so that any dialogs will print
  // the correct coords

  updateMarkerCBs(&userMarkers);
  updateMarkerCBs(&catalogMarkers);
}

void Frame::updateColorCells(unsigned short* index, 
			     unsigned char* cells, int cnt)
{
  // the colorbar widget will pass us a pointer to the indexCells

  colorCount = cnt;
  if (indexCells)
    delete [] indexCells;
  indexCells = new unsigned short[cnt];
  if (!indexCells) {
    internalError("Unable to Alloc indexCells");
    return;
  }
  memcpy(indexCells, index, cnt*sizeof(unsigned short));

  // copy the rgb vales to the colorCells array (for postscript printing)

  if (colorCells)
    delete [] colorCells;
  colorCells = new unsigned char[cnt*3];
  if (!colorCells) {
    internalError("Unable to Alloc colorCells");
    return;
  }
  memcpy(colorCells, cells, cnt*3);
}

void Frame::updateMagnifierMatrices(Matrix& refToMagnifier)
{
  FitsImage* ptr = context.fits;
  while (ptr) {
    FitsImage* sptr = ptr;
    while (sptr) {
      sptr->updateMagnifierMatrices(refToMagnifier);
      sptr = sptr->nextSlice();
    }
    ptr = ptr->nextMosaic();
  }
}

void Frame::updateMatrices()
{
  // very important! do this first
  FrameBase::updateMatrices();

  // alway identity
  Matrix rgbToRef; 

  FitsImage* ptr = context.fits;
  while (ptr) {
    FitsImage* sptr = ptr;
    while (sptr) {
      sptr->updateMatrices(rgbToRef, refToUser, userToWidget, 
			   widgetToCanvas, userToPanner);
      sptr = sptr->nextSlice();
    }
    ptr = ptr->nextMosaic();
  }
}

void Frame::updatePS(float scale, int width, int height)
{
  Matrix mx = psMatrix(scale, width, height);

  FitsImage* ptr = context.cfits;
  while (ptr) {
    ptr->updatePS(mx);
    ptr = ptr->nextMosaic();
  }
}

void Frame::unloadAllFits()
{
  unloadFits();
}

void Frame::unloadFits()
{
  if (DebugPerf)
    cerr << "Frame::unloadFits" << endl;

  // clean up from iis if needed
  if (isIIS()) {
    context.frScale.setClipMode(FrScale::MINMAX);
    context.frScale.setMMMode(FrScale::SAMPLE);
    context.frScale.setULow(DEFAULTLOW);
    context.frScale.setUHigh(DEFAULTHIGH);
    context.frScale.setColorScaleType(FrScale::LINEARSCALE);
  }

  context.unload();

  FrameBase::unloadFits();
}

// Commands

void Frame::getColorbarCmd()
{
  ostringstream str;
  str << cmapID << ' ' << bias << ' ' << contrast << ' ' << invert << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Frame::getRGBChannelCmd()
{
  Tcl_AppendResult(interp, "red", NULL);
}

void Frame::getRGBViewCmd()
{
  Tcl_AppendResult(interp, "1 1 1", NULL);
}

void Frame::getRGBSystemCmd()
{
  Tcl_AppendResult(interp, "image", NULL);
}

void Frame::getTypeCmd()
{
  Tcl_AppendResult(interp, "base", NULL);
}

void Frame::getWCSAlignCmd()
{
  ostringstream str;
  str << (unsigned short*)context.fits << ends;

  Tcl_AppendResult(interp, (wcsAlignment ? "1" : "0"), " ", 
		   coordSystemStr_[wcsSystem-IMAGE], " ",
		   skyFrameStr_[wcsSky-FK4], " ",
		   str.str().c_str(), NULL);
}

void Frame::iisCmd(int width, int height)
{
  unloadAllFits();

  // force clip mode to user between 1 to 200
  context.frScale.setClipMode(FrScale::USERCLIP);
  context.frScale.setMMMode(FrScale::SCAN);
  context.frScale.setULow(IISMIN);
  context.frScale.setUHigh(IISMAX);
  context.frScale.setColorScaleType(FrScale::IISSCALE);

  FitsImage* img = new FitsImageIIS(this, width, height);

  loadDone(currentContext->load(this, ALLOC, "", img, IMG),IMG);
}

void Frame::iisEraseCmd()
{
  if (context.cfits)
    ((FitsImageIIS*)context.cfits)->iisErase();
}

void Frame::iisGetCmd(char* ptr, int x, int y, int dx, int dy)
{
  if (context.cfits)
    memcpy(ptr, ((FitsImageIIS*)context.cfits)->iisGet(x,y), dx*dy);
}

void Frame::iisSetCmd(const char* data, int x, int y, int dx, int dy)
{
  if (context.cfits)
    ((FitsImageIIS*)context.cfits)->iisSet(x, y, data, dx*dy);
}

void Frame::iisWCSCmd(const Matrix& mx, const Vector& z, int zt)
{
  if (context.cfits)
    ((FitsImageIIS*)context.cfits)->iisWCS(mx, z, zt);
}

void Frame::sliceCmd(int ii, int ss)
{
  context.updateSlice(ii, ss);

  // execute any update callbacks
  updateCBMarkers();

  update(BASE);
}



