// Copyright (C) 1999-2010
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "frame3d.h"
#include "fitsimage.h"
#include "ps.h"
#include "NaN.h"

#include "sigbus.h"

Frame3d::Frame3d(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item) 
  : Frame3dBase(i,c,item)
{
  currentContext = &context;
  keyContext = &context;
  keyContextSet =1;

  cmapID = 1;
  bias = 0.5;
  contrast = 1.0;

  colorCount = 0;
  colorScale = NULL;
  colorCells = NULL;
  indexCells = NULL;
}

Frame3d::~Frame3d()
{
  if (colorScale)
    delete colorScale;

  if (colorCells)
    delete [] colorCells;

  if (indexCells)
    delete [] indexCells;
}

void Frame3d::alignWCS()
{
  if (!wcsAlignment || !context.cfits || !hasWCS(wcsSystem)) {
    wcsOrientation = NORMAL;
    wcsOrientationMatrix.identity();
    wcsRotation = 0;
    return;
  }

  calcAlignWCS(context.cfits, wcsSystem, wcsSky,
	       &wcsOrientation, &wcsOrientationMatrix, &wcsRotation);
}

void Frame3d::alignWCS(FitsImage* ptr)
{
  if (!wcsAlignment || !ptr || !context.cfits || !hasWCS(wcsSystem)) {
    wcsOrientation = NORMAL;
    wcsOrientationMatrix.identity();
    wcsRotation = 0;
    return;
  }

  Vector zoom(1,1);
  calcAlignWCS(ptr, context.cfits, wcsSystem, wcsSky,
	       &wcsOrientation, &wcsOrientationMatrix, &wcsRotation, &zoom);
}

unsigned char* Frame3d::fillImage(int width, int height,
				  double* (FitsImage::*getMatrix)())
{
  depth = zbufDepth_;

  if (getMatrix == &FitsImage::getmWidgetToData3d) {
    zbuf = &zbufWidget_;
    mkzbuf = &mkzbufWidget_;
  }
  else if (getMatrix == &FitsImage::getmPannerToData3d) {
    zbuf = &zbufPanner_;
    mkzbuf = &mkzbufPanner_;
  }
  else if (getMatrix == &FitsImage::getmPSToData3d) {
    zbuf = &zbufPS_;
    mkzbuf = &mkzbufPS_;
  }

  unsigned char* img =NULL;
  switch (renderMethod_) {
  case MIP:
    img = fillImageMIP(width, height, depth, getMatrix);
    break;
  case AIP:
    img = fillImageAIP(width, height, depth, getMatrix);
    break;
  case THRESHOLD:
    img = fillImageThreshold(width, height, depth, getMatrix);
    break;
  case BLEND:
    img = fillImageBlend(width, height, depth, getMatrix);
    break;
  }

  // this can be a memory hog, free it!
  if (getMatrix == &FitsImage::getmPSToData3d) {
    if (zbufPS_)
      delete [] zbufPS_;
    zbufPS_ = NULL;
    if (mkzbufPS_)
      delete [] mkzbufPS_;
    mkzbufPS_ = NULL;
  }

  if (!img) {
    internalError("fillImage Error");
    return NULL;
  }

  return img;
}

unsigned char* Frame3d::fillImageBlend(int width, int height, int depth,
				       double* (FitsImage::*getMatrix)())
{
  if (!*zbuf) {
    *zbuf = new float[width*height*depth];
    if (!(*zbuf))
      return NULL;
    memset(*zbuf, 0, width*height*depth*sizeof(float));

    *mkzbuf = new unsigned char[width*height*depth];
    if (!(*mkzbuf))
      return NULL;
    memset(*mkzbuf, 0, width*height*depth);

    FitsBound* params = context.fits->getDataParams(context.frScale.scanMode());
    int srcw = context.fits->width();
    int srch = context.fits->height();
    int srcd = context.fits->depth();

    double* mm = (context.fits->*getMatrix)();

    float* dest = *zbuf;
    unsigned char* mkdest = *mkzbuf;

    SETSIGBUS
    for (int jj=0; jj<height; jj++) {
      for (int ii=0; ii<width; ii++) {

	for (int kk=0; kk<depth; kk++, dest++, mkdest++) {
	  double xx = ii*mm[0] + jj*mm[4] + kk*mm[8]  + mm[12];
	  double yy = ii*mm[1] + jj*mm[5] + kk*mm[9]  + mm[13];
	  double zz = ii*mm[2] + jj*mm[6] + kk*mm[10] + mm[14];

	  if (xx>=params->xmin && xx<params->xmax && 
	      yy>=params->ymin && yy<params->ymax &&
	      zz>=0 && zz<srcd) {

	    float value = context.fits->getValueDouble(long(zz)*srcw*srch + 
					       long(yy)*srcw + long(xx));

	    if (!isnanf(value)) {
	      *dest =value;
	      *mkdest=1;
	    }
	  }
	}
      }
    }
    CLEARSIGBUS
  }

  // we go one more deep for bg
  int dd = depth+1;

  // create Z-buffer with alpha
  // stay with floating point
  unsigned char* acc = new unsigned char[width*height*dd*3];
  if (!acc)
    return NULL;
  memset(acc, 0, width*height*dd*3);

  float* alpha = new float[width*height*dd];
  if (!alpha)
    return NULL;
  memset(alpha, 0, width*height*dd*sizeof(float));

  unsigned char* mkacc = new unsigned char[width*height];
  if (!mkacc)
    return NULL;
  memset(mkacc, 0, width*height);

  {
    FitsBound* params = context.fits->getDataParams(context.frScale.scanMode());
    int srcw = context.fits->width();
    int srch = context.fits->height();
    int srcd = context.fits->depth();

    double* mm = (context.fits->*getMatrix)();

    int length = colorScale->size() - 1;
    const unsigned char* table = colorScale->psColors();

    double ll = context.fits->getLowDouble();
    double hh = context.fits->getHighDouble();
    double diff = hh - ll;

    unsigned char* dest = acc;
    float* alpha_dest = alpha;
    unsigned char* mkdest = mkacc;
    float* src = *zbuf;
    unsigned char* mksrc = *mkzbuf;
    for (int jj=0; jj<height; jj++) {
      for (int ii=0; ii<width; ii++, mkdest++) {
	for (int kk=0; kk<depth; kk++, dest+=3, alpha_dest++, src++, mksrc++) {
	  if (*mksrc) {
	    float value = *src;

	    if ((value>=ll) && (value<=hh)) {
	      int l = (int)(((value - ll)/diff * length) + .5);
	      float aa = (value-ll)/diff;
	      *dest = table[l*3+2]*aa; // red
	      *(dest+1) = table[l*3+1]*aa; // green
	      *(dest+2) = table[l*3]*aa; // blue
	      *alpha_dest = aa; // alpha

	      *mkdest=1;
	    }
	  }
	}

	// bg
	*dest++ = (unsigned char)bgColor->red;
	*dest++ = (unsigned char)bgColor->green;
	*dest++ = (unsigned char)bgColor->blue;
	*alpha_dest++ = 1.;
      }
    }
  }

  // bring color forward in Z-buffer
  // use front to back blend equation
  {
    unsigned char* fg = acc;
    float* alpha_fg = alpha;
    unsigned char* mksrc = mkacc;
    for (int jj=0; jj<height; jj++) {
      for (int ii=0; ii<width; ii++, mksrc++) {
	if (*mksrc) {
	  for (int kk=0; kk<depth; kk++) {
	    unsigned char* bg = fg+3;
	    float* alpha_bg = alpha_fg+1;

	    float tt = 1-*alpha_fg;
	    *bg++ = *fg++ + tt* *bg; // red
	    *bg++ = *fg++ + tt* *bg; // green
	    *bg++ = *fg++ + tt* *bg; // blue
	    *alpha_bg = *alpha_fg++ + tt* *alpha_bg; // alpha
	  }
	  fg+=3; // skip over last layer
	  alpha_fg++;
	}
	else {
	  fg+=(depth+1)*3;
	  alpha_fg+=(depth+1);
	}
      }
    }
  }

  // convert from float to unsigned char
  // image is in last layer of acc
  unsigned char* img = new unsigned char[width*height*3];
  if (!img)
    return NULL;
  memset(img, 0, width*height*3);

  {
    unsigned char* dest = img;
    unsigned char* src = acc + (depth*3);
    unsigned char* mksrc = mkacc;
    for (int jj=0; jj<height; jj++) {
      for (int ii=0; ii<width; ii++, src+=dd*3, mksrc++) {
	if (*mksrc) {
	  *dest++ = *src; // red
	  *dest++ = *(src+1); // green
	  *dest++ = *(src+2); // blue
	}
	else {
	  *dest++ = (unsigned char)bgColor->red;
	  *dest++ = (unsigned char)bgColor->green;
	  *dest++ = (unsigned char)bgColor->blue;
	}
      }
    }
  }

  if (acc)
    delete [] acc;

  if (alpha)
    delete [] alpha;

  if (mkacc)
    delete [] mkacc;

  return img;
}

unsigned char* Frame3d::fillImageColor(int width, int height)
{
  unsigned char* img = new unsigned char[width*height*3];
  memset(img, 0, width*height*3);

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  double ll = context.fits->getLowDouble();
  double hh = context.fits->getHighDouble();
  double diff = hh - ll;

  register unsigned char red = (unsigned char)bgColor->red;
  register unsigned char green = (unsigned char)bgColor->green;
  register unsigned char blue = (unsigned char)bgColor->blue; 

  unsigned char* dest = img;
  float* src = *zbuf;
  unsigned char* mksrc = *mkzbuf;

  for (int jj=0; jj<height; jj++) {
    for (int ii=0; ii<width; ii++, dest+=3, src++, mksrc++) {
      *dest = red;
      *(dest+1) = green;
      *(dest+2) = blue;

      if (*mksrc) {
	float value = *src;
	if ((value>=ll) && (value<=hh)) {
	  int l = (int)(((value - ll)/diff * length) + .5);
	  *dest = table[l*3+2]; // red
	  *(dest+1) = table[l*3+1]; // green
	  *(dest+2) = table[l*3]; // blue
	}
      }
    }
  }

  return img;
}

unsigned char* Frame3d::fillImageAIP(int width, int height, int depth,
				     double* (FitsImage::*getMatrix)())
{
  if (!*zbuf) {
    *zbuf = new float[width*height];
    memset(*zbuf, 0, width*height*sizeof(float));

    *mkzbuf = new unsigned char[width*height];
    memset(*mkzbuf, 0, width*height);

    FitsBound* params = context.fits->getDataParams(context.frScale.scanMode());
    int srcw = context.fits->width();
    int srch = context.fits->height();
    int srcd = context.fits->depth();

    double* mm = (context.fits->*getMatrix)();

    float* dest = *zbuf;
    unsigned char* mkdest = *mkzbuf;

    SETSIGBUS
    for (int jj=0; jj<height; jj++) {
      for (int ii=0; ii<width; ii++, dest++, mkdest++) {
	int cnt=0;
	float acc=0;

	for (int kk=0; kk<depth; kk++) {
	  double xx = ii*mm[0] + jj*mm[4] + kk*mm[8]  + mm[12];
	  double yy = ii*mm[1] + jj*mm[5] + kk*mm[9]  + mm[13];
	  double zz = ii*mm[2] + jj*mm[6] + kk*mm[10] + mm[14];

	  if (xx>=params->xmin && xx<params->xmax && 
	      yy>=params->ymin && yy<params->ymax &&
	      zz>=0 && zz<srcd) {

	    float value = context.fits->getValueDouble(long(zz)*srcw*srch + 
					       long(yy)*srcw + long(xx));

	    if (!isnanf(value)) {
	      cnt++;
	      acc+=value;
	    }
	  }
	}

	if (cnt) {
	  *dest =acc/cnt;
	  *mkdest=1;
	}
      }
    }
    CLEARSIGBUS
  }

  return fillImageColor(width, height);
}

unsigned char* Frame3d::fillImageMIP(int width, int height, int depth,
				     double* (FitsImage::*getMatrix)())
{
  if (!*zbuf) {
    *zbuf = new float[width*height];
    memset(*zbuf, 0, width*height*sizeof(float));

    *mkzbuf = new unsigned char[width*height];
    memset(*mkzbuf, 0, width*height);

    FitsBound* params = context.fits->getDataParams(context.frScale.scanMode());
    int srcw = context.fits->width();
    int srch = context.fits->height();
    int srcd = context.fits->depth();

    double* mm = (context.fits->*getMatrix)();

    float* dest = *zbuf;
    unsigned char* mkdest = *mkzbuf;

    SETSIGBUS
    for (int jj=0; jj<height; jj++) {
      for (int ii=0; ii<width; ii++, dest++, mkdest++) {
	float max = -FLT_MAX;

	for (int kk=0; kk<depth; kk++) {
	  double xx = ii*mm[0] + jj*mm[4] + kk*mm[8]  + mm[12];
	  double yy = ii*mm[1] + jj*mm[5] + kk*mm[9]  + mm[13];
	  double zz = ii*mm[2] + jj*mm[6] + kk*mm[10] + mm[14];

	  if (xx>=params->xmin && xx<params->xmax && 
	      yy>=params->ymin && yy<params->ymax &&
	      zz>=0 && zz<srcd) {

	    float value = context.fits->getValueDouble(long(zz)*srcw*srch + 
					       long(yy)*srcw + long(xx));

	    if (!isnanf(value) && (value>max)) {
	      *dest = max = value;
	      *mkdest=1;
	    }
	  }
	}
      }
    }
    CLEARSIGBUS
  }

  return fillImageColor(width, height);
}

unsigned char* Frame3d::fillImageThreshold(int width, int height, int depth,
					   double* (FitsImage::*getMatrix)())
{
  if (!*zbuf) {
    *zbuf = new float[width*height];
    memset(*zbuf, 0, width*height*sizeof(float));

    *mkzbuf = new unsigned char[width*height];
    memset(*mkzbuf, 0, width*height);

    FitsBound* params = context.fits->getDataParams(context.frScale.scanMode());
    int srcw = context.fits->width();
    int srch = context.fits->height();
    int srcd = context.fits->depth();

    double* mm = (context.fits->*getMatrix)();

    float* dest = *zbuf;
    unsigned char* mkdest = *mkzbuf;

    SETSIGBUS
    for (int jj=0; jj<height; jj++) {
      for (int ii=0; ii<width; ii++, dest++, mkdest++) {

	for (int kk=0; kk<depth; kk++) {
	  double xx = ii*mm[0] + jj*mm[4] + kk*mm[8]  + mm[12];
	  double yy = ii*mm[1] + jj*mm[5] + kk*mm[9]  + mm[13];
	  double zz = ii*mm[2] + jj*mm[6] + kk*mm[10] + mm[14];

	  if (xx>=params->xmin && xx<params->xmax && 
	      yy>=params->ymin && yy<params->ymax &&
	      zz>=0 && zz<srcd) {

	    float value = context.fits->getValueDouble(long(zz)*srcw*srch + 
					       long(yy)*srcw + long(xx));

	    if (!isnanf(value) && (value>=threshold_)) {
	      *dest =value;
	      *mkdest=1;
	    }
	  }
	}
      }
    }
    CLEARSIGBUS
  }

  return fillImageColor(width, height);
}

Vector3d Frame3d::imageSize()
{
  // return in IMAGE coords and extends edge to edge
  Vector3d rr(100,100,100);
  FitsImage* ptr = context.fits;
  if (ptr) {
    FitsBound* params = ptr->getDataParams(context.frScale.scanMode());
    rr =  Vector3d(params->xmax-params->xmin, params->ymax-params->ymin,
		   ptr->depth());
  }
  return rr;
}

void Frame3d::reset()
{
  cmapID = 1;
  bias = 0.5;
  contrast = 1.0;

  Frame3dBase::reset();
}

void Frame3d::updateBin(const Matrix& mx)
{
  context.updateBin(this,mx);

  updateColorScale();
  updateNow(MATRIX); // because we have changed zoom
}

void Frame3d::updateColorCells(unsigned short* index, 
			     unsigned char* cells, int cnt)
{
  // the colorbar widget will pass us a pointer to the indexCells

  colorCount = cnt;
  if (indexCells)
    delete [] indexCells;
  indexCells = new unsigned short[cnt];
  if (!indexCells) {
    internalError("Unable to Alloc indexCells");
    return;
  }
  memcpy(indexCells, index, cnt*sizeof(unsigned short));

  // copy the rgb vales to the colorCells array (for postscript printing)

  if (colorCells)
    delete [] colorCells;
  colorCells = new unsigned char[cnt*3];
  if (!colorCells) {
    internalError("Unable to Alloc colorCells");
    return;
  }
  memcpy(colorCells, cells, cnt*3);
}

void Frame3d::updateMagnifierMatrices(Matrix3d& refToMagnifier3d)
{
  FitsImage* ptr = context.fits;
  while (ptr) {
    FitsImage* sptr = ptr;
    while (sptr) {
      sptr->updateMagnifierMatrices(refToMagnifier3d);
      sptr = sptr->nextSlice();
    }
    ptr = ptr->nextMosaic();
  }
}

void Frame3d::updateMatrices()
{
  // very important! do this first
  Frame3dBase::updateMatrices();

  // alway identity
  FitsImage* ptr = context.fits;
  while (ptr) {
    FitsImage* sptr = ptr;
    while (sptr) {
      sptr->updateMatrices(threedToRef3d, refToUser3d, 
			   userToWidget3d, userToPanner3d);
      sptr = sptr->nextSlice();
    }
    ptr = ptr->nextMosaic();
  }
}

void Frame3d::unloadAllFits()
{
  unloadFits();
}

void Frame3d::unloadFits()
{
  if (DebugPerf)
    cerr << "Frame3d::unloadFits" << endl;

  context.unload();

  Base::unloadFits();
}

void Frame3d::updatePS(float scale, int width, int height)
{
  Matrix3d mx = psMatrix(scale, width, height);

  FitsImage* ptr = context.cfits;
  while (ptr) {
    ptr->updatePS(mx);
    ptr = ptr->nextMosaic();
  }
}

// Commands

void Frame3d::getColorbarCmd()
{
  ostringstream str;
  str << cmapID << ' ' << bias << ' ' << contrast << ' ' << invert << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Frame3d::getRGBChannelCmd()
{
  Tcl_AppendResult(interp, "red", NULL);
}

void Frame3d::getRGBViewCmd()
{
  Tcl_AppendResult(interp, "1 1 1", NULL);
}

void Frame3d::getRGBSystemCmd()
{
  Tcl_AppendResult(interp, "image", NULL);
}

void Frame3d::getTypeCmd()
{
  Tcl_AppendResult(interp, "3d", NULL);
}

void Frame3d::getWCSAlignCmd()
{
  ostringstream str;
  str << (unsigned short*)context.fits << ends;

  Tcl_AppendResult(interp, (wcsAlignment ? "1" : "0"), " ", 
		   coordSystemStr_[wcsSystem-IMAGE], " ",
		   skyFrameStr_[wcsSky-FK4], " ",
		   str.str().c_str(), NULL);
}

void Frame3d::sliceCmd(int ii, int ss)
{
  context.updateSlice(ii, ss);
  update(BASE);
}
