// Copyright (C) 1999-2010
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "framebase.h"
#include "context.h"
#include "fitsimage.h"
#include "NaN.h"
#include "blt.h"
#include "bltVector.h"
#include "projection.h"
#include "circle3d.h"

#include "sigbus.h"

void FrameBase::bltProjection(char* xname, char* yname, 
			      char* xcname, char* ycname,
			      Projection* pp)
{
  double* x;
  double* y;
  double* xc;
  double* yc;

  int num = bltProjFits(&x, &y, &xc, &yc, pp);

  Blt_Vector* xx;
  if (Blt_GetVector(interp, xname, &xx) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xx, x, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* yy;
  if (Blt_GetVector(interp, yname, &yy) != TCL_OK) 
    goto error;

  if (Blt_ResetVector(yy, y, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* xxc;
  if (Blt_GetVector(interp, xcname, &xxc) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xxc, xc, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* yyc;
  if (Blt_GetVector(interp, ycname, &yyc) != TCL_OK) 
    goto error;

  if (Blt_ResetVector(yyc, yc, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  return;

 error:
  result = TCL_ERROR;
  return;
}

int FrameBase::bltProjFits(double** x, double** y, 
			   double** xc, double** yc, 
			   Projection* pp)
{
  // does not extend across mosaic boundries
  // uses currentContext, currentSlice

  Vector cc = pp->getCenter();
  Vector p1 = pp->getP1();
  Vector p2 = pp->getP2(); 
  int width = pp->getWidth();

  FitsImage* ptr = currentContext->cfits;
  while (ptr) {
    Vector img = cc * ptr->getRefToData();
    FitsBound* params = ptr->getDataParams(currentContext->frScale.scanMode());

    if (img[0]>=params->xmin && img[0]<params->xmax && 
	img[1]>=params->ymin && img[1]<params->ymax)
      break;

    ptr = ptr->nextMosaic();
  }
  if (!ptr)
    ptr = currentContext->cfits;

  FitsBound* params = ptr->getDataParams(currentContext->frScale.scanMode());

  Vector vv = p2-p1; 
  int num = vv.length() +1;
  Vector s = vv.normalize();
  int cnt[num];

  *x = (double*)malloc(num*sizeof(double));
  *y = (double*)malloc(num*sizeof(double));
  *xc = (double*)malloc(num*sizeof(double));
  *yc = (double*)malloc(num*sizeof(double));

  // main loop

  SETSIGBUS
  for (long ii=0; ii<num; ii++) {
    Vector t = p1 + s*ii;

    (*x)[ii] = ii+1;
    (*y)[ii] = 0;
    cnt[ii] = 0;

    Vector tv = ptr->mapFromRef(t, pp->getSystem(), pp->getSkyFrame());
    (*xc)[ii] = tv[0];
    (*yc)[ii] = tv[1];

    Vector z = t * ptr->getRefToData();

    if (z[0]>=params->xmin && z[0]<params->xmax && 
	z[1]>=params->ymin && z[1]<params->ymax) {

      // check for nan
      double v = ptr->getValueDouble(z);
      if (!isnand(v)) {
	(*y)[ii] = v;
	cnt[ii] = 1;
      }

      Vector ss = Vector(-s[1],s[0]);

      for (long jj=1; jj<width; jj++) {
	Vector tt = p1 + s*ii + ss*jj;

	Vector zz = tt * ptr->getRefToData();
		
	if (zz[0]>=params->xmin && zz[0]<params->xmax && 
	    zz[1]>=params->ymin && zz[1]<params->ymax) {

	  double vvalue = ptr->getValueDouble(zz);
	  // check for nan
	  if (!isnand(vvalue)) {
	    (*y)[ii] += vvalue;
	    cnt[ii]++;
	  }
	}
      }
    }
  }

  // average if needed
  if (pp->getMethod())
    for (long ii=0; ii<num; ii++)
      if (!isnand((*y)[ii]) && cnt[ii]!=0)
	(*y)[ii] /= cnt[ii];

  CLEARSIGBUS

  return num;
}

void FrameBase::bltCircle3d(char* xname, char* yname, 
				char* xcname, Circle3d* pp)
{
  double* x;
  double* y;
  double* xc;

  int num = bltCircle3dFits(&x, &y, &xc, pp);

  Blt_Vector* xx;
  if (Blt_GetVector(interp, xname, &xx) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xx, x, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* yy;
  if (Blt_GetVector(interp, yname, &yy) != TCL_OK) 
    goto error;

  if (Blt_ResetVector(yy, y, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* xxc;
  if (Blt_GetVector(interp, xcname, &xxc) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xxc, xc, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  return;

 error:
  result = TCL_ERROR;
  return;
}

int FrameBase::bltCircle3dFits(double** x, double** y, 
			     double** xc, Circle3d* pp)
{
  // does not extend across mosaic boundries
  // uses currentContext, firstSlice

  Vector cc = pp->getCenter();

  FitsImage* ptr = currentContext->fits;
  while (ptr) {
    Vector img = cc * ptr->getRefToData();
    FitsBound* params = ptr->getDataParams(currentContext->frScale.scanMode());

    if (img[0]>=params->xmin && img[0]<params->xmax && 
	img[1]>=params->ymin && img[1]<params->ymax)
      break;

    ptr = ptr->nextMosaic();
  }
  if (!ptr)
    ptr = currentContext->fits;

  int num = currentContext->calcSlice();
  for (int ii=1; ii<num; ii++)
    if (ptr)
      ptr = ptr->nextSlice();

  FitsBound* params = ptr->getDataParams(currentContext->frScale.scanMode());
  long srcw = ptr->width();
  long srch = ptr->height();
  long srcd = ptr->depth();

  int cnt[srcd];
  *x = (double*)malloc(srcd*sizeof(double));
  *y = (double*)malloc(srcd*sizeof(double));
  *xc = (double*)malloc(srcd*sizeof(double));

  Vector cd = cc * ptr->getRefToData();
  float rr = (pp->annuli(0))[0];
  float rr2 = rr*rr;

  // main loop

  SETSIGBUS
  for (long kk=0; kk<srcd; kk++) {
    (*x)[kk] = kk+1;
    (*y)[kk] = 0;
    cnt[kk] = 0;

    (*xc)[kk] = ptr->mapFromImagex(kk+1, pp->getSystem(), 2);

    for (long jj=-rr; jj<=rr; jj++) {
      for (long ii=-rr; ii<=rr; ii++) {
	Vector aa = cd + Vector(ii,jj);

	if (aa[0]>=params->xmin && aa[0]<params->xmax && 
	    aa[1]>=params->ymin && aa[1]<params->ymax &&
	    kk>=0 && kk<srcd) {

	  if (ii*ii+jj*jj <= rr2) {
	    double val = 
	      ptr->getValueDouble(kk*srcw*srch + int(aa[1])*srcw + int(aa[0]));

	    // check for nan
	    if (!isnand(val)) {
	      (*y)[kk] += val;
	      cnt[kk]++;
	    }
	  }
	}
      }
    }
  }

  // average if needed
  if (pp->getMethod())
    for (long kk=0; kk<srcd; kk++)
      if (cnt[kk]!=0)
	(*y)[kk] /= cnt[kk];

  CLEARSIGBUS

  return srcd;
}

void FrameBase::bltCut(char* xname, char* yname, 
		       Orientation axis, const Vector& rr)
{
  // Vector r is in Widget coords
  Vector r = rr;

  int size;
  if (axis == XX)
    size = options->width;
  else
    size = options->height;

  long length = (size+1) * 2;
  double* x = (double*)malloc(sizeof(double)*length);
  double* y = (double*)malloc(sizeof(double)*length);

  if (!currentContext->cfits) {
    for (int i=0; i<=size; i++) {
      x[i*2] = i;
      x[i*2+1] = i;
      y[i*2] = 0;
      y[i*2+1] = 0;
    }
  }
  else
    bltCutFits(x, y, size, r, axis);

  Blt_Vector* xx;
  if (Blt_GetVector(interp, xname, &xx) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xx, x, length, length*sizeof(double), TCL_DYNAMIC) != 
      TCL_OK)
    goto error;

  Blt_Vector* yy;
  if (Blt_GetVector(interp, yname, &yy) != TCL_OK)
    goto error;

  if (Blt_ResetVector(yy, y, length, length*sizeof(double), TCL_DYNAMIC) != 
      TCL_OK)
    goto error;

  return;

 error:
    result = TCL_ERROR;
    return;
}

void FrameBase::bltCutFits(double* x, double* y, int size, Vector& r, 
			   Orientation axis)
{
  // basics
  FitsImage* sptr = currentContext->cfits;
  int mosaic = isMosaic();

  // variable
  FitsBound* params = sptr->getDataParams(currentContext->frScale.scanMode());
  double prev = currentContext->frScale.low();

  // main loop

  SETSIGBUS
  for (int ii=0; ii<=size; ii++) {
    double v = currentContext->frScale.low();

    if (mosaic) {
      sptr = currentContext->cfits;
      params = sptr->getDataParams(currentContext->frScale.scanMode());
    }

    do {
      Vector img;
      if (axis == XX)
	img = Vector(1+ii,r[1]) * sptr->getWidgetToData();
      else
	img = Vector(r[0],1+ii) * sptr->getWidgetToData();

      if (img[0]>=params->xmin && img[0]<params->xmax && 
	  img[1]>=params->ymin && img[1]<params->ymax) {
	double value = sptr->getValueDouble(img);

	if (!isnand(value))
	  v = value;

	break;
      }
      else {
	if (mosaic) {
	  sptr = sptr->nextMosaic();
	  if (sptr)
	    params = sptr->getDataParams(currentContext->frScale.scanMode());
	}
      }
    }
    while (mosaic && sptr);

    x[2*ii] = ii;
    x[2*ii +1] = ii;
    
    y[2*ii] = prev;
    y[2*ii +1] = v;
    prev = v;
  }
  CLEARSIGBUS
}

