// Copyright (C) 1999-2010
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include <fstream>
#include "fdstream.hpp"

#include "tcl.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "framebase.h"
#include "context.h"
#include "fitsimage.h"

#include "sigbus.h"

void FrameBase::contourAuxHeadCmd()
{
  currentContext->contourAuxHead();
}

void FrameBase::contourAuxNextCmd()
{
  currentContext->contourAuxNext();
}

void FrameBase::contourAuxSaveCmd(const char* fn, CoordSystem sys, SkyFrame sky)
{
  if (currentContext->cfits && hasContourAux()) {
    ofstream str(fn);

    if (str) {
      List<Vertex>& c = (List<Vertex>&)(currentContext->auxcontours.current()->contours());

      if (c.head())
	do {
	  Vector v = c.current()->vector;
	  if (v[0] != DBL_MAX)
	    str << setiosflags(ios::scientific) 
		<< setprecision(8) 
		<< currentContext->cfits->mapFromRef(v, sys, sky);
	  str << endl;
	}
	while (c.next());
    }
  }
}

void FrameBase::contourCopyCmd(CoordSystem sys, SkyFrame sky)
{
  if (currentContext->cfits && hasContour()) {
    List<Vertex>* v = new List<Vertex>(currentContext->contour->contours());
    if (v->head())
      do {
	Vector& w = v->current()->vector;
	if (w[0] != DBL_MAX)
	  w = currentContext->cfits->mapFromRef(w, sys, sky);
      } while (v->next());

    ostringstream str;
    str << hex << v << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }
}

void FrameBase::contourCreateCmd(const char* color, int width, int dash,
				 FVContour::Method method, int numlevel, 
				 int smooth, 
				 FrScale::ColorScaleType colorScaleType,
				 float expo,
				 float clipMode, Vector limits, 
				 const char* level)
{
  if (DebugPerf)
    cerr << "contourCreate" << endl;

  // make sure we have up-to-date matrices
  updateMatrices();

  currentContext->clearContour();

  InverseScale* scale =NULL;
  if (level && strlen(level)>0) {
    int cnt = 0;
    double levels[100];
    string x(level);
    istringstream str(x);

    while ((cnt<100) && (str >> levels[cnt]))
      cnt++;

    scale = new InverseScale(cnt, levels);
  }
  else {
    switch (colorScaleType) {
    case FrScale::IISSCALE:
    case FrScale::LINEARSCALE:
      scale = new LinearInverseScale(numlevel, limits[0], limits[1]);
      break;
    case FrScale::LOGSCALE:
      scale =  new LogInverseScale(numlevel, limits[0], limits[1], expo);
      break;
    case FrScale::POWSCALE:
      scale =  new PowInverseScale(numlevel, limits[0], limits[1], expo);
      break;
    case FrScale::SQRTSCALE:
      scale =  new SqrtInverseScale(numlevel, limits[0], limits[1]);
      break;
    case FrScale::SQUAREDSCALE:
      scale =  new SquaredInverseScale(numlevel, limits[0], limits[1]);
      break;
    case FrScale::HISTEQUSCALE:
      scale =  new HistEquInverseScale(numlevel, limits[0], limits[1], 
				       currentContext->histequ(), HISTEQUSIZE);
      break;
    }
  }

  if (!isMosaic()) {
    if (currentContext->cfits)
      currentContext->setContour(new FVContour(this, currentContext->cfits, color, width, dash, method, numlevel, smooth, level, colorScaleType, expo, clipMode, limits, scale));
  }
  else {
    if (currentContext->fits) {
      currentContext->setContour(new FVContour(this, currentContext->fits, color, width, dash, method, numlevel, smooth, level, colorScaleType, expo, clipMode, limits, scale));

      FitsImage* ptr = currentContext->fits->nextMosaic();
      while (ptr) {
	currentContext->contour->append(ptr);
	ptr = ptr->nextMosaic();
      }
    }
  }

  update(BASE);
}

void FrameBase::contourDeleteCmd()
{
  currentContext->clearContour();

  update(BASE);
}

void FrameBase::contourDeleteAllCmd()
{
  currentContext->clearContour();
  currentContext->auxcontours.deleteAll();

  update(BASE);
}

void FrameBase::contourLoadCmd(const char* color, int w, int d, const char* fn,
			       CoordSystem sys, SkyFrame sky)
{
  if (!currentContext->cfits)
    return;

  ifstream str(fn);

  if (str) {
    List<Vertex> c;
 
    while (!str.eof()) {
      Vertex* vv;

      char buf[64];
      str.getline(buf,64,'\n');
      if (strlen(buf) > 0) {
	Vector v;
	string x(buf);
	istringstream sstr(x);

	sstr >> v[0] >> v[1];
	vv = new Vertex(currentContext->fits->mapToRef(v, sys, sky));
      }
      else
	vv = new Vertex(Vector(DBL_MAX, DBL_MAX));

      c.append(vv);
    }

    currentContext->auxcontours.append(new Contour(this, color, w, d, c));
  }

  update(BASE);
}

void FrameBase::contourPasteCmd(const char* c, int w, int d, void* u, 
			    CoordSystem sys, SkyFrame sky)
{
  if (!currentContext->cfits)
    return;

  List<Vertex>* v = (List<Vertex>*)u;

  if (v->head())
    do {
      Vector& w = v->current()->vector;
      if (w[0] != DBL_MAX)
	w = currentContext->fits->mapToRef(w, sys, sky);
    } while (v->next());
  
  currentContext->auxcontours.append(new Contour(this, c, w, d, *v));
  delete v;

  update(BASE);
}

void FrameBase::contourSaveCmd(const char* fn, CoordSystem sys, SkyFrame sky)
{
  if (currentContext->cfits && hasContour()) {
    ofstream str(fn);

    if (str) {
      List<Vertex>& c = (List<Vertex>&)(currentContext->contour->contours());

      if (c.head())
	do {
	  Vector v = (c.current())->vector;
	  if (v[0] != DBL_MAX && v[1] != DBL_MAX)
	    str << setiosflags(ios::scientific) 
		<< setprecision(8) 
		<< currentContext->cfits->mapFromRef(v, sys, sky) 
		<< endl;
	  else
	    str << endl;
	}
	while (c.next());
    }
  }
}

void FrameBase::contourSetColorCmd(const char* clr)
{
  if (hasContour()) {
    currentContext->contour->setColor(clr);
    update(BASE);
  }
}

void FrameBase::contourSetDashCmd(int d)
{
  if (hasContour()) {
    currentContext->contour->setDash(d);
    update(BASE);
  }
}

void FrameBase::contourSetLineWidthCmd(int w)
{
  if (hasContour()) {
    currentContext->contour->setLineWidth(w);
    update(BASE);
  }
}

void FrameBase::crosshairCmd(int which)
{
  useCrosshair = which ? 1 : 0;
  update(PIXMAP);
}

void FrameBase::crosshairCmd(const Vector& v, InternalSystem sys)
{
  useCrosshair = 1;
  crosshair = mapToRef(v, sys);
  update(PIXMAP);
}

void FrameBase::crosshairCmd(const Vector& v, CoordSystem sys, SkyFrame sky)
{
  useCrosshair = 1;
  if (currentContext->cfits)
    crosshair = currentContext->cfits->mapToRef(v, sys, sky);
  update(PIXMAP);
}

void FrameBase::crosshairBeginCmd(const Vector& v, InternalSystem sys)
{
  // This routine is optimalized for rendering an interactive cursor. First,
  // a cursorless pixmap is generated. Then as the cursor is moved about, the 
  // cursor is written directly to the display, and erase by XCopyArea from the
  // widget pixmap.

  // turn cursor off and generate pixmap
  useCrosshair = 0;
  updateNow(PIXMAP);

  // turn cursor on, and draw directly to display
  useCrosshair = 1;
  crosshair = mapToRef(v, sys);
  drawCrosshair();
}

void FrameBase::crosshairBeginCmd(const Vector& v, CoordSystem sys)
{
  // This routine is optimalized for rendering an interactive cursor. First,
  // a cursorless pixmap is generated. Then as the cursor is moved about, the 
  // cursor is written directly to the display, and erase by XCopyArea from the
  // widget pixmap.

  // turn cursor off and generate pixmap
  useCrosshair = 0;
  updateNow(PIXMAP);

  // turn cursor on, and draw directly to display
  useCrosshair = 1;
  if (currentContext->cfits)
    crosshair = currentContext->cfits->mapToRef(v, sys, FK5);
  drawCrosshair();
}

void FrameBase::crosshairMotionCmd(const Vector& v, InternalSystem sys)
{
  eraseCrosshair();
  crosshair = mapToRef(v, sys);
  drawCrosshair();
}

void FrameBase::crosshairMotionCmd(const Vector& v, CoordSystem sys)
{
  eraseCrosshair();
  if (currentContext->cfits)
    crosshair = currentContext->cfits->mapToRef(v, sys, FK5);
  drawCrosshair();
}

void FrameBase::crosshairWarpCmd(const Vector& v)
{
  useCrosshair = 1;

  Vector r = crosshair * refToCanvas;
  r += v;
  crosshair = r * canvasToRef;

  update(PIXMAP);
  updateMagnifier();
}

void FrameBase::getColorMapLevelCmd(int count, const Vector& v,
				      InternalSystem ref)
{
  if (currentContext->cfits)
    switch (ref) {
    case CANVAS:
      getColorMapLevel(count, v, &FitsImage::getCanvasToData);
      break;
    case PANNER:
      getColorMapLevel(count, v, &FitsImage::getPannerToData);
      break;
    }
  else
    Base::getColorMapLevelCmd(count, currentContext->frScale.low(), 
			      currentContext->frScale.high(),
			      currentContext->frScale.colorScaleType(),
			      currentContext->frScale.expo());
}

void FrameBase::getCoordCmd(const Vector& v, InternalSystem in, 
			CoordSystem out, SkyFrame sky, SkyFormat format)
{
  switch (in) {
  case CANVAS:
    if (FitsImage* ptr=isInFits(v, &FitsImage::getCanvasToData, NULL))
      printFromRef(ptr, v * canvasToRef, out, sky, format, DEFAULT);
    break;
  case PANNER:
    if (FitsImage* ptr=isInFits(v, &FitsImage::getPannerToData, NULL))
      printFromRef(ptr, v * pannerToRef, out, sky, format, DEFAULT);
    break;
  }
}

void FrameBase::getCoordFromRefCmd(double v, CoordSystem out, int ss)
{
  if (currentContext->cfits)
    printDouble(currentContext->cfits->mapFromImagex(v,out,ss), DEFAULT);
  else
    printDouble(0,DEFAULT);
}

void FrameBase::getCoordToRefCmd(double v, CoordSystem out, int ss)
{
  if (currentContext->cfits)
    printDouble(currentContext->cfits->mapToImagex(v,out,ss), DEFAULT);
  else
    printDouble(0,DEFAULT);
}

void FrameBase::getCrosshairCmd(InternalSystem sys)
{
  printVector(mapFromRef(crosshair, sys), DEFAULT);
}

void FrameBase::getCrosshairCmd(CoordSystem sys, SkyFrame sky, SkyFormat format,
				Precision p)
{
  if (currentContext->cfits)
    printFromRef(currentContext->cfits, crosshair, sys, sky, format, p);
  else
    printVector(Vector(), DEFAULT);
}

void FrameBase::getCrosshairStatusCmd()
{
  if (useCrosshair)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void FrameBase::getCursorCmd(InternalSystem sys)
{
  // make sure we have up-to-date matrices
  updateNow(MATRIX);

  if (currentContext->cfits)
    switch (sys) {
    case CANVAS:
      printVector(cursor * userToCanvas, DEFAULT);
      break;
    case PANNER:
      break;
      printVector(cursor * userToPanner, DEFAULT);
    }
  else
    printVector(Vector(), DEFAULT);
} 

void FrameBase::getCursorCmd(CoordSystem sys, SkyFrame sky, SkyFormat format,
			 Precision p)
{
  // make sure we have up-to-date matrices
  updateNow(MATRIX);
  
  if (currentContext->cfits)
    printFromRef(currentContext->cfits, cursor * userToRef, sys, sky, format, p);
  else
    printVector(Vector(), DEFAULT);
} 

void FrameBase::getDataValuesCmd(int which, const Vector& v, 
			     CoordSystem sys, SkyFrame sky,
			     const Vector& ddd, char* var)
{
  // clear an previous values
  Tcl_UnsetVar(interp,var,0);

  // find anchor point
  FitsImage* base = currentContext->cfits;
  for (int i=1; i<which; i++)
    if (base)
      base = base->nextMosaic();

  if (!base) {
    Tcl_SetVar(interp,var,"",0);
    result = TCL_ERROR;
    return;
  }
  Vector d = base->mapLenToRef(ddd,sys,DEGREES);
  Vector ll = base->mapToRef(v,sys,sky);

  SETSIGBUS
  for (int i=0; i<d[0]; i++) {
    for (int j=0; j<d[1]; j++) {
      Vector rr = ll+Vector(i,j);

      // index is in terms of native coords
      Vector in = base->mapFromRef(rr,sys,sky);

      ostringstream str;
      str << setprecision(12) << in[0] << ',' << in[1] << ends;

      int found = 0;
      FitsImage* ptr = currentContext->fits;
      while (ptr) {
	Vector dd = rr * ptr->getRefToData();
	FitsBound* params = 
	  ptr->getDataParams(currentContext->frScale.scanMode());

	if (dd[0]>=params->xmin && dd[0]<params->xmax && 
	    dd[1]>=params->ymin && dd[1]<params->ymax) {
	  Tcl_SetVar2(interp,var,str.str().c_str(),(char*)ptr->getValue(dd),0);
	  found = 1;
	  break;
	}
	ptr = ptr->nextMosaic();
      }

      if (!found)
	Tcl_SetVar2(interp,var,str.str().c_str(),"",0);
    }
  }
  CLEARSIGBUS
}

void FrameBase::getDataValuesCmd(const Vector& v, InternalSystem sys,
				 const Vector& b)
{
  Vector d = b;
  FitsImage* ptr = NULL;
  Vector ll;

  // find anchor point
  switch (sys) {
  case CANVAS:
    if (ptr=isInFits(v, &FitsImage::getCanvasToData, NULL))
      ll = v*canvasToRef - Vector((d[0]-1)/2,(d[1]-1)/2);
    break;
  case PANNER:
    if (ptr=isInFits(v, &FitsImage::getPannerToData, NULL))
      ll = v*pannerToRef - Vector((d[0]-1)/2,(d[1]-1)/2);
    break;
  }
  if (!ptr)
    return;

  SETSIGBUS
  for (int j=0; j<d[1]; j++) {
    for (int i=0; i<d[0]; i++) {
      Vector dd = (ll+Vector(i,j)) * ptr->getRefToData();
      FitsBound* params = 
	ptr->getDataParams(currentContext->frScale.scanMode());

      if (dd[0]>=params->xmin && dd[0]<params->xmax && 
	  dd[1]>=params->ymin && dd[1]<params->ymax)
	Tcl_AppendResult(interp, (char*)ptr->getValue(dd), " ", NULL);
    }
  }
  CLEARSIGBUS
}

void FrameBase::getFitsExtCmd(const Vector& v, InternalSystem ref)
{
  switch (ref) {
  case CANVAS:
    getFitsExt(v, ref, &FitsImage::getCanvasToData);
    break;
  case PANNER:
    getFitsExt(v, ref, &FitsImage::getPannerToData);
    break;
  }
}

void FrameBase::getFitsFileNameCmd(const Vector& v, InternalSystem sys,
			       FileNameType type)
{
  switch (sys) {
  case CANVAS:
    if (FitsImage* ptr = isInFits(v, &FitsImage::getCanvasToData, NULL)) {
      switch (type) {
      case ROOTBASE:
	Tcl_AppendResult(interp, ptr->getRootBaseFileName(), NULL);
	break;
      case FULLBASE:
	Tcl_AppendResult(interp, ptr->getFullBaseFileName(), NULL);
	break;
      case ROOT:
	Tcl_AppendResult(interp, ptr->getRootFileName(), NULL);
	break;
      case FULL:
	Tcl_AppendResult(interp, ptr->getFullFileName(), NULL);
	break;
      }
    }
    break;
  case PANNER:
    if (FitsImage* ptr = isInFits(v, &FitsImage::getPannerToData, NULL)) {
      switch (type) {
      case ROOTBASE:
	Tcl_AppendResult(interp, ptr->getRootBaseFileName(), NULL);
	break;
      case FULLBASE:
	Tcl_AppendResult(interp, ptr->getFullBaseFileName(), NULL);
	break;
      case ROOT:
	Tcl_AppendResult(interp, ptr->getRootFileName(), NULL);
	break;
      case FULL:
	Tcl_AppendResult(interp, ptr->getFullFileName(), NULL);
	break;
      }
    }
    break;
  }
}

void FrameBase::getFitsCenterCmd(CoordSystem sys, SkyFrame sky, 
				 SkyFormat format, Precision p)
{
  // make sure we have up-to-date matrices
  updateNow(MATRIX);
  
  if (keyContext && keyContext->fits)
    printFromRef(keyContext->fits, imageCenter() * keyContext->fits->getImageToRef(), sys, sky, format, p);
  else
    printVector(Vector(), DEFAULT);
}

void FrameBase::getFitsSizeCmd(CoordSystem sys, SkyFrame sky, 
			       SkyFormat format, Precision pp)
{
  // make sure we have up-to-date matrices
  updateNow(MATRIX);
  
  if (keyContext->fits) {
    if (!keyContext->fits->hasWCSEqu(sys)) {
      printVector(Vector(), DEFAULT);
      return;
    }

    BBox bb = imageBBox();
    Matrix mm = keyContext->fits->getImageToRef();

    Vector ll = bb.ll   * mm;
    Vector lr = bb.lr() * mm;
    Vector ur = bb.ur   * mm;
    Vector ul = bb.ul() * mm;

    Vector ww[4];
    ww[0] = keyContext->fits->mapFromRef(ll,sys,sky);
    ww[1] = keyContext->fits->mapFromRef(lr,sys,sky);
    ww[2] = keyContext->fits->mapFromRef(ur,sys,sky);
    ww[3] = keyContext->fits->mapFromRef(ul,sys,sky);

    // we need to check for the case of crossing 0 in ra
    // since ra is returned as 0 > ra > 360
    {
      float min = 360;
      float max = 0;
      for (int ii=0; ii<4; ii++) {
	if (ww[ii][0]<min)
	  min=ww[ii][0];
	if (ww[ii][0]>max)
	  max=ww[ii][0];
      }

      // ok, we have a problem
      if (max-min > 180)
	for (int ii=0; ii<4; ii++)
	  if (ww[ii][0] > 180)
	    ww[ii][0] -= 360;
    }

    BBox wbb(ww[0],ww[0]);
    for (int ii=1; ii<4; ii++)
      wbb.bound(ww[ii]);

    Vector ss(keyContext->fits->wcsdist(wbb.ll,wbb.lr(),sys),
	      keyContext->fits->wcsdist(wbb.ll,wbb.ul(),sys));

    switch (format) {
    case DEGREES:
    case SEXAGESIMAL:
      break;
    case ARCMIN:
      ss *= 60;
      break;
    case ARCSEC:
      ss *= 60*60;
      break;
    }

    printVector(ss,pp);
  }
  else
    printVector(Vector(), DEFAULT);
}

void FrameBase::updateFitsCmd(int which, BBox bb, int now)
{
  // Note: bb is in IMAGE coords
  FitsImage* ptr = currentContext->fits;
  if (which > 0) {
    for (int i=0; i<(which-1); i++) {
      if (ptr)
	ptr = ptr->nextMosaic();
    }
  }

  if (ptr) {
    Vector ll = bb.ll * ptr->getImageToCanvas();
    Vector lr = bb.lr() * ptr->getImageToCanvas();
    Vector ur = bb.ur * ptr->getImageToCanvas();
    Vector ul = bb.ul() * ptr->getImageToCanvas();
    BBox rr(ll);
    rr.bound(lr);
    rr.bound(ur);
    rr.bound(ul);

    if (now)
      updateNow(BASE, rr);
    else
      update(BASE, rr);
  }
}

void FrameBase::getGridCmd()
{
  if (grid) {
    Tcl_AppendElement(interp, coordSystemStr_[(grid->system())-IMAGE]);

    Tcl_AppendElement(interp, skyFrameStr_[(grid->sky())-FK4]);

    Tcl_AppendElement(interp, skyFormatStr_[grid->skyFormat()]);

    switch (grid->type()) {
    case Grid::ANALYSIS:
      Tcl_AppendElement(interp, "analysis");
      break;
    case Grid::PUBLICATION:
      Tcl_AppendElement(interp, "publication");
      break;
    }
  }
  else
    Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getGridOptionCmd()
{
  if (grid)
    Tcl_AppendResult(interp, grid->option(), NULL);
  else
    Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getHorzCutCmd(char* x, char* y, const Vector& v, 
			      InternalSystem in)
{
  switch (in) {
  case CANVAS:
    bltCut(x, y, XX, v * canvasToWidget);
    break;
  case PANNER:
    bltCut(x, y, XX, v * pannerToWidget);
    break;
  }
}

void FrameBase::getPixelTableCmd(const Vector& v, InternalSystem sys,
			     int size, char* var)
{
  switch (sys) {
  case CANVAS:
    getPixelTable(v, &FitsImage::getCanvasToData, size, var);
    break;
  case PANNER:
    getPixelTable(v, &FitsImage::getPannerToData, size, var);
    break;
  }
}

void FrameBase::getValueCmd(const Vector& v, InternalSystem sys)
{
  Vector r;

  SETSIGBUS
  switch (sys) {
  case CANVAS:
    if (FitsImage* ptr=isInFits(v, &FitsImage::getCanvasToData, &r))
      Tcl_AppendResult(interp, ptr->getValue(r), NULL);
    break;
  case PANNER:
    if (FitsImage* ptr=isInFits(v, &FitsImage::getPannerToData, &r))
      Tcl_AppendResult(interp, ptr->getValue(r), NULL);
    break;
  }
  CLEARSIGBUS
}

void FrameBase::getVertCutCmd(char* x, char* y, const Vector& v,
			      InternalSystem in)
{
  switch (in) {
  case CANVAS:
    bltCut(x, y, YY, v * canvasToWidget);
    break;
  case PANNER:
    bltCut(x, y, YY, v * pannerToWidget);
    break;
  }
}

void FrameBase::get3dRenderMethodCmd()
{
  Tcl_AppendResult(interp, "mip", NULL);
}

void FrameBase::get3dThresholdCmd()
{
  Tcl_AppendResult(interp, "50", NULL);
}

void FrameBase::get3dViewCmd()
{
  Tcl_AppendResult(interp, "0 0", NULL);
}

void FrameBase::gridCmd(CoordSystem sys, SkyFrame sky, SkyFormat format, 
			Grid::GridType type, const char* ops)
{
  if (grid)
    delete grid;
  grid = new Grid(this, sys, sky, format, type, ops);

  update(BASE);
}

void FrameBase::gridDeleteCmd()
{
  if (grid)
    delete grid;
  grid = NULL;

  update(BASE);
}

void FrameBase::iisCursorModeCmd(int state)
{
  if (state) {
    // save cursor pos
    Window root, child;
    int rootx,rooty,winx,winy;
    unsigned int msk;
    XQueryPointer(display,Tk_WindowId(tkwin),&root,&child,
		  &rootx,&rooty,&winx,&winy,&msk);

    if (!iisInteractive) {
      iisSaveRoot = root;
      iisSaveCursor = Vector(rootx,rooty);
    }
    iisInteractive++;

    // set iisLastCursor if first time thru
    if (iisLastCursor[0] == 0 && iisLastCursor[1] == 0)
      iisLastCursor = Vector(options->width,options->height)/2;

    // and move to last cursor position
    Vector r = iisLastCursor * widgetToWindow;
    XWarpPointer(display, None,Tk_WindowId(tkwin),0,0,0,0,(int)r[0],(int)r[1]);
  }
  else {
    // restore cursor pos
    if (!(--iisInteractive)) {
      XWarpPointer(display, None, iisSaveRoot, 0, 0, 0, 0, 
		   (int)iisSaveCursor[0], (int)iisSaveCursor[1]);
      iisSaveRoot = 0;
      iisSaveCursor = Vector();
    }
  }
}

void FrameBase::iisGetCursorCmd()
{
  printVector(iisLastCursor,DEFAULT);
}

void FrameBase::iisGetFileNameCmd()
{
  FitsImage* ptr = currentContext->fits;
  while (ptr && ptr->nextMosaic())
    ptr = ptr->nextMosaic();

  if (ptr)
    Tcl_AppendResult(interp, ptr->iisGetFileName(), NULL);
}

void FrameBase::iisGetFileNameCmd(int which)
{
  FitsImage* ptr = currentContext->fits;
  for (int i=0; i<(which-1); i++) {
    if (ptr)
      ptr = ptr->nextMosaic();
  }

  if (ptr)
    Tcl_AppendResult(interp, ptr->iisGetFileName(), NULL);
}

void FrameBase::iisGetFileNameCmd(const Vector& v)
{
  if (FitsImage* ptr = isInFits(v, &FitsImage::getCanvasToData, NULL))
    Tcl_AppendResult(interp, ptr->iisGetFileName(), NULL);
}

void FrameBase::iisMessageCmd(const char* ptr)
{
  if (!currentContext->cfits)
    return;

  // do we have something?
  if (!(ptr && *ptr))
    return;

  // filename
  // note: the file name is the second value passed
  char fn[PATH_MAX];
  string x(ptr);
  istringstream str(x);
  str >> fn >> fn;
  currentContext->cfits->setFileName(fn);

  // object name
  // first go to end
  char* sptr = (char*)ptr;
  while (*sptr)
    sptr++;
  // now work backwards to last '-'
  while (sptr != ptr) {
    if (*sptr == '-')
      break;
    sptr--;
  }
  // is it good?
  if (sptr != ptr) {
    // go forward 2
    sptr += 2;
    currentContext->cfits->setObjectName(sptr);
  }
}

void FrameBase::iisSetCursorCmd(const Vector& v, InternalSystem sys)
{
  // assume canvas
  if (!currentContext->cfits)
    return;

  iisLastCursor = v * canvasToWidget;

  Vector r = iisLastCursor * widgetToWindow;
  XWarpPointer(display, None, Tk_WindowId(tkwin), 0,0,0,0,(int)r[0],(int)r[1]);
}

void FrameBase::iisSetCursorCmd(const Vector& v, CoordSystem sys)
{
  if (!currentContext->cfits)
    return;

  iisLastCursor = currentContext->cfits->mapToRef(v,sys) * refToWidget;

  Vector r = iisLastCursor * widgetToWindow;
  XWarpPointer(display, None, Tk_WindowId(tkwin), 0,0,0,0,(int)r[0],(int)r[1]);
}

void FrameBase::iisSetFileNameCmd(const char* fn)
{
  FitsImage* ptr = currentContext->fits;
  while (ptr && ptr->nextMosaic())
    ptr = ptr->nextMosaic();

  if (ptr)
    ptr->iisSetFileName(fn);
}

void FrameBase::iisSetFileNameCmd(const char* fn, int which)
{
  FitsImage* ptr = currentContext->fits;
  for (int i=0; i<(which-1); i++) {
    if (ptr)
      ptr = ptr->nextMosaic();
  }

  if (ptr)
    ptr->iisSetFileName(fn);
}

void FrameBase::panCmd(const Vector& v1, const Vector& v2)
{
  if (!currentContext->cfits)
    return;

  Vector start = v1 * canvasToUser;
  Vector stop = v2 * canvasToUser;
  cursor -= stop - start;
  update(MATRIX);
}

void FrameBase::panCmd(const Vector& v, InternalSystem sys)
{
  if (!currentContext->cfits)
    return;

  switch (sys) {
  case CANVAS:
    {
      Vector u = cursor * userToCanvas;
      u += v;
      cursor = u * canvasToUser;
    }
    break;
  case PANNER: 
    {
      Vector u = cursor * userToPanner;
      u += v;
      cursor = u * pannerToUser;
    }
    break;
  }

  setBinCursor(cursor);
  update(MATRIX);
}

void FrameBase::panCmd(const Vector& v, CoordSystem sys, SkyFrame sky)
{
  if (!currentContext->cfits)
    return;

  Vector u = currentContext->cfits->mapFromRef(cursor * userToRef, sys, sky);
  u += v;
  cursor = currentContext->cfits->mapToRef(u, sys, sky) * refToUser;
  
  setBinCursor(cursor);
  update(MATRIX);
}

void FrameBase::panBBoxCmd(const Vector& v)
{
  if (!currentContext->cfits)
    return;

  // we want to round to nearest pixel center
  Vector r = v * currentContext->cfits->getPannerToData();
  cursor = (r.floor() + Vector(.5,.5)) * currentContext->cfits->getDataToUser();

  setBinCursor(cursor);
  update(MATRIX);
}

void FrameBase::panBeginCmd(const Vector& v)
{
  if (!currentContext->cfits)
    return;

  // save the current cursor
  panCursor = v;

  // copy tmp pixmap
  panPM = Tk_GetPixmap(display, Tk_WindowId(tkwin),
		       options->width, options->height, depth);
  if (!panPM) {
    internalError("Unable to Create Pan Motion Pixmap");
    return;
  }
  XCopyArea(display, pixmap, panPM, gc, 0, 0, options->width, 
	    options->height, 0,0);
}

void FrameBase::panMotionCmd(const Vector& v)
{
  if (!currentContext->cfits)
    return;

  if (visible) {

    // Clear
    Vector diff = ((v * canvasToWidget) - (panCursor * canvasToWidget));

    BBox hh,vv;
    if (diff[0]>0 && diff[1]>0) {
      hh = BBox(Vector(0,0), Vector(options->width, diff[1]));
      vv = BBox(Vector(0,0), Vector(diff[0], options->height));
    } else if (diff[0]>0 && diff[1]<0) {
      hh = BBox(Vector(options->width,options->height), 
		Vector(0,options->height+diff[1]));
      vv = BBox(Vector(0,0), Vector(diff[0], options->height));
    } else if (diff[0]<0 && diff[1]>0) {
      hh = BBox(Vector(0,0), Vector(options->width, diff[1]));
      vv = BBox(Vector(options->width,options->height), 
		Vector(options->width+diff[0], 0));
    } else if (diff[0]<0 && diff[1]<0) {
      hh = BBox(Vector(options->width,options->height), 
		Vector(0,options->height+diff[1]));
      vv = BBox(Vector(options->width,options->height), 
		Vector(options->width+diff[0], 0));
    }

    hh = hh * widgetToWindow;
    vv = vv * widgetToWindow;
    
    XSetForeground(display, gc, bgColor->pixel);

    Vector hs = hh.size();
    XFillRectangle(display, Tk_WindowId(tkwin), gc, 
		   (int)hh.ll[0], (int)hh.ll[1], (int)hs[0], (int)hs[1]);

    Vector vs = vv.size();
    XFillRectangle(display, Tk_WindowId(tkwin), gc, 
		   (int)vv.ll[0], (int)vv.ll[1], (int)vs[0], (int)vs[1]);

    // display tmp pixmap at new location
    Vector dd = ((v * canvasToWidget) - (panCursor * canvasToWidget)) * 
      widgetToWindow;
    XCopyArea(display, panPM, Tk_WindowId(tkwin), panGCXOR, 
	      0, 0, options->width, options->height, dd[0], dd[1]);
  }
}

void FrameBase::panEndCmd(const Vector& v)
{
  if (!currentContext->cfits)
    return;

  // delete tmp pixmap
  if (panPM)
    Tk_FreePixmap(display, panPM);
  panPM = 0;

  Vector start = panCursor * canvasToUser;
  Vector stop = v * canvasToUser;
  cursor -= stop - start;

  setBinCursor(cursor);
  update(MATRIX);
}

void FrameBase::panToCmd(const Vector& v, InternalSystem sys)
{
  if (!currentContext->cfits)
    return;

  switch (sys) {
  case CANVAS:
    cursor = v * canvasToUser;
    break;
  case PANNER: 
    cursor = v * pannerToUser;
    break;
  }

  setBinCursor(cursor);
  update(MATRIX);
}

void FrameBase::panToCmd(const Vector& v, CoordSystem sys, SkyFrame sky)
{
  if (!currentContext->cfits)
    return;

  cursor = currentContext->cfits->mapToRef(v, sys, sky) * refToUser;

  setBinCursor(cursor);
  update(MATRIX);
}

void FrameBase::rotateBeginCmd()
{
  // save the current rotation
  rotateRotation = rotation;

  // Create src XImage
  if (!(rotateSrcXM = XGetImage(display, pixmap, 0, 0,
				options->width, options->height, 
				AllPlanes, ZPixmap))) {
    internalError("Unable to Create Rotate XImage");
    return;
  }

  // Create dest XImage
  if (!(rotateDestXM = XGetImage(display, pixmap, 0, 0,
				 options->width, options->height, 
				 AllPlanes, ZPixmap))) {
    internalError("Unable to Create Rotate XImage");
    return;
  }

  // Create dest Pixmap
  rotatePM = Tk_GetPixmap(display, Tk_WindowId(tkwin), 
			  options->width, options->height, depth);
  if (!rotatePM) {
    internalError("Unable to Create Rotate Motion Pixmap");
    return;
  }
}

void FrameBase::rotateMotionCmd(double angle)
{
  rotation = rotateRotation + angle;
  rotateMotion();
}

void FrameBase::rotateEndCmd()
{
  // Clean up
  if (rotateSrcXM)
    XDestroyImage(rotateSrcXM);
 
  if (rotateDestXM)
    XDestroyImage(rotateDestXM);

  if (rotatePM)
    Tk_FreePixmap(display, rotatePM);

  update(MATRIX);
}

void FrameBase::zoomAboutCmd(const Vector& z, const Vector& v, 
			     InternalSystem sys)
{
  // assume canvas coords
  Vector az = ((Vector&)z).abs();
  zoom_[0] *= az[0];
  zoom_[1] *= az[1];

  if (currentContext->cfits) {
    switch (sys) {
    case CANVAS:
      cursor = v * canvasToUser;
      break;
    case PANNER:
      cursor = v * pannerToUser;
      break;
    }
    setBinCursor(cursor);
  }

  update(MATRIX);
}

void FrameBase::zoomAboutCmd(const Vector& z, const Vector& v, 
			     CoordSystem sys, SkyFrame sky)
{
  Vector az = ((Vector&)z).abs();
  zoom_[0] *= az[0];
  zoom_[1] *= az[1];

  if (currentContext->cfits) {
    cursor = currentContext->cfits->mapToRef(v, sys, sky) * refToUser;
    setBinCursor(cursor);
  }

  update(MATRIX);
}

void FrameBase::zoomToAboutCmd(const Vector& z, const Vector& v,
			       InternalSystem sys)
{
  // assume canvas coords
  zoom_ = ((Vector&)z).abs();

  if (!currentContext->cfits) {
    switch (sys) {
    case CANVAS:
      cursor = v * canvasToUser;
      break;
    case PANNER:
      cursor = v * pannerToUser;
      break;
    }
    setBinCursor(cursor);
  }

  update(MATRIX);
}

void FrameBase::zoomToAboutCmd(const Vector& z, const Vector& v, 
			       CoordSystem sys, SkyFrame sky)
{
  zoom_ = ((Vector&)z).abs();
  if (currentContext->cfits) {
    cursor = currentContext->cfits->mapToRef(v, sys, sky) * refToUser;
    setBinCursor(cursor);
  }

  update(MATRIX);
}

void FrameBase::zoomToFitCmd(double s)
{
  if (keyContext->fits) {
    // center the image exactly, do not round
    // centerImage();
    cursor = imageCenter() * imageToData;
    double zz = s*calcZoom(Vector(options->width,options->height),imageSize());
    zoom_ = Vector(zz,zz);
    update(MATRIX);
  }
}







