// Copyright (C) 1999-2010
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "grid.h"
#include "context.h"
#include "framebase.h"
#include "fitsimage.h"

extern "C" {
  #include "ast.h"
}

extern GridBase* astGridPtr;

Grid::Grid(Widget* p, 
	   CoordSystem sys, SkyFrame sky, SkyFormat format, 
	   GridType t, const char* o) 
  : GridBase(p,o), system_(sys), sky_(sky), skyFormat_(format), type_(t)
{
}

int Grid::render()
{
  FrameBase* pp = (FrameBase*)parent;

  // X11 vars
  gc_ = pp->gridGC;
  pixmap_ = pp->basePixmap;

  return doit(X11);
}

int Grid::ps(int mode)
{
  FrameBase* pp = (FrameBase*)parent;

  // ps vars
  matrix_ = pp->getWidgetToCanvas();
  mode_ = mode;

  return doit(PS);
}

#ifdef _MACOSX
int Grid::macosx()
{
  FrameBase* pp = (FrameBase*)parent;

  matrix_ = pp->getWidgetToCanvas();

  return doit(MACOSX);
}
#endif

#ifdef _WIN32
int Grid::win32()
{
  FrameBase* pp = (FrameBase*)parent;

  matrix_ = pp->getWidgetToCanvas();

  return doit(GWIN32);
}
#endif

int Grid::doit(RenderMode rm)
{
  FrameBase* pp = (FrameBase*)parent;
  FitsImage* fits = pp->keyContext->fits;

  if (!fits)
    return 1;

  int width = fits->width();
  int height = fits->height();

  astClearStatus; // just to make sure
  astBegin; // start memory management

  AstFrameSet* frameSet = NULL;
  AstPlot* plot = NULL;

  if (!(frameSet = astFrameSet(astFrame(2,"Domain=WIDGET"),"")))
    goto error;

  // map from Widget to Image
  matrixMap(frameSet,fits->getWidgetToImage(),"Domain=IMAGE");

  switch (system_) {
  case IMAGE:
    break;

  case PHYSICAL:
    matrixMap(frameSet,fits->getImageToPhysical(),"Domain=PHYSICAL");
    break;

  case AMPLIFIER:
    matrixMap(frameSet,fits->getImageToAmplifier(),"Domain=AMPLIFIER");
    break;

  case DETECTOR:
    matrixMap(frameSet,fits->getImageToDetector(),"Domain=DETECTOR");
    break;

  default:
    if (!fits->hasWCS(system_))
      goto error;

    if (fits->hasWCSEqu(system_)) {
      // map from Image to Equatorial WCS

      WorldCoor* wcs = fits->getWCS(system_);

      AstFrameSet* wcsfs = NULL;

      // we can handle most projections, but not these
      if (wcs->prjcode==WCS_PIX ||
	  wcs->prjcode==WCS_LIN ||
	  wcs->prjcode==WCS_PLT)
	goto error;

      // try to build a valid header, so that wcssubs and ast will agree.
      // however, for certain projections, just feed the header directly to
      // ast.
      if (wcs->prjcode==WCS_DSS || wcs->prjcode==WCS_TNX) 
	wcsfs = (AstFrameSet*)fits2ast();
      else
      	wcsfs = (AstFrameSet*)wcs2ast(wcs);

      if (!wcsfs)
	goto error;

      // set desired WCS system

      switch (sky_) {
      case FK4:
	astSet(wcsfs, "System=FK4, Equinox=B1950");
	break;
      case FK5:
	astSet(wcsfs, "System=FK5, Equinox=J2000");
	break;
      case ICRS:
	astSet(wcsfs, "System=FK5, Equinox=J2000");
	break;
      case GALACTIC:
	astSet(wcsfs, "System=Galactic");
	break;
      case ECLIPTIC:
	astSet(wcsfs, "System=Ecliptic");
	// this is a kludge to get AST to agree with wcssubs
	astSetD(wcsfs, "Equinox", astGetD(wcsfs, "Epoch"));
	break;
      }

      // invert wcsfs  base/current 
      astInvert(wcsfs);

      // add wcsfs to frameset
      // this will link frame 2 of frameset to frame 3 wcsfs
      astAddFrame(frameSet,2,astUnitMap(2,""),wcsfs);
      // set the current of frameset to 4 (was 3)
      astSetI(frameSet,"current",4);
    }
    else {
      // map from Image to Linear WCS
      WorldCoor* wcs = fits->getWCS(system_);
      Matrix m(wcs->cd[0],wcs->cd[1],wcs->cd[2],wcs->cd[3],
	       wcs->crval[0]-
	       (wcs->cd[0]*wcs->crpix[0])-(wcs->cd[2]*wcs->crpix[1]),
	       wcs->crval[1]-
	       (wcs->cd[1]*wcs->crpix[0])-(wcs->cd[3]*wcs->crpix[1]));
      matrixMap(frameSet,m,"Domain=WCS");
    }
    break;
  }

  astSet(frameSet,"Title=%s", " ");

  if (DebugAST) {
    int status = astStatus;
    astClearStatus;
    astShow(frameSet);
    astSetStatus(status);
  }

  // create astPlot
  float gbox[4];
  double pbox[4];

  switch (type_) {
  case ANALYSIS:
    {
      gbox[0] = pbox[0] = 0;
      gbox[1] = pbox[1] = 0;
      gbox[2] = pbox[2] = pp->options->width-1;
      gbox[3] = pbox[3] = pp->options->height-1;
    }
    break;
  case PUBLICATION:
    {
      Matrix m = fits->getImageToWidget();
      BBox b = pp->imageBBox() * fits->getRefToData();

      Vector ll = b.ll   * m;
      Vector lr = b.lr() * m;
      Vector ur = b.ur   * m;
      Vector ul = b.ul() * m;
      
      BBox bb(ll,ll);
      bb.bound(lr);
      bb.bound(ur);
      bb.bound(ul);
      bb += Vector(.5,.5);

      gbox[0] = pbox[0] = bb.ll[0];
      gbox[1] = pbox[1] = bb.ll[1];
      gbox[2] = pbox[2] = bb.ur[0];
      gbox[3] = pbox[3] = bb.ur[1];
    }
    break;
  }

  plot = astPlot(frameSet, gbox, pbox, option_);

  // and now create astGrid

  astGridPtr = this;
  renderMode = rm;
  astGrid(plot);

  astEnd; // now, clean up memory
  return 1;

error:
  astEnd;
  return 0;
}

// AST support

void* Grid::fits2ast() 
{
  FrameBase* pp = (FrameBase*)parent;

  // read FITS header into astChannel

  // we may have an error, just reset
  astClearStatus;

  // new fitschan
  AstFitsChan* chan = astFitsChan(NULL, NULL, "");
  if (!astOK || chan == AST__NULL)
    return NULL;

  // no warning messages
  astClear(chan,"Warnings");

  // fill up chan
  char* header = pp->currentContext->fits->getCards();
  int ncards = pp->currentContext->fits->getNCards();
  for (int i=0; i<ncards; i++) {
    char buf[81];
    strncpy(buf,header+(i*80),80);
    buf[80] = '\0';

    astPutFits(chan, buf, 0);
    // sometimes, we get a bad parse, just ignore
    if (!astOK)
      astClearStatus;
  }

  // parse FITS header for WCS

  // we may have an error, just reset
  astClearStatus;
  astClear(chan, "Card");

  // parse header
  AstFrameSet* frameSet = (AstFrameSet*)astRead(chan);

  // do we have anything?
  if (!astOK || frameSet == AST__NULL || 
      strcmp(astGetC(frameSet,"Class"), "FrameSet"))
    return NULL;

  return frameSet;
}

void* Grid::wcs2ast(void* vwcs) 
{
  FrameBase* pp = (FrameBase*)parent;

  // read wcs struct into astChannel

  // we may have an error, just reset
  astClearStatus;

  // new fitschan
  AstFitsChan* chan;
  chan = astFitsChan(NULL, NULL, "");
  if (!astOK || chan == AST__NULL)
    return NULL;

  // no warning messages
  astClear(chan,"Warnings");

  // fill up chan

  FitsFile* fits = (pp->keyContext->fits)->imageFile();
  WorldCoor* wcs = (WorldCoor*)vwcs;

  if (!(strncmp("RA---TNX", wcs->ctype[0], 8))) {
    // not used, use fits2ast
    /*
    putFitsCard(chan, "CTYPE1", "RA---TAN");
    putFitsCard(chan, "CTYPE2", "DEC--TAN");

    for (int i=0; i<=2; i++)
      for (int m=0; m<=9; m++) {
	ostringstream str;
	str << "WAT" << i << '_' << "00" << m << ends;
	if (fits->find(str.str().c_str())) {
	  float val = fits->getReal(str.str().c_str(),0);
	  putFitsCard(chan, str.str().c_str(), val);
	}
      }
    */
  }
  else if (!(strncmp("XLON", wcs->ctype[0], 4))) {
    // this is really just ra/dec with north pole of value 0
    putFitsCard(chan, "CTYPE1", "RA---TAN");
    putFitsCard(chan, "CTYPE2", "DEC--TAN");
    putFitsCard(chan, "LATPOLE", -90);
  }
  else {
    putFitsCard(chan, "CTYPE1", wcs->ctype[0]);
    putFitsCard(chan, "CTYPE2", wcs->ctype[1]);
  }
 
  putFitsCard(chan, "CRPIX1", wcs->crpix[0]);
  putFitsCard(chan, "CRPIX2", wcs->crpix[1]);

  putFitsCard(chan, "CRVAL1", wcs->crval[0]);
  putFitsCard(chan, "CRVAL2", wcs->crval[1]);

  // This is very complicated/. AST is very, very, very picky as to which
  // keywords it use...
  //
  //   first see if any CD's present
  //     if so, if there is a rotation, or LONPOLE/LATPOLE, use CD's,
  //       else use just CDELTS
  //   else use CDELT's and
  //     see if any PC's present, use them
  //       else see if any CROTA's present, use them

  if (fits->find("CD1_1")) {
    if (!wcs->cd[1] && !wcs->cd[2] && 
	wcs->latpole == 999 && wcs->longpole == 999) {
      putFitsCard(chan, "CDELT1", wcs->cdelt[0]);
      putFitsCard(chan, "CDELT2", wcs->cdelt[1]);
    }
    else {
      putFitsCard(chan, "CD1_1", wcs->cd[0]);
      putFitsCard(chan, "CD1_2", wcs->cd[1]);
      putFitsCard(chan, "CD2_1", wcs->cd[2]);
      putFitsCard(chan, "CD2_2", wcs->cd[3]);
    }
  }
  else if (fits->find("CDELT1")) {
    putFitsCard(chan, "CDELT1", wcs->cdelt[0]);
    putFitsCard(chan, "CDELT2", wcs->cdelt[1]);

    if (fits->find("PC1_1")) {
      putFitsCard(chan, "PC1_1", wcs->pc[0]);
      putFitsCard(chan, "PC1_2", wcs->pc[1]);
      putFitsCard(chan, "PC2_1", wcs->pc[2]);
      putFitsCard(chan, "PC2_2", wcs->pc[3]);
    }
    else if (fits->find("PC001001")) {
      putFitsCard(chan, "PC001001", wcs->pc[0]);
      putFitsCard(chan, "PC001002", wcs->pc[1]);
      putFitsCard(chan, "PC002001", wcs->pc[2]);
      putFitsCard(chan, "PC002002", wcs->pc[3]);
    }
    else {
      if (fits->find("CROTA1"))
	putFitsCard(chan, "CROTA1", wcs->rot);
      if (fits->find("CROTA2"))
	putFitsCard(chan, "CROTA2", wcs->rot);
    }
  }

  // equiniox
  putFitsCard(chan, "EQUINOX", wcs->equinox);

  // from wcssub/wcsinit.c line 800
  // wcs->epoch = 1900.0 + (mjd - 15019.81352) / 365.242198781;
  putFitsCard(chan, "MJD-OBS", (wcs->epoch-1900)*365.242198781+15019.81352);

  if (!strncmp("RA",wcs->ctype[0],2) || !strncmp("RA",wcs->ctype[1],2)) {
    if (!strncmp("FK4",wcs->radecsys,3) ||
	!strncmp("FK5",wcs->radecsys,3) ||
	!strncmp("ICRS",wcs->radecsys,3))
      putFitsCard(chan, "RADESYS", wcs->radecsys);
  }

  if (wcs->latpole != 999)
    putFitsCard(chan, "LATPOLE", wcs->latpole);
  if (wcs->longpole != 999)
    putFitsCard(chan, "LONPOLE", wcs->longpole);

  // Projection parameters
  // PVi_m
  for (int i=1; i<=2; i++)
    for (int m=0; m<=9; m++) {
      ostringstream str;
      str << "PV" << i << '_' << m << ends;
      if (fits->find(str.str().c_str())) {
	float val = fits->getReal(str.str().c_str(),0);
	putFitsCard(chan, str.str().c_str(), val);
      }
    }

  // PSi_m
  for (int i=1; i<=2; i++)
    for (int m=0; m<=9; m++) {
      ostringstream str;
      str << "PS" << i << '_' << m << ends;
      if (fits->find(str.str().c_str())) {
	float val = fits->getReal(str.str().c_str(),0);
	putFitsCard(chan, str.str().c_str(), val);
      }
    }
   
  // rewind chan
  astClear(chan, "Card");

  // parse header
  AstFrameSet* frameSet = (AstFrameSet*)astRead(chan);

  // do we have anything?
  if (!astOK || frameSet == AST__NULL || 
      strcmp(astGetC(frameSet,"Class"), "FrameSet"))
    return NULL;

  if (wcs->coorflip) {
    int orr[] = {2,1};
    astPermAxes(frameSet,orr);
  }

  return frameSet;
}

void Grid::putFitsCard(void* chan, const char* key, const char* value)
{
  char buf[80];
  memset(buf,'\0', 80);

  ostringstream str;
  str.setf(ios::left,ios::adjustfield);
  str.width(8);
  str << key << "= '" << value << "'";
  memcpy(buf,str.str().c_str(),str.str().length());

  astPutFits(chan, buf, 0);
  astClearStatus;

  if (DebugAST)
    cerr << buf << endl;
}

void Grid::putFitsCard(void* chan, const char* key, double value)
{
  char buf[80];
  memset(buf,'\0', 80);

  ostringstream str;
  str.setf(ios::left,ios::adjustfield);
  str.setf(ios::scientific,ios::floatfield);
  str.width(8);
  str.precision(12);
  str << key << "= " << value;
  memcpy(buf,str.str().c_str(),str.str().length());

  astPutFits(chan, buf, 0);
  astClearStatus;

  if (DebugAST)
    cerr << buf << endl;
}

