// Copyright (C) 1999-2010
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "marker.h"
#include "framebase.h"
#include "fitsimage.h"

static int markerSeqID = 1;

// Marker Members Public

Marker::Marker(const Marker& a)
{
  id = a.id;
  strcpy(type, a.type);
  parent = a.parent;

  center = a.center;
  bbox = a.bbox;
  allBBox = a.allBBox;

  angle = a.angle;
  flip_ = a.flip_;

  numHandle = a.numHandle;
  if (numHandle) {
    handle = new Vector[numHandle];
    for (int i=0; i<numHandle; i++)
      handle[i] = a.handle[i];
  }
  else
    handle = NULL;

  colorName = dupstr(a.colorName);
  color = a.color;
  lineWidth = a.lineWidth;
  properties = a.properties;
  selected = a.selected;
  highlited = a.highlited;

  dashlist[0] = a.dashlist[0];
  dashlist[1] = a.dashlist[1];

  text = dupstr(a.text);
  tkfont_ = Tk_GetFont(parent->interp, parent->tkwin, Tk_NameOfFont(a.tkfont_));
  psfont_ = Tk_GetFont(parent->interp, parent->tkwin, Tk_NameOfFont(a.psfont_));
  comment = dupstr(a.comment);

  display = a.display;
  gc = a.gc;
  gcxor = a.gcxor;
  
  for (int ii=0; ii<XMLNUMCOL; ii++)
    XMLCol[ii] = NULL;

  // disable Callbacks by default
  doCB = 0;

  tags = a.tags;
  callbacks = a.callbacks;

  previous_ = NULL;
  next_ = NULL;
}

Marker::Marker(FrameBase* p, const Vector& ctr, 
	       double ang,
	       const char* clr, int* dsh, 
	       int w, const char* f, const char* t, 
	       unsigned short prop, const char* c, 
	       const List<Tag>& tg, const List<CallBack>& cb)
{
  id = markerSeqID++;
  type[0] = '\0';
  parent = p;

  center = ctr;

  angle = ang;
  if (!parent->isIIS())
    flip_ = FlipY();

  handle = NULL;
  numHandle = 0;

  colorName = dupstr(clr);
  color = parent->getColor(colorName);
  lineWidth = w;
  properties = prop;
  selected = 0;
  highlited = 0;

  dashlist[0] = dsh[0];
  dashlist[1] = dsh[1];

  text = dupstr(t);
  tkfont_ =NULL;
  psfont_ =NULL;
  initFonts(f);

  comment = dupstr(c);

  display = parent->display;
  gc = parent->markerGC;
  gcxor = parent->markerGCXOR;

  for (int ii=0; ii<XMLNUMCOL; ii++)
    XMLCol[ii] = NULL;

  doCB = 1;

  tags = tg;
  callbacks = cb;

  previous_ = NULL;
  next_ = NULL;
}

Marker::~Marker()
{
  if (colorName)
    delete [] colorName;

  if (text)
    delete [] text;

  if (comment)
    delete [] comment;

  if (tkfont_)
    Tk_FreeFont(tkfont_);

  if (psfont_)
    Tk_FreeFont(psfont_);

  if (handle)
    delete [] handle;

  for (int ii=0; ii<XMLNUMCOL; ii++)
    if (XMLCol[ii])
      delete [] XMLCol[ii];

  doCallBack(CallBack::DELETECB);
}

// draw

void Marker::draw(Drawable drawable, int tt)
{
  renderXHandles(drawable);
  if (tt)
    renderXText(drawable, parent->refToWidget, SRC);
  renderX(drawable, parent->refToWidget, SRC);
}

void Marker::drawXOR()
{
  renderXText(Tk_WindowId(parent->tkwin), parent->refToWindow, XOR);
  renderX(Tk_WindowId(parent->tkwin), parent->refToWindow, XOR);
}

void Marker::drawMagnifier(const Matrix& mx, int tt)
{
  if (tt)
    renderXText(parent->magnifierPixmap, mx, SRC);
  renderX(parent->magnifierPixmap, mx, SRC);
}

void Marker::renderXHandles(Drawable drawable)
{
  XSetForeground(display, gc, color);

  // handles are of length 5
  if (selected && canSelect())
    for (int i=0; i<numHandle; i++) {
      Vector v = (handle[i]*parent->canvasToWidget - Vector(2,2)).round();
      XFillRectangle(display, drawable, gc, v[0], v[1], 5, 5);
    }
}

void Marker::renderXText(Drawable drawable, const Matrix& mx, RenderMode mode)
{
  GC lgc;
  switch (mode) {
  case SRC:
    lgc = gc;
    XSetForeground(display, gc, color);
    break;
  case XOR:
    lgc = gcxor;
    break;
  }

  if (tkfont_)
    XSetFont(display, lgc, Tk_FontId(tkfont_));

  if (text && tkfont_) {
    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(tkfont_, &metrics);
    int width = Tk_TextWidth(tkfont_, text, strlen(text));

    BBox bb = bbox * parent->canvasToRef * mx;
    Vector bbc = bb.center();
    Vector t =  Vector(bbc[0], bb.ll[1]) * 
      Translate(-width/2., -metrics.descent);
    Tk_DrawChars(display, drawable, lgc, tkfont_, text, strlen(text), 
		 t[0], t[1]);
  }
}

void Marker::renderXArrow(Drawable drawable, const Vector& a, const Vector& b,
			  GC lgc)
{
  Vector* c = arrow(a,b);
  XPoint cc[6];
  for (int i=0; i<6; i++) {
    cc[i].x = (short)c[i][0];
    cc[i].y = (short)c[i][1];
  }

  XFillPolygon(display, drawable, lgc, cc, 6, Nonconvex, CoordModeOrigin);
  delete [] c;
}

GC Marker::renderXGC(RenderMode mode)
{
  // set width, color, dash
  switch (mode) {
  case SRC:
    XSetForeground(display, gc, color); 
    if ((properties & SOURCE) && !(properties & DASH))
      renderXLineNoDash(gc);
    else
      renderXLineDash(gc);
    return gc;

  case XOR:
    renderXLineDash(gcxor);
    return gcxor;
  }
}

void Marker::renderXLineDash(GC lgc)
{
  char dlist[2];
  dlist[0] = dashlist[0];
  dlist[1] = dashlist[1];
  int ww = (highlited && canHighlite()) ? lineWidth*2 : lineWidth;
  XSetDashes(display, lgc, 0, dlist, 2);
  XSetLineAttributes(display, lgc, ww, LineOnOffDash, CapButt, JoinMiter);
}

void Marker::renderXLineNoDash(GC lgc)
{
  int ww = (highlited && canHighlite()) ? lineWidth*2 : lineWidth;
  XSetLineAttributes(display, lgc, ww, LineSolid, CapButt, JoinMiter);
}

// ps

void Marker::ps(int mode, int tt)
{
  if (tt)
    renderPSText(mode);
  renderPS(mode);
}

void Marker::renderPSText(int mode)
{
  // set color
  renderPSColor(mode, parent->getXColor(colorName));

  // do text
  if (text && psfont_) {
    ostringstream str;

    Tcl_DString psdstr;
    Tcl_DStringInit(&psdstr);
    int psSize = Tk_PostscriptFontName(psfont_, &psdstr) *
      parent->getDisplayRatio();

    str << '/' << Tcl_DStringValue(&psdstr) 
	<< " findfont " << psSize << " scalefont setfont" << endl;
    Tcl_DStringFree(&psdstr);

    Vector bbc = bbox.center();
    Vector t =  Vector(bbc[0], bbox.ll[1]);

    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(psfont_, &metrics);

    str << t.TkCanvasPs(parent->canvas) << " moveto" << endl
	<< '(' << psQuote(text) << ')' << endl
	<< "dup stringwidth pop 2 div 0 exch sub " 
	<< metrics.descent << " rmoveto show" << endl
	<< ends;

    Tcl_AppendResult(parent->interp, (char*)str.str().c_str(), NULL);
  }
}

void Marker::renderPSArrow(const Vector& a, const Vector& b)
{
  Vector d = (a-b).normalize();
  Vector* c = arrow(a,b);
  for (int i=0; i<6; i++) {
    ostringstream str;
    if (i==0)
      str << "newpath " << endl
	  << c[i].TkCanvasPs(parent->canvas) << " moveto" << endl << ends;
    else
      str << c[i].TkCanvasPs(parent->canvas) << " lineto" << endl << ends;

    Tcl_AppendResult(parent->interp, (char*)str.str().c_str(), NULL);
  }

  ostringstream str;
  str << "closepath fill" << endl << ends;
  Tcl_AppendResult(parent->interp, (char*)str.str().c_str(), NULL);

  delete [] c;
}

void Marker::renderPSGC(int mode)
{
  // set width, color, dash
  renderPSColor(mode, parent->getXColor(colorName));
  if ((properties & SOURCE) && !(properties & DASH))
    renderPSLineNoDash();
  else
    renderPSLineDash();
}

void Marker::renderPSLineDash()
{
  ostringstream str;
  str << lineWidth << " setlinewidth" << endl;
  str << "[" << dashlist[0] << " " << dashlist[1] << "] 0 setdash" << endl << ends;
  Tcl_AppendResult(parent->interp, (char*)str.str().c_str(), NULL);
}

void Marker::renderPSLineNoDash()
{
  ostringstream str;
  str << lineWidth << " setlinewidth" << endl;
  str << "[] 0 setdash" << endl << ends;
  Tcl_AppendResult(parent->interp, (char*)str.str().c_str(), NULL);
}

void Marker::renderPSColor(int mode, XColor* clr)
{
  ostringstream str;

  switch ((Widget::PSColorSpace)mode) {
  case Widget::BW:
  case Widget::GRAY:
    psColorGray(clr, str);
    str << " setgray";
    break;
  case Widget::RGB:
    psColorRGB(clr, str);
    str << " setrgbcolor";
    break;
  case Widget::CMYK:
    psColorCMYK(clr, str);
    str << " setcmykcolor";
    break;
  }
  str << endl << ends;

  Tcl_AppendResult(parent->interp, (char*)str.str().c_str(), NULL);
}

#ifdef _MACOSX
void Marker::macosx(int tt)
{
  if (tt)
    renderMACOSXText();
  renderMACOSX();
}

void Marker::renderMACOSXText()
{
  // set color
  macosxColor(parent->getXColor(colorName));

  // do text
  if (text && psfont_) {
    Tcl_DString psdstr;
    Tcl_DStringInit(&psdstr);
    int psSize = Tk_PostscriptFontName(psfont_, &psdstr);
    macosxFont(Tcl_DStringValue(&psdstr), psSize);
    Tcl_DStringFree(&psdstr);

    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(psfont_, &metrics);
    int width = Tk_TextWidth(psfont_, text, strlen(text));

    Vector bbc = bbox.center();
    Vector t =  Vector(bbc[0], bbox.ll[1]) * 
      Translate(-width/2., -metrics.descent);

    macosxDrawText(t, 0, text);
  }
}

void Marker::renderMACOSXArrow(const Vector& a, const Vector& b)
{
  Vector* cc = arrow(a,b);
  Vector vv[6];
  for (int ii=0; ii<6; ii++)
    vv[ii] = cc[ii];

  macosxFillPolygon(vv,6);
  delete [] cc;
}

void Marker::renderMACOSXGC()
{
  // set width, color, dash
  macosxColor(parent->getXColor(colorName));

  if ((properties & SOURCE) && !(properties & DASH))
    renderMACOSXLineNoDash();
  else
    renderMACOSXLineDash();
}

void Marker::renderMACOSXLineDash()
{
  macosxWidth(lineWidth);
  float dlist[2];
  dlist[0] = dashlist[0];
  dlist[1] = dashlist[1];
  macosxDash(dlist,2);
}

void Marker::renderMACOSXLineNoDash()
{
  macosxWidth(lineWidth);
  macosxDash(NULL,0);
}
#endif

#ifdef _WIN32
void Marker::win32(int tt)
{
  if (tt)
    renderWIN32Text();
  renderWIN32();
}

void Marker::renderWIN32Text()
{
  // set color
  win32Color(parent->getXColor(colorName));

  // do text
  if (text && psfont_) {
    Tcl_DString psdstr;
    Tcl_DStringInit(&psdstr);
    int psSize = Tk_PostscriptFontName(psfont_, &psdstr);
    win32Font(Tcl_DStringValue(&psdstr), psSize);
    Tcl_DStringFree(&psdstr);

    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(psfont_, &metrics);
    int width = Tk_TextWidth(psfont_, text, strlen(text));

    Vector bbc = bbox.center();
    Vector t =  Vector(bbc[0], bbox.ll[1]) * 
      Translate(-width/2., -metrics.descent);

    win32DrawText(t, 0, text);
  }
}

void Marker::renderWIN32Arrow(const Vector& a, const Vector& b)
{
  Vector* cc = arrow(a,b);
  Vector vv[6];
  for (int ii=0; ii<6; ii++)
    vv[ii] = cc[ii];

  win32FillPolygon(vv,6);
  delete [] cc;
}

void Marker::renderWIN32GC()
{
  // set width, color, dash
  win32Color(parent->getXColor(colorName));

  if ((properties & SOURCE) && !(properties & DASH))
    renderWIN32LineNoDash();
  else
    renderWIN32LineDash();
}

void Marker::renderWIN32LineDash()
{
  win32Width(lineWidth);
  float dlist[2];
  dlist[0] = dashlist[0];
  dlist[1] = dashlist[1];
  win32Dash(dlist,2);
}

void Marker::renderWIN32LineNoDash()
{
  win32Width(lineWidth);
  win32Dash(NULL,0);
}
#endif

// other

void Marker::deleteCBs()
{
  callbacks.deleteAll();
}

void Marker::newIdentity()
{
  id = markerSeqID++;

  doCB = 1;
  //  deleteCBs();
  updateBBox();
}

void Marker::updateCoords(const Matrix& mx)
{
  center*=mx;
  updateBBox();
}

void Marker::centroid()
{
  center = parent->centroid(center);
  updateBBox();
  doCallBack(CallBack::MOVECB);
}

void Marker::moveTo(const Vector& v)
{
  center=v;
  updateBBox();
  doCallBack(CallBack::MOVECB);
}

void Marker::moveBegin()
{
  doCallBack(CallBack::MOVEBEGINCB);
}

void Marker::move(const Vector& v)
{
  center+=v;
  updateBBox();
  doCallBack(CallBack::MOVECB);
}

void Marker::moveEnd()
{
  doCallBack(CallBack::MOVEENDCB);
}

void Marker::editBegin(int)
{
  doCallBack(CallBack::EDITBEGINCB);
}

void Marker::edit(const Vector& v, int h)
{
  doCallBack(CallBack::EDITCB);
}

void Marker::editEnd()
{
  doCallBack(CallBack::EDITENDCB);
}

void Marker::rotateBegin()
{
  doCallBack(CallBack::ROTATEBEGINCB);
}

void Marker::rotate(const Vector& v, int h)
{
  // v is in ref coords
  // handles are in canvas coords

  double a = (v * Translate(-center) * flip_).angle();
  double b = ((handle[h-1]*parent->canvasToRef * Translate(-center) * flip_)).angle();
  angle -= a-b;

  updateBBox();
  doCallBack(CallBack::ROTATECB);
}

void Marker::rotateEnd()
{
  doCallBack(CallBack::ROTATEENDCB);
}

void Marker::select() {
  // only call the CB if not already selected
  if (!selected)
    doCallBack(CallBack::SELECTCB);
  selected = 1;
}

void Marker::unselect() {
  // only call the CB if already selected
  if (selected)
    doCallBack(CallBack::UNSELECTCB);
  selected = 0;
}

void Marker::toggleSelect() {
  selected = !selected;
  if (selected)
    doCallBack(CallBack::SELECTCB);
  else
    doCallBack(CallBack::UNSELECTCB);
}

void Marker::highlite() {
  // only call the CB if not already highlited
  if (!highlited)
    doCallBack(CallBack::HIGHLITECB);
  highlited = 1;
}

void Marker::unhighlite() {
  // only call the CB if already highlited
  if (highlited)
    doCallBack(CallBack::UNHIGHLITECB);
  highlited = 0;
}

void Marker::toggleHighlite() {
  highlited = !highlited;
  if (highlited)
    doCallBack(CallBack::HIGHLITECB);
  else
    doCallBack(CallBack::UNHIGHLITECB);
}

void Marker::key() {
  doCallBack(CallBack::KEYCB);
}

Vector Marker::getHandle(int h)
{
  if (h<numHandle)
    return handle[h];
  else
    return Vector();
}

void Marker::setAngle(double a)
{
  angle = a;
  updateBBox();

  doCallBack(CallBack::ROTATECB);
}

void Marker::setColor(const char* clr)
{
  if (colorName)
    delete [] colorName;

  colorName = dupstr(clr);
  color = parent->getColor(colorName);

  doCallBack(CallBack::COLORCB);
}

void Marker::setLineWidth(int w)
{
  lineWidth = w;

  doCallBack(CallBack::LINEWIDTHCB);
}

void Marker::setFont(const char* f)
{
  initFonts(f);
  updateBBox();
  doCallBack(CallBack::FONTCB);
}

void Marker::initFonts(const char* ff)
{
  if (tkfont_)
    Tk_FreeFont(tkfont_);
  tkfont_ = NULL;
  if (psfont_)
    Tk_FreeFont(psfont_);
  psfont_ = NULL;

  char* dd = "helvetica 9 roman normal";
  if (!ff)
    ff = dd;

  psfont_ = Tk_GetFont(parent->interp, parent->tkwin, ff);

  char* ptr = (char*)ff;
  while (*ptr && *ptr != ' ')
    ptr++;

  ostringstream fstr;
  if (!strncmp(ff,"helvetica",4))
    fstr << parent->options->helvetica << ptr << ends;
  else if (!strncmp(ff,"times",4))
    fstr << parent->options->times << ptr << ends;
  else if (!strncmp(ff,"courier",4))
    fstr << parent->options->courier << ptr << ends;
  else
    fstr << parent->options->helvetica << ptr << ends;

  tkfont_ = Tk_GetFont(parent->getInterp(), parent->getTkwin(), 
		       fstr.str().c_str());
}

const char* Marker::getFont()
{
  if (psfont_)
    return Tk_NameOfFont(psfont_);
  else
    return NULL;
}

void Marker::addTag(const char* tg)
{
  Tag* t = new Tag(tg);
  tags.append(t);
}

void Marker::editTag(const char* from, const char* to)
{
  // change any tags
  {
    Tag* t = tags.head();
    while (t) {
      if (!strcmp(t->tag(),from)) {
	t->set(to);
      }
      t=t->next();
    }
  }

  // now, remove duplicates
  {
    Tag* t = tags.head();
    while (t) {
      Tag* tt=t->next();
      while (tt) {
	if (!strcmp(t->tag(),tt->tag())) {
	  Tag* ntt = tags.extractNext(tt);
	  delete tt;
	  tt = ntt;
	}
	else
	  tt=tt->next();
      }
      t=t->next();
    }
  }
}

void Marker::deleteTags()
{
  tags.deleteAll();
}

void Marker::deleteTag(const char* tg)
{
  Tag* t = tags.head();
  while (t) {
    if (!strcmp(t->tag(),tg)) {
      tags.extractNext(t);
      delete t;
      return;
    }
    t = t->next();
  }
}

void Marker::deleteTag(int w)
{
  Tag* t = tags.head();
  for (int i=0; i<w; i++)
    if (t)
      t = t->next();
    else
      break;

  if (t) {
    tags.extractNext(t);
    delete t;
  }
}

const char* Marker::getTag()
{
  Tag* t = tags.head();
  if (t)
    return t->tag();
  else
    return NULL;
}

const char* Marker::getNextTag()
{
  Tag* t = tags.next();
  if (t)
    return t->tag();
  else
    return NULL;
}

const char* Marker::getTag(int w)
{
  Tag* t = tags.head();
  for (int i=0; i<w; i++)
    if (t)
      t = t->next();
    else
      break;

  if (t)
    return t->tag();
  else
    return NULL;
}

int Marker::hasTag(const char* tg)
{
  Tag* t = tags.head();
  while (t) {
    if (!strcmp(t->tag(),tg))
      return 1;
    t = t->next();
  }
  return 0;
}

int Marker::onHandle(const Vector& v)
{
  // return handle number
  // work last to first for annuli

  for (int i=numHandle-1; i>=0; i--) {
    BBox bb(handle[i]);
    bb.expand(2);
    if (bb.isIn(v))
      return i+1;
  }
  return 0;
}

int Marker::getProperty(unsigned short which)
{
  return (properties & which) ? 1 : 0;
}

void Marker::setText(const char* str)
{
  if (text)
    delete [] text;
  text = dupstr(str);

  updateBBox();
  doCallBack(CallBack::TEXTCB);
}

void Marker::setProperty(unsigned short prop, int value)
{
  if (value)
    properties |= prop;
  else
    properties &= ~prop;

  if (prop == FIXED) // bbox will change
    updateBBox();

  doCallBack(CallBack::PROPERTYCB);
}

int Marker::addCallBack(CallBack::Type t, const char* proc, const char* arg)
{
  CallBack* cb = new CallBack(parent->interp, t, proc, arg);
  if (cb) {
    callbacks.append(cb);
    return TCL_OK;
  }

  return TCL_ERROR;
}

void Marker::deleteCallBack(CallBack::Type t)
{
  CallBack* cb = callbacks.head();
  while (cb) {
    if (cb->type() == t) {
      CallBack* next = callbacks.extractNext(cb);
      delete cb;
      cb = next;
    }
    else
      cb = cb->next();
  }
}

int Marker::deleteCallBack(CallBack::Type t, const char* proc)
{
  CallBack* cb = callbacks.head();
  while (cb) {
    if (cb->type() == t && (!strcmp(cb->proc(), proc))) {
      callbacks.extractNext(cb);
      delete cb;
      return TCL_OK;
    }
    else
      cb = cb->next();
  }

  return TCL_ERROR;
}

int Marker::isVisible(const BBox& b)
{
  // assume visible, prove otherwise
  // all coords are in canvas coords

  BBox bb(b);
  return 
    !((allBBox.ur[0] < bb.ll[0]) ||
    (allBBox.ll[0] > bb.ur[0]) ||
    (allBBox.ur[1] < bb.ll[1]) ||
    (allBBox.ll[1] > bb.ur[1]));
}

void Marker::doCallBack(CallBack::Type t)
{
  if (!doCB)
    return;

  ostringstream str;

  str << id << ends;

  CallBack* cb=callbacks.head();
  while (cb) {
    if (cb->type() == t)
      if (cb->eval(str.str().c_str())) {
	ostringstream estr;
	estr << "Unable to eval Marker CallBack "
	     << cb->proc() << " : " << parent->interp->result << ends;
	internalError(estr.str().c_str());
      }
    cb=cb->next();
  }
}

void Marker::internalError(const char* msg)
{
  if (parent)
    parent->internalError(msg);
}

// Protected

double Marker::calcAngle()
{
  switch (parent->getOrientation()) {
  case NONE:
  case XY:
    return angle + parent->getRotation();
  case XX:
  case YY:
    return -angle + parent->getRotation();
  }
}

double Marker::calcAngle(double a)
{
  switch (parent->getOrientation()) {
  case NONE:
  case XY:
    return a + parent->getRotation();
  case XX:
  case YY:
    return -a + parent->getRotation();
  }
}

void Marker::calcAllBBox()
{
  allBBox = bbox;

  if (text && tkfont_) {
    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(tkfont_, &metrics);
    int width = Tk_TextWidth(tkfont_, text, strlen(text));

    Vector bbc = bbox.center();
    Vector ll =  Vector(bbc[0], bbox.ll[1]) * Translate(-width/2., 0);
    Vector ur = ll * Translate(width, -(metrics.linespace));

    allBBox.bound(ll);
    allBBox.bound(ur);
  }
}

Vector Marker::modifyArrow(const Vector& a, const Vector& b)
{
  const int tip = 6;  // length from end of line to tip of arrow

  double ang = ((b-a).normalize()).angle();
  return b * Translate(-a) * Rotate(ang) * Translate(-tip,0) * 
      Rotate(-ang) * Translate(a);
}

Vector* Marker::arrow(const Vector& a, const Vector& b)
{
  Vector d = (a-b).normalize();

  const int tip = 6;  // length from end of line to tip of arrow
  const int tail = 2; // length from end of line to tails of arrow
  const int wc = 2;   // width of arrow at end of line
  const int wt = 3;   // width of arrow at tails

  Vector* c = new Vector[6];
  c[0] = Vector(0, tip);
  c[1] = Vector(-wc, 0);
  c[2] = Vector(-wt, -tail);
  c[3] = Vector(0, 0);
  c[4] = Vector(wt, -tail);
  c[5] = Vector(wc, 0);

  Matrix m = Translate(0,-tip) * 
    Scale(1.5) * 
    Rotate(-M_PI/2) * 
    Rotate(-d.angle()) * 
    Translate(b);

  for (int i=0; i<6; i++) {
    c[i] = (c[i] * m).round();
  }

  return c;
}

double Marker::mapLenFromRef(double d, CoordSystem sys, SkyFormat format)
{
  FitsImage* ptr = parent->findFits(sys,center);
  return (ptr ? ptr->mapLenFromRef(d,sys,format) : 0);
}

Matrix Marker::fwdRefMatrix()
{
  return Rotate(angle) * flip_ * Translate(center);
}

Matrix Marker::fwdCanvasMatrix()
{
  return Rotate(angle) * flip_ * Translate(center) * parent->refToCanvas;
}

Matrix Marker::bckRefMatrix()
{
  return parent->canvasToRef * Translate(-center) * flip_ * Rotate(-angle);
}

Matrix Marker::bckCanvasMatrix()
{
  return Translate(-center) * flip_ * Rotate(-angle);
}

// list

void Marker::listPre(ostream& str, CoordSystem sys, SkyFrame sky, 
		     FitsImage* ptr, int strip, int hash)
{
  // no props for semicolons
  if (!strip) {
    FitsImage* fits = parent->findFits();
    if (fits && fits->nextMosaic()) {
      switch (sys) {
      case IMAGE:
      case PHYSICAL:
      case DETECTOR:
      case AMPLIFIER:
	str << "# tile " << parent->findFits(ptr) << endl;

	break;
      default:
	if (!parent->findFits()->hasWCSEqu(sys))
	  str << "# tile " << parent->findFits(ptr) << endl;
      }
    }

    if (hash)
      str << "# ";
  }
	
  if (!(properties&INCLUDE))
    str << '-';
}

void Marker::listPost(ostream& str, int conj, int strip)
{
  // no props for semicolons
  if (!strip) {
    if (conj)
      str << " ||";

    listProperties(str, 1);
  }
  else {
    if (conj)
      str << "||";
    else
      str << ';';
  }
}

void Marker::listProperties(ostream& str, int hash)
{
  if (strcmp("green",colorName) ||
      dashlist[0] != 8 ||
      dashlist[1] != 3 ||
      (lineWidth != 1) ||
      strcmp("helvetica 10 normal roman",getFont()) ||
      text[0] ||
      !(properties&SELECT) ||
      !(properties&HIGHLITE) ||
      (properties&DASH) ||
      (properties&FIXED) ||
      !(properties&EDIT) ||
      !(properties&MOVE) ||
      !(properties&ROTATE) ||
      !(properties&DELETE) ||
      !(properties&SOURCE) ||
      (tags.count() > 0) ||
      (comment && *comment)) {

    if (hash)
      str << " #";
    listProps(str);
  }

  str << endl;
}

void Marker::listProps(ostream& str)
{
  if (strcmp("green",colorName))
    str << " color=" << colorName;

  if (dashlist[0] != 8 || dashlist[1] != 3)
    str << " dashlist=" << dashlist[0] << ' ' << dashlist[1];

  if (lineWidth != 1)
    str << " width=" << lineWidth;

  if (strcmp("helvetica 10 normal roman", getFont()))
    str << " font=\"" << getFont() << '"';

  if (text && *text) // only list text if there is something to list
    str << " text={" << text << '}';

  if (!(properties&SELECT))
    str << " select=0";

  if (!(properties&HIGHLITE))
    str << " highlite=0";

  if (properties&DASH)
    str << " dash=1";

  if (properties&FIXED)
    str << " fixed=1";

  if (!(properties&EDIT))
    str << " edit=0";

  if (!(properties&MOVE))
    str << " move=0";

  if (!(properties&ROTATE))
    str << " rotate=0";

  if (!(properties&DELETE))
    str << " delete=0";

  if (!(properties&SOURCE))
    str << " background";

  // tags
  Tag* t = tags.head();
  while (t) {
    str << " tag={" << t->tag() << '}';
    t = t->next();
  }

  if (comment && *comment) 
    str << ' ' << comment;
}

void Marker::listCiaoPre(ostream& str)
{
  if (!(properties&INCLUDE))
    str << '-';
}

void Marker::listCiaoPost(ostream& str, int strip)
			  
{
  str << (strip ? ';' : '\n');
}

void Marker::listProsPost(ostream& str, int strip)
{
  str << (strip ? ';' : '\n');
}

void Marker::listSAOtngPre(ostream& str, int strip)
{
  if (!strip && text[0])
    str << '#' << text << endl;

  if (properties&INCLUDE)
    str << '+';
  else
    str << '-';
}

void Marker::listSAOtngPost(ostream& str, int strip)
{
  if (!strip) {
    str << " # ";
    if (comment && *comment)
      str << comment;
    else if (!(properties&SOURCE))
      str << "background";
    else
      str << colorName;
  }

  str << (strip ? ';' : '\n');
}

void Marker::listSAOimagePre(ostream& str)
{
  if (!(properties&INCLUDE))
    str << '-';
}

void Marker::listSAOimagePost(ostream& str, int strip)
{
  str << (strip ? ';' : '\n');
}

void Marker::listXY(ostream& str, CoordSystem sys, SkyFrame sky,
		    SkyFormat format, int strip)
{
  FitsImage* ptr = parent->findFits();

  switch (sys) {
  case IMAGE:
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    str << setprecision(8) << ptr->mapFromRef(center,sys);
    break;
  default:
    if (ptr->hasWCS(sys)) {
      if (ptr->hasWCSEqu(sys)) {
	switch (format) {
	case DEGREES:
	  str << setprecision(8) << ptr->mapFromRef(center,sys,sky);
	  break;
	case SEXAGESIMAL:
	  {
	    char buf[64];
	    ptr->mapFromRef(center,sys,sky,format,buf,64);
	    char ra[16];
	    char dec[16];
	    string x(buf);
	    istringstream wcs(x);
	    wcs >> ra >> dec;
	    str << ra << ' ' << dec;
	  }
	  break;
	}
      }
      else {
	str << setprecision(8) << ptr->mapFromRef(center,sys);
      }
    }
    break;
  }

  str << (strip ? ';' : '\n');
}

void Marker::XMLRowInit()
{
  for (int ii=0; ii<XMLNUMCOL; ii++) {
    if (XMLCol[ii])
      delete [] XMLCol[ii];
    XMLCol[ii] = NULL;
  }
}

void Marker::XMLRow(XMLColName col, int val)
{
  ostringstream str;
  str << val << ends;

  if (XMLCol[col])
    delete [] XMLCol[col];
  XMLCol[col] = dupstr(str.str().c_str());
}

void Marker::XMLRow(XMLColName col, int* val, int cnt)
{
  ostringstream str;
  for (int ii=0; ii<cnt; ii++) {
    str << val[ii];
    if (ii!=cnt-1)
      str << ' ';
    else
      str << ends;
  }

  if (XMLCol[col])
    delete [] XMLCol[col];
  XMLCol[col] = dupstr(str.str().c_str());
}

void Marker::XMLRow(XMLColName col, double val)
{
  ostringstream str;
  str << setprecision(8) << val << ends;

  if (XMLCol[col])
    delete [] XMLCol[col];
  XMLCol[col] = dupstr(str.str().c_str());
}

void Marker::XMLRow(XMLColName col, double* val, int cnt)
{
  ostringstream str;
  for (int ii=0; ii<cnt; ii++) {
    str << setprecision(8) << val[ii];
    if (ii!=cnt-1)
      str << ' ';
    else
      str << ends;
  }

  if (XMLCol[col])
    delete [] XMLCol[col];
  XMLCol[col] = dupstr(str.str().c_str());
}

void Marker::XMLRow(XMLColName col, char* val)
{
  if (XMLCol[col])
    delete [] XMLCol[col];
  XMLCol[col] = dupstr(val);
}

void Marker::XMLRow(XMLColName col, char** val, int cnt)
{
  ostringstream str;
  for (int ii=0; ii<cnt; ii++) {
    str << val[ii];
    if (ii!=cnt-1)
      str << ' ';
    else
      str << ends;
  }

  if (XMLCol[col])
    delete [] XMLCol[col];
  XMLCol[col] = dupstr(str.str().c_str());
}

void Marker::XMLRowProps(FitsImage* ptr, CoordSystem sys)
{
  // tile
  {
    ostringstream str;
    XMLColName col = XMLTILE;

    FitsImage* fits = parent->findFits();
    if (fits && fits->nextMosaic()) {
      switch (sys) {
      case IMAGE:
      case PHYSICAL:
      case DETECTOR:
      case AMPLIFIER:
	str << parent->findFits(ptr) << ends;
	break;
      default:
	if (!parent->findFits()->hasWCSEqu(sys))
	  str << parent->findFits(ptr) << ends;
	break;
      }
    }
    if (XMLCol[col])
      delete [] XMLCol[col];
    XMLCol[col] = dupstr(str.str().c_str());
  }

  // color
  {
    XMLColName col = XMLCOLOR;
    if (XMLCol[col])
      delete [] XMLCol[col];
    XMLCol[col] = dupstr(colorName);
  }

  // width
  {
    ostringstream str;
    str << lineWidth << ends;
    XMLColName col = XMLWIDTH;
    if (XMLCol[col])
      delete [] XMLCol[col];
    XMLCol[col] = dupstr(str.str().c_str());
  }

  // text
  if (text && *text) {
    XMLColName col = XMLTEXT;
    if (XMLCol[col])
      delete [] XMLCol[col];
    XMLCol[col] = dupstr(text);
  }

  // font
  {
    XMLColName col = XMLFONT;
    if (XMLCol[col])
      delete [] XMLCol[col];
    XMLCol[col] = dupstr(getFont());
  }

  XMLRowProp(XMLSELECT,SELECT);
  XMLRowProp(XMLHIGHLITE,HIGHLITE);
  XMLRowProp(XMLEDIT,EDIT);
  XMLRowProp(XMLMOVE,MOVE);
  XMLRowProp(XMLROTATE,ROTATE);
  XMLRowProp(XMLDELETE,DELETE);
  XMLRowProp(XMLFIXED,FIXED);
  XMLRowProp(XMLINCLUDE,INCLUDE);
  XMLRowProp(XMLSOURCE,SOURCE);
  XMLRowProp(XMLDASH,DASH);

  // dashlist
  {
    ostringstream str;
    str << dashlist[0] << ',' << dashlist[1] << ends;
    XMLColName col = XMLDASHLIST;
    if (XMLCol[col])
      delete [] XMLCol[col];
    XMLCol[col] = dupstr(str.str().c_str());
  }

  // tags
  {
    ostringstream str;
    Tag* start = tags.head();
    Tag* ptr = start;
    while (ptr) {
      if (ptr != start)
	str << ' ';
      char* tag = (char*)ptr->tag();
      while (*tag) {
	if (*tag == ' ')
	  str << "&#160;";
	else
	  str << *tag;
	tag++;
      }
      ptr = ptr->next();
    }
    str << ends;

    XMLColName col = XMLTAG;
    if (XMLCol[col])
      delete [] XMLCol[col];
    XMLCol[col] = dupstr(str.str().c_str());
  }

  // comment
  if (comment && *comment) {
    XMLColName col = XMLCOMMENT;
    if (XMLCol[col])
      delete [] XMLCol[col];
    XMLCol[col] = dupstr(comment);
  }
}

void Marker::XMLRowEnd(ostream& str)
{
  str << "<TR>";
  for (int ii=0; ii<XMLNUMCOL; ii++) {
    str << "<TD>";
    if (XMLCol[ii]) {
      char* ss = XMLQuote(XMLCol[ii]);
      str << ss;
      delete [] XMLCol[ii];
      delete [] ss;
      XMLCol[ii] = NULL;
    }
    str << "</TD>";
  }
  str << "</TR>" << endl;
}

void Marker::XMLRowProp(XMLColName col, Property prop)
{
  ostringstream str;
  if (properties & prop)
      str << "1";
    else
      str << "0" << ends;

  XMLCol[col] = dupstr(str.str().c_str());
}

void Marker::XMLRowPoint(FitsImage* ptr, CoordSystem sys, SkyFrame sky, 
			 SkyFormat format, Vector vv)
{
  switch (sys) {
  case IMAGE:
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    {
      Vector v = ptr->mapFromRef(vv,sys);
      XMLRow(XMLX,v[0]);
      XMLRow(XMLY,v[1]);
    }
    break;
  default:
    if (ptr->hasWCS(sys)) {
      if (ptr->hasWCSEqu(sys)) {
	switch (format) {
	case DEGREES:
	  {
	    Vector v = ptr->mapFromRef(vv,sys,sky);
	    XMLRow(XMLX,v[0]);
	    XMLRow(XMLY,v[1]);
	  }
	  break;
	case SEXAGESIMAL:
	  {
	    char buf[64];
	    ptr->mapFromRef(vv,sys,sky,format,buf,64);
	    char ra[16];
	    char dec[16];
	    string x(buf);
	    istringstream wcs(x);
	    wcs >> ra >> dec;

	    XMLRow(XMLX,ra);
	    XMLRow(XMLY,dec);
	  }
	  break;
	}
      }
      else {
	Vector v = ptr->mapFromRef(vv,sys);
	XMLRow(XMLX,v[0]);
	XMLRow(XMLY,v[1]);
      }
    }
  }
}

void Marker::XMLRowPoint(FitsImage* ptr, CoordSystem sys, SkyFrame sky, 
			 SkyFormat format, Vector* vv, int cnt)
{
  switch (sys) {
  case IMAGE:
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    {
      double xx[cnt];
      double yy[cnt];
      for (int ii=0; ii<cnt; ii++) {
	Vector v = ptr->mapFromRef(vv[ii],sys);
	xx[ii] = v[0];
	yy[ii] = v[1];
      }
      XMLRow(XMLXV,xx,cnt);
      XMLRow(XMLYV,yy,cnt);
    }
    break;
  default:
    if (ptr->hasWCS(sys)) {
      if (ptr->hasWCSEqu(sys)) {
	switch (format) {
	case DEGREES:
	  {
	    double xx[cnt];
	    double yy[cnt];
	    for (int ii=0; ii<cnt; ii++) {
	      Vector v = ptr->mapFromRef(vv[ii],sys,sky);
	      xx[ii] = v[0];
	      yy[ii] = v[1];
	    }
	    XMLRow(XMLXV,xx,cnt);
	    XMLRow(XMLYV,yy,cnt);
	  }
	  break;
	case SEXAGESIMAL:
	  {
	    char* xx[cnt];
	    char* yy[cnt];
	    for (int ii=0; ii<cnt; ii++) {
	      char buf[64];
	      ptr->mapFromRef(vv[ii],sys,sky,format,buf,64);
	      char ra[16];
	      char dec[16];
	      string x(buf);
	      istringstream wcs(x);
	      wcs >> ra >> dec;

	      xx[ii] = dupstr(ra);
	      yy[ii] = dupstr(dec);
	    }
	    XMLRow(XMLXV,xx,cnt);
	    XMLRow(XMLYV,yy,cnt);

	    for (int ii=0; ii<cnt; ii++) {
	      delete [] xx[ii];
	      delete [] yy[ii];
	    }
	  }
	  break;
	}
      }
      else {
	double xx[cnt];
	double yy[cnt];
	for (int ii=0; ii<cnt; ii++) {
	  Vector v = ptr->mapFromRef(vv[ii],sys);
	  xx[ii] = v[0];
	  yy[ii] = v[1];
	}
	XMLRow(XMLXV,xx,cnt);
	XMLRow(XMLYV,yy,cnt);
      }
    }
  }
}

void Marker::XMLRowRadiusX(FitsImage* ptr, CoordSystem sys, Vector vv)
{
  double rr = ptr->mapLenFromRef(vv[0],sys,ARCSEC);
  XMLRow(XMLR,rr);
}

void Marker::XMLRowRadiusX(FitsImage* ptr, CoordSystem sys, Vector* vv, int cnt)
{
  double rr[cnt];
  for (int ii=0; ii<cnt; ii++)
    rr[ii] = ptr->mapLenFromRef(vv[ii][0],sys,ARCSEC);

  XMLRow(XMLRV,rr,cnt);
}

void Marker::XMLRowRadius(FitsImage* ptr, CoordSystem sys, Vector vv)
{
  Vector v = ptr->mapLenFromRef(vv,sys,ARCSEC);
  XMLRow(XMLR,v[0]);
  XMLRow(XMLR2,v[1]);
}

void Marker::XMLRowRadius(FitsImage* ptr, CoordSystem sys, Vector* vv, int cnt)
{
  double rr[cnt];
  double rr2[cnt];
  for (int ii=0; ii<cnt; ii++) {
    Vector v = ptr->mapLenFromRef(vv[ii],sys,ARCSEC);
    rr[ii] = v[0];
    rr2[ii] = v[1];
  }

  XMLRow(XMLRV,rr,cnt);
  XMLRow(XMLRV2,rr2,cnt);
}

void Marker::XMLRowAng(CoordSystem sys, SkyFrame sky)
{
  XMLRow(XMLANG, radToDeg(parent->mapAngleFromRef(angle,sys,sky)));
}

void Marker::XMLRowAng(CoordSystem sys, SkyFrame sky, double* ang, int cnt)
{
  double aa[cnt];
  for (int ii=0; ii<cnt; ii++)
    aa[ii] = radToDeg(parent->mapAngleFromRef(ang[ii],sys,sky));

  XMLRow(XMLANGV,aa,cnt);
}

char* Marker::XMLQuote(char* src)
{
  char* dest = new char[strlen(src)*7+1];
  char* sptr = src;
  char* dptr = dest;
  while (sptr && *sptr) {
    if (*sptr == '&') {
      // special case, char entities
      if (*(sptr+1) == '#') {
	*dptr++ = *sptr;
      }
      else {
	memcpy(dptr,"&amp;",5);
	dptr += 5;
      }
    }
    else if (*sptr == '<') {
      memcpy(dptr,"&lt;",4);
      dptr += 4;
    }
    else if (*sptr == '>') {
      memcpy(dptr,"&gt;",4);
      dptr += 4;
    }
    else if (*sptr == '\'') {
      memcpy(dptr,"&apos;",6);
      dptr += 6;
    }
    else if (*sptr == '"') {
      memcpy(dptr,"&quot;",6);
      dptr += 6;
    }
    else
     *dptr++ = *sptr;

    sptr++;
  }
  *dptr = NULL;

  return dest;
}

// special composite funtionallity

void Marker::setComposite(const Matrix& mx, double aa)
{
  center *= mx;
  angle += aa;
  updateBBox();
}

void Marker::setComposite(unsigned short p, const char* clr, int w, int h)
{
  properties = p;
  lineWidth = w;
  if (colorName)
    delete [] colorName;

  colorName = dupstr(clr);
  color = parent->getColor(colorName);
  highlited = h;
}

