// Copyright (C) 1999-2010
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "point.h"
#include "framebase.h"
#include "fitsimage.h"

Point::Point(const Point& a) : Marker(a)
{
  strncpy(pointtype,a.pointtype,16);
  size_ = a.size_;
}

Point::Point(FrameBase* p, const Vector& ctr, int size,
	     const char* clr, int* dsh, 
	     int wth, const char* fnt, const char* txt,
	     unsigned short prop, const char* cmt,
	     const List<Tag>& tg, const List<CallBack>& cb)
  : Marker(p, ctr, 0, clr, dsh, wth, fnt, txt, prop, cmt, tg, cb)
{
  pointtype[0] = '\0';
  size_ = size;

  handle = new Vector[4];
  numHandle = 4;

  updateBBox();
}

void Point::updateBBox()
{
  // bound marker

  Vector h = Vector(size_, size_)/2;
  Vector c = center * parent->refToCanvas;
  bbox = BBox(c-h,c+h);

  // generate handles

  handle[0] = bbox.ll;
  handle[1] = bbox.lr();
  handle[2] = bbox.ur;
  handle[3] = bbox.ul();

  // make room for handles

  bbox.expand(3);

  // calculate overall bbox

  calcAllBBox();
}

void Point::setSize(int size)
{
  size_ = size;
  updateBBox();
  doCallBack(CallBack::EDITCB);
}

void Point::renderXLineDash(GC lgc)
{
  char dlist[2];
  dlist[0] = ceil(dashlist[0]/2.)+'0';
  dlist[1] = ceil(dashlist[1]/2.)+'0';
  int ww = (highlited && canHighlite()) ? lineWidth*2 : lineWidth;
  XSetDashes(display, lgc, 0, dlist, 2);
  XSetLineAttributes(display, lgc, ww, LineOnOffDash, CapButt, JoinMiter);
}

void Point::renderXInclude(Drawable drawable, const Matrix& mx, 
			   RenderMode mode)
{
  if (!(properties & INCLUDE)) {
    Vector r1 = (center*mx - Vector(size_,size_)/2).round();
    Vector r2 = (center*mx + Vector(size_,size_)/2).round();

    GC lgc = renderXGC(mode);
    if (mode == SRC)
      XSetForeground(display, gc, parent->getColor("red"));

    XDrawLine(display, drawable, lgc, r1[0], r1[1], r2[0], r2[1]);    
  }
}

void Point::renderPSLineDash()
{
  ostringstream str;
  str << lineWidth << " setlinewidth" << endl;
  str << "[4 2] 0 setdash" << endl << ends;
  Tcl_AppendResult(parent->interp, (char*)str.str().c_str(), NULL);
}

void Point::renderPSInclude(int mode)
{
  if (!(properties & INCLUDE)) {
    Matrix m = parent->refToCanvas;
    Vector r1 = center*m - Vector(size_,size_)/2;
    Vector r2 = center*m + Vector(size_,size_)/2;

    renderPSColor(mode, parent->getXColor("red"));

    ostringstream str;
    str << "newpath " 
	<< r1.TkCanvasPs(parent->canvas) << " moveto" << endl
	<< r2.TkCanvasPs(parent->canvas) << " lineto" << endl
	<< " stroke" << endl << ends;
    Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);
  }
}

#ifdef _MACOSX
void Point::renderMACOSXLineDash()
{
  macosxWidth(lineWidth);
  float dlist[2] = {4,2};
  macosxDash(dlist,2);
}

void Point::renderMACOSXInclude()
{
  if (!(properties & INCLUDE)) {
    Matrix m = parent->refToCanvas;
    Vector r1 = center*m - Vector(size_,size_)/2;
    Vector r2 = center*m + Vector(size_,size_)/2;

    macosxColor(parent->getXColor("red"));
    macosxDrawLine(r1,r2);
  }
}
#endif

#ifdef _WIN32
void Point::renderWIN32LineDash()
{
  win32Width(lineWidth);
  float dlist[2] = {4,2};
  win32Dash(dlist,2);
}

void Point::renderWIN32Include()
{
  if (!(properties & INCLUDE)) {
    Matrix m = parent->refToCanvas;
    Vector r1 = center*m - Vector(size_,size_)/2;
    Vector r2 = center*m + Vector(size_,size_)/2;

    win32Color(parent->getXColor("red"));
    win32DrawLine(r1,r2);
  }
}
#endif

CirclePoint::CirclePoint(const CirclePoint& a) : Point(a) {}

CirclePoint::CirclePoint(FrameBase* p, const Vector& ctr, int size,
			 const char* clr, int* dsh, 
			 int wth, const char* fnt, const char* txt,
			 unsigned short prop, const char* cmt,
			 const List<Tag>& tg, const List<CallBack>& cb)
  : Point(p, ctr, size, clr, dsh, wth, fnt, txt, prop, cmt, tg, cb)
{
  strcpy(type,"circle point");
  strcpy(pointtype,"circle");
}

void CirclePoint::renderX(Drawable drawable, const Matrix& mx, RenderMode mode)
{
  GC lgc = renderXGC(mode);

  Vector v = (center * mx * Translate(-Vector(size_,size_)/2)).round();
#ifndef _WIN32
  XDrawArc(display, drawable, lgc, v[0], v[1], size_, size_, 0, 360*64);
#else
  int ss = int(size_/2.+.5)*2;
  XDrawArc(display, drawable, lgc, v[0], v[1], ss, ss, 0, 360*64);
#endif

  renderXInclude(drawable, mx, mode);
}

void CirclePoint::renderPS(int mode)
{
  renderPSGC(mode);

  Vector cc = center  * parent->refToCanvas;
  ostringstream str;
  str << "newpath " << cc.TkCanvasPs(parent->canvas) << size_/2.
      << " 0 360 arc stroke" << endl << ends;
  Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);

  renderPSInclude(mode);
}

#ifdef _MACOSX
void CirclePoint::renderMACOSX()
{
  renderMACOSXGC();

  Vector cc = center  * parent->refToCanvas;
  macosxDrawArc(cc, size_/2., 0, 2*M_PI);
  renderMACOSXInclude();
}
#endif

#ifdef _WIN32
void CirclePoint::renderWIN32()
{
  renderWIN32GC();

  Vector cc = center  * parent->refToCanvas;
  win32DrawArc(cc, size_/2., 0, 2*M_PI);
  renderWIN32Include();
}
#endif

BoxPoint::BoxPoint(const BoxPoint& a) : Point(a) {}

BoxPoint::BoxPoint(FrameBase* p, const Vector& ctr, int size,
		   const char* clr, int* dsh, 
		   int wth, const char* fnt, const char* txt,
		   unsigned short prop, const char* cmt,
		   const List<Tag>& tg, const List<CallBack>& cb)
  : Point(p, ctr, size, clr, dsh, wth, fnt, txt, prop, cmt, tg, cb)
{
  strcpy(type, "box point");
  strcpy(pointtype, "box");
}

void BoxPoint::renderX(Drawable drawable, const Matrix& mx, RenderMode mode)
{
  GC lgc = renderXGC(mode);

  Vector v = (center * mx * Translate(-Vector(size_,size_)/2)).round();
  XDrawRectangle(display, drawable, lgc, v[0], v[1], size_, size_);

  renderXInclude(drawable, mx, mode);
}

void BoxPoint::renderPS(int mode)
{
  renderPSGC(mode);

  Vector v = center*parent->refToCanvas - Vector(size_,-size_)/2;
  ostringstream str;
  str << "newpath " << v.TkCanvasPs(parent->canvas) 
      << Vector(size_,size_)
      << " rectstroke" << endl << ends;
  Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);

  renderPSInclude(mode);
}

#ifdef _MACOSX
void BoxPoint::renderMACOSX()
{
  renderMACOSXGC();

  Vector v = center*parent->refToCanvas - Vector(size_,-size_)/2;
  macosxDrawRect(v, Vector(size_,size_));
  renderMACOSXInclude();
}
#endif

#ifdef _WIN32
void BoxPoint::renderWIN32()
{
  renderWIN32GC();

  Vector v = center*parent->refToCanvas - Vector(size_,-size_)/2;
  win32DrawRect(v, Vector(size_,size_));
  renderWIN32Include();
}
#endif

DiamondPoint::DiamondPoint(const DiamondPoint& a) : Point(a) {}

DiamondPoint::DiamondPoint(FrameBase* p, const Vector& ctr, int size,
			   const char* clr, int* dsh, 
			   int wth, const char* fnt, const char* txt,
			   unsigned short prop, const char* cmt,
			   const List<Tag>& tg, const List<CallBack>& cb)
  : Point(p, ctr, size, clr, dsh, wth, fnt, txt, prop, cmt, tg, cb)
{
  strcpy(type, "diamond point");
  strcpy(pointtype, "diamond");
}

void DiamondPoint::renderX(Drawable drawable, const Matrix& mx, RenderMode mode)

{
  GC lgc = renderXGC(mode);

  int l = size_-1;
  Vector cc = center * mx;
  Vector v1 = (cc - Vector(l,0)/2).round();
  Vector v2 = (cc - Vector(0,l)/2).round();
  Vector v3 = (cc + Vector(l,0)/2).round();
  Vector v4 = (cc + Vector(0,l)/2).round();

  XPoint p[5];
  p[0].x = (short)v1[0];
  p[0].y = (short)v1[1];
  p[1].x = (short)v2[0];
  p[1].y = (short)v2[1];
  p[2].x = (short)v3[0];
  p[2].y = (short)v3[1];
  p[3].x = (short)v4[0];
  p[3].y = (short)v4[1];
  p[4] = p[0];

  XDrawLines(display, drawable, lgc, p, 5, CoordModeOrigin);

  renderXInclude(drawable, mx, mode);
}

void DiamondPoint::renderPS(int mode)
{
  renderPSGC(mode);

  Matrix m = parent->refToCanvas;

  Vector v1 = center*m - Vector(size_, 0)/2;
  Vector v2 = center*m - Vector(0, size_)/2;
  Vector v3 = center*m + Vector(size_, 0)/2;
  Vector v4 = center*m + Vector(0, size_)/2;

  ostringstream str;
  str << "newpath " << v1.TkCanvasPs(parent->canvas) << " moveto" << endl
      << v2.TkCanvasPs(parent->canvas) << " lineto" << endl
      << v3.TkCanvasPs(parent->canvas) << " lineto" << endl
      << v4.TkCanvasPs(parent->canvas) << " lineto" << endl
      << "closepath stroke" << endl
      << ends;
  Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);

  renderPSInclude(mode);
}

#ifdef _MACOSX
void DiamondPoint::renderMACOSX()
{
  renderMACOSXGC();

  Matrix m = parent->refToCanvas;

  Vector vv[5];
  vv[0] = center*m - Vector(size_, 0)/2;
  vv[1] = center*m - Vector(0, size_)/2;
  vv[2] = center*m + Vector(size_, 0)/2;
  vv[3] = center*m + Vector(0, size_)/2;
  vv[4] = vv[0];

  macosxDrawLines(vv,5);
  renderMACOSXInclude();
}
#endif

#ifdef _WIN32
void DiamondPoint::renderWIN32()
{
  renderWIN32GC();

  Matrix m = parent->refToCanvas;

  Vector vv[5];
  vv[0] = center*m - Vector(size_, 0)/2;
  vv[1] = center*m - Vector(0, size_)/2;
  vv[2] = center*m + Vector(size_, 0)/2;
  vv[3] = center*m + Vector(0, size_)/2;
  vv[4] = vv[0];

  win32DrawLines(vv,5);
  renderWIN32Include();
}
#endif

CrossPoint::CrossPoint(const CrossPoint& a) : Point(a) {}

CrossPoint::CrossPoint(FrameBase* p, const Vector& ctr, int size,
		       const char* clr, int* dsh, 
		       int wth, const char* fnt, const char* txt,
		       unsigned short prop, const char* cmt,
		       const List<Tag>& tg, const List<CallBack>& cb)
  : Point(p, ctr, size, clr, dsh, wth, fnt, txt, prop, cmt, tg, cb)
{
  strcpy(type, "cross point");
  strcpy(pointtype, "cross");
}

void CrossPoint::renderX(Drawable drawable, const Matrix& mx, RenderMode mode)
{
  GC lgc = renderXGC(mode);

  Vector cc = center * mx;
  Vector v1 = (cc - Vector(size_,0)/2).round();
  Vector v2 = (cc + Vector(size_,0)/2).round();
  Vector v3 = (cc - Vector(0,size_)/2).round();
  Vector v4 = (cc + Vector(0,size_)/2).round();

  XDrawLine(display, drawable, lgc, v1[0], v1[1], v2[0], v2[1]);
  XDrawLine(display, drawable, lgc, v3[0], v3[1], v4[0], v4[1]);

  renderXInclude(drawable, mx, mode);
}

void CrossPoint::renderPS(int mode)
{
  renderPSGC(mode);

  Matrix m = parent->refToCanvas;

  Vector v1 = center*m - Vector(size_,0)/2;
  Vector v2 = center*m + Vector(size_,0)/2;
  Vector v3 = center*m - Vector(0,size_)/2;
  Vector v4 = center*m + Vector(0,size_)/2;

  ostringstream str;
  str << "newpath " << v1.TkCanvasPs(parent->canvas) << " moveto" << endl
      << v2.TkCanvasPs(parent->canvas) << " lineto stroke" << endl
      << "newpath " << v3.TkCanvasPs(parent->canvas) << " moveto" << endl
      << v4.TkCanvasPs(parent->canvas) << " lineto stroke" << endl
      << ends;
  Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);

  renderPSInclude(mode);
}

#ifdef _MACOSX
void CrossPoint::renderMACOSX()
{
  renderMACOSXGC();

  Matrix m = parent->refToCanvas;

  Vector v1 = center*m - Vector(size_,0)/2;
  Vector v2 = center*m + Vector(size_,0)/2;
  Vector v3 = center*m - Vector(0,size_)/2;
  Vector v4 = center*m + Vector(0,size_)/2;

  macosxDrawLine(v1,v2);
  macosxDrawLine(v3,v4);

  renderMACOSXInclude();
}
#endif

#ifdef _WIN32
void CrossPoint::renderWIN32()
{
  renderWIN32GC();

  Matrix m = parent->refToCanvas;

  Vector v1 = center*m - Vector(size_,0)/2;
  Vector v2 = center*m + Vector(size_,0)/2;
  Vector v3 = center*m - Vector(0,size_)/2;
  Vector v4 = center*m + Vector(0,size_)/2;

  win32DrawLine(v1,v2);
  win32DrawLine(v3,v4);

  renderWIN32Include();
}
#endif

ExPoint::ExPoint(const ExPoint& a) : Point(a) {}

ExPoint::ExPoint(FrameBase* p, const Vector& ctr, int size,
		 const char* clr, int* dsh, 
		 int wth, const char* fnt, const char* txt,
		 unsigned short prop, const char* cmt,
		 const List<Tag>& tg, const List<CallBack>& cb)
  : Point(p, ctr, size, clr, dsh, wth, fnt, txt, prop, cmt, tg, cb)
{
  strcpy(type, "x point");
  strcpy(pointtype, "x");
}

void ExPoint::renderX(Drawable drawable, const Matrix& mx, RenderMode mode)
{
  GC lgc = renderXGC(mode);

  Vector cc = center * mx;
  Vector v1 = (cc + Vector(-size_,-size_)/2).round();
  Vector v2 = (cc + Vector( size_,-size_)/2).round();
  Vector v3 = (cc + Vector( size_, size_)/2).round();
  Vector v4 = (cc + Vector(-size_, size_)/2).round();

  XDrawLine(display, drawable, lgc, v1[0], v1[1], v3[0], v3[1]);
  XDrawLine(display, drawable, lgc, v2[0], v2[1], v4[0], v4[1]);

  renderXInclude(drawable, mx, mode);
}

void ExPoint::renderPS(int mode)
{
  renderPSGC(mode);

  Matrix m = parent->refToCanvas;

  Vector v1 = center*m + Vector(-size_, -size_)/2;
  Vector v2 = center*m + Vector(size_, -size_)/2;
  Vector v3 = center*m + Vector(size_, size_)/2;
  Vector v4 = center*m + Vector(-size_, size_)/2;

  ostringstream str;
  str << "newpath " << v1.TkCanvasPs(parent->canvas) << " moveto" << endl
      << v3.TkCanvasPs(parent->canvas) << " lineto stroke" << endl
      << "newpath " << v2.TkCanvasPs(parent->canvas) << " moveto" << endl
      << v4.TkCanvasPs(parent->canvas) << " lineto stroke" << endl
      << ends;
  Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);

  renderPSInclude(mode);
}

#ifdef _MACOSX
void ExPoint::renderMACOSX()
{
  renderMACOSXGC();

  Matrix m = parent->refToCanvas;

  Vector v1 = center*m + Vector(-size_, -size_)/2;
  Vector v2 = center*m + Vector(size_, -size_)/2;
  Vector v3 = center*m + Vector(size_, size_)/2;
  Vector v4 = center*m + Vector(-size_, size_)/2;

  macosxDrawLine(v1,v3);
  macosxDrawLine(v2,v4);

  renderMACOSXInclude();
}
#endif

#ifdef _WIN32
void ExPoint::renderWIN32()
{
  renderWIN32GC();

  Matrix m = parent->refToCanvas;

  Vector v1 = center*m + Vector(-size_, -size_)/2;
  Vector v2 = center*m + Vector(size_, -size_)/2;
  Vector v3 = center*m + Vector(size_, size_)/2;
  Vector v4 = center*m + Vector(-size_, size_)/2;

  win32DrawLine(v1,v3);
  win32DrawLine(v2,v4);

  renderWIN32Include();
}
#endif

ArrowPoint::ArrowPoint(const ArrowPoint& a) : Point(a) {}

ArrowPoint::ArrowPoint(FrameBase* p, const Vector& ctr, int size,
		       const char* clr, int* dsh, 
		       int wth, const char* fnt, const char* txt,
		       unsigned short prop, const char* cmt,
		       const List<Tag>& tg, const List<CallBack>& cb)
  : Point(p, ctr, size, clr, dsh, wth, fnt, txt, prop, cmt, tg, cb)
{
  strcpy(type, "arrow point");
  strcpy(pointtype, "arrow");
}

void ArrowPoint::renderX(Drawable drawable, const Matrix& mx, RenderMode mode)
{
  GC lgc = renderXGC(mode);

  Vector cc = center * mx;
  Vector v1 = (cc + Vector(-size_, 0)/2).round();
  Vector v2 = (cc + Vector(0, -size_)/2).round();
  Vector v3 = (cc - Vector(size_, size_)/2).round();
  Vector v4 = cc.round();

  XDrawLine(display, drawable, lgc, v1[0], v1[1], v4[0], v4[1]);
  XDrawLine(display, drawable, lgc, v2[0], v2[1], v4[0], v4[1]);
  XDrawLine(display, drawable, lgc, v3[0], v3[1], v4[0], v4[1]);

  renderXInclude(drawable, mx, mode);
}

void ArrowPoint::renderPS(int mode)
{
  renderPSGC(mode);

  Matrix m = parent->refToCanvas;

  Vector v1 = center*m + Vector(-size_, 0)/2;
  Vector v2 = center*m + Vector(0, -size_)/2;
  Vector v3 = center*m - Vector(size_, size_)/2;
  Vector v4 = center*m;

  ostringstream str;
  str << "newpath " << v1.TkCanvasPs(parent->canvas) << " moveto" << endl
      << v4.TkCanvasPs(parent->canvas) << " lineto stroke" << endl
      << "newpath " << v2.TkCanvasPs(parent->canvas) << " moveto" << endl
      << v4.TkCanvasPs(parent->canvas) << " lineto stroke" << endl
      << "newpath " << v3.TkCanvasPs(parent->canvas) << " moveto" << endl
      << v4.TkCanvasPs(parent->canvas) << " lineto stroke" << endl
      << ends;
  Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);

  renderPSInclude(mode);
}

#ifdef _MACOSX
void ArrowPoint::renderMACOSX()
{
  renderMACOSXGC();

  Matrix m = parent->refToCanvas;

  Vector v1 = center*m + Vector(-size_, 0)/2;
  Vector v2 = center*m + Vector(0, -size_)/2;
  Vector v3 = center*m - Vector(size_, size_)/2;
  Vector v4 = center*m;

  macosxDrawLine(v1,v4);
  macosxDrawLine(v2,v4);
  macosxDrawLine(v3,v4);

  renderMACOSXInclude();
}
#endif

#ifdef _WIN32
void ArrowPoint::renderWIN32()
{
  renderWIN32GC();

  Matrix m = parent->refToCanvas;

  Vector v1 = center*m + Vector(-size_, 0)/2;
  Vector v2 = center*m + Vector(0, -size_)/2;
  Vector v3 = center*m - Vector(size_, size_)/2;
  Vector v4 = center*m;

  win32DrawLine(v1,v4);
  win32DrawLine(v2,v4);
  win32DrawLine(v3,v4);

  renderWIN32Include();
}
#endif

BoxCirclePoint::BoxCirclePoint(const BoxCirclePoint& a) : Point(a) {}

BoxCirclePoint::BoxCirclePoint(FrameBase* p, const Vector& ctr, int size,
			       const char* clr, int* dsh, 
			       int wth, const char* fnt, const char* txt,
			       unsigned short prop, const char* cmt,
			       const List<Tag>& tg, const List<CallBack>& cb)
  : Point(p, ctr, size, clr, dsh, wth, fnt, txt, prop, cmt, tg, cb)
{
  strcpy(type, "boxcircle point");
  strcpy(pointtype, "boxcircle");
}

void BoxCirclePoint::renderX(Drawable drawable, const Matrix& mx,
			     RenderMode mode)
{
  GC lgc = renderXGC(mode);

  Vector cc = center * mx;

#ifndef _WIN32
  Vector v1 = (cc * Translate(-Vector(size_-2,size_-2)/2)).round();
  Vector v2 = (cc * Translate(-Vector(size_,size_)/2)).round();
  XDrawArc(display, drawable, lgc, v1[0], v1[1], size_-2, size_-2, 0, 
	   360*64);
  XDrawRectangle(display, drawable, lgc, v2[0], v2[1], size_, size_);
#else
  int ss = int(size_/2.+.5)*2;
  Vector v1 = (cc * Translate(-Vector(ss-2,ss-2)/2)).round();
  Vector v2 = (cc * Translate(-Vector(ss,ss)/2)).round();
  XDrawArc(display, drawable, lgc, v1[0], v1[1], ss-2, ss-2, 0, 360*64);
  XDrawRectangle(display, drawable, lgc, v2[0], v2[1], ss, ss);
#endif



  renderXInclude(drawable, mx, mode);
}

void BoxCirclePoint::renderPS(int mode)
{
  renderPSGC(mode);

  Vector v1 = center*parent->refToCanvas;
  Vector v2 = center*parent->refToCanvas - Vector(size_,-size_)/2;

  ostringstream str;
  str << "newpath " << v1.TkCanvasPs(parent->canvas) << (size_-2)/2.
      << " 0 360 arc stroke" << endl
      << "newpath " << v2.TkCanvasPs(parent->canvas) 
      << Vector(size_,size_) 
      << " rectstroke" << endl
      << ends;
  Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);

  renderPSInclude(mode);
}

#ifdef _MACOSX
void BoxCirclePoint::renderMACOSX()
{
  renderMACOSXGC();

  Vector v1 = center*parent->refToCanvas;
  Vector v2 = center*parent->refToCanvas - Vector(size_,-size_)/2;

  macosxDrawArc(v1, (size_-2)/2., 0, 2*M_PI);
  macosxDrawRect(v2, Vector(size_,size_));

  renderMACOSXInclude();
}
#endif

#ifdef _WIN32
void BoxCirclePoint::renderWIN32()
{
  renderWIN32GC();

  Vector v1 = center*parent->refToCanvas;
  Vector v2 = center*parent->refToCanvas - Vector(size_,-size_)/2;

  win32DrawArc(v1, (size_-2)/2., 0, 2*M_PI);
  win32DrawRect(v2, Vector(size_,size_));

  renderWIN32Include();
}
#endif

// list

void Point::list(ostream& str, CoordSystem sys, SkyFrame sky, 
		 SkyFormat format, int conj, int strip)
{
  FitsImage* ptr = parent->findFits(sys,center);
  listPre(str, sys, sky, ptr, strip, 0);

  switch (sys) {
  case IMAGE:
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    {
      Vector v = ptr->mapFromRef(center,sys);
      str << "point" << '(' << setprecision(8) << v[0] << ',' << v[1] << ')';
    }
    break;
  default:
    if (ptr->hasWCS(sys)) {
      if (ptr->hasWCSEqu(sys)) {
	switch (format) {
	case DEGREES:
	  {
	    Vector v = ptr->mapFromRef(center,sys,sky);
	    str << "point" << '(' << setprecision(8) 
		<< v[0] << ',' << v[1] << ')';
	  }
	  break;
	case SEXAGESIMAL:
	  {
	    char buf[64];
	    ptr->mapFromRef(center,sys,sky,format,buf,64);
	    char ra[16];
	    char dec[16];
	    string x(buf);
	    istringstream wcs(x);
	    wcs >> ra >> dec;
	    str << "point" << '(' << ra << ',' << dec << ')';
	  }
	  break;
	}
      }
      else {
	Vector v = ptr->mapFromRef(center,sys);
	str << "point" << '(' << setprecision(8) << v[0] << ',' << v[1] << ')';
      }
    }
  }

  listPost(str, conj, strip);
}

void Point::listPost(ostream& str, int conj, int strip)
{
  // no props for semicolons
  if (!strip) {
    if (conj)
      str << " ||";

    str << " # point=" << pointtype;
    if (size_ != POINTSIZE)
      str << ' ' << size_;
    listProperties(str, 0);
  }
  else {
    if (conj)
      str << "||";
    else
      str << ';';
  }
}

void Point::listXML(ostream& str, CoordSystem sys, SkyFrame sky, 
		    SkyFormat format)
{
  FitsImage* ptr = parent->findFits(sys,center);
  
  XMLRowInit();
  XMLRow(XMLSHAPE,"point");

  XMLRowCenter(ptr,sys,sky,format);
  XMLRow(XMLPARAM,pointtype);
  XMLRow(XMLPARAM2,size_);

  XMLRowProps(ptr,sys);
  XMLRowEnd(str);
}

void Point::listCiao(ostream& str, CoordSystem sys, int strip)
{
  FitsImage* ptr = parent->findFits();
  listCiaoPre(str);

  switch (sys) {
  case IMAGE:
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    {
      Vector v = ptr->mapFromRef(center,PHYSICAL);
      str << "point(" << setprecision(8) << v[0] << ',' << v[1] << ')';
    }
    break;
  default:
    if (ptr->hasWCSEqu(sys)) {
      char buf[64];
      ptr->mapFromRef(center,sys,FK5,SEXAGESIMAL,buf,64);
      char ra[16];
      char dec[16];
      string x(buf);
      istringstream wcs(x);
      wcs >> ra >> dec;
      str << "point" << '(' << ra << ',' << dec << ')';
    }
    break;
  }

  listCiaoPost(str, strip);
}

void Point::listPros(ostream& str, CoordSystem sys, SkyFrame sky,
		     SkyFormat format, int strip)
{
  FitsImage* ptr = parent->findFits();

  switch (sys) {
  case IMAGE:
  case DETECTOR:
  case AMPLIFIER:
    sys = IMAGE;
  case PHYSICAL:
    {
      listProsCoordSystem(str,sys,sky);
      str << "; ";
      Vector v = ptr->mapFromRef(center,sys);
      str << "point " << setprecision(8) << v;
    }
    break;
  default:
    if (ptr->hasWCSEqu(sys)) {
      listProsCoordSystem(str,sys,sky);
      str << "; ";

      switch (format) {
      case DEGREES:
	{
	  Vector v = ptr->mapFromRef(center,sys,sky);
	  str << "point " << setprecision(8) << v[0] << "d " << v[1] << "d ";
	}
	break;
      case SEXAGESIMAL:
	{
	  char buf[64];
	  ptr->mapFromRef(center,sys,sky,format,buf,64);
	  char ra[16];
	  char dec[16];
	  string x(buf);
	  istringstream wcs(x);
	  wcs >> ra >> dec;
	  if (dec[0]=='+')
	    str << "point " << ra << ' ' << dec+1;
	  else
	    str << "point " << ra << ' ' << dec;
	}
	break;
      }
    }
  }

  listProsPost(str, strip);
}

void Point::listSAOtng(ostream& str, CoordSystem sys, SkyFrame sky,
		       SkyFormat format, int strip)
{
  FitsImage* ptr = parent->findFits();
  listSAOtngPre(str, strip);

  switch (sys) {
  case IMAGE:
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    {
      Vector v = ptr->mapFromRef(center,IMAGE);
      str << "point(" << setprecision(8) << v[0] << ',' << v[1] << ')';
    }
    break;
  default:
    if (ptr->hasWCSEqu(sys)) {
      switch (format) {
      case DEGREES:
	{
	  Vector v = ptr->mapFromRef(center,sys,sky);
	  str << "point(" << setprecision(8) << v[0] << ',' << v[1] << ')';
	}
	break;
      case SEXAGESIMAL:
	{
	  char buf[64];
	  ptr->mapFromRef(center,sys,sky,format,buf,64);
	  char ra[16];
	  char dec[16];
	  string x(buf);
	  istringstream wcs(x);
	  wcs >> ra >> dec;
	  str << "point(" << ra << ',' << dec << ')';
	}
	break;
      }
    }
  }

  listSAOtngPost(str,strip);
}

void Point::listSAOimage(ostream& str, int strip)
{
  FitsImage* ptr = parent->findFits();
  listSAOimagePre(str);

  // all coords are in image coords

  Vector v = ptr->mapFromRef(center,IMAGE);
  str << "point(" << setprecision(8) << v[0] << ',' << v[1] << ')';

  listSAOimagePost(str, strip);
}
