// Copyright (C) 1999-2010
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "text.h"
#include "framebase.h"
#include "fitsimage.h"
#include "rotstr.h"

#include "tcl.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

Text::Text(const Text& a) : Marker(a) 
{
  rotate = a.rotate;
}

Text::Text(FrameBase* p, const Vector& ctr,
	   double ang, int rot,
	   const char* clr, int* dsh,
	   int wth, const char* fnt, const char* txt, 
	   unsigned short prop, const char* cmt,
	   const List<Tag>& tg, const List<CallBack>& cb) 
  : Marker(p, ctr, ang, clr, dsh, wth, fnt, txt, prop, cmt, tg, cb)
{
  strcpy(type,"text");
  handle = new Vector[4];
  numHandle = 4;
  rotate = rot;

  updateBBox();
}

void Text::draw(Drawable drawable, int tt)
{
  // always render regardless of tt
  renderXHandles(drawable);
  renderX(drawable, parent->refToWidget, SRC);
}

void Text::drawXOR()
{
  renderX(Tk_WindowId(parent->tkwin), parent->refToWindow, XOR);
}

void Text::drawMagnifier(const Matrix& mx, int tt)
{
  // always render regardless of tt
  renderX(parent->magnifierPixmap, mx, SRC);
}

void Text::renderX(Drawable drawable, const Matrix& mx, RenderMode mode)
{
  GC lgc = renderXGC(mode);

  double ang = rotate ? calcAngle() : 0;
  if (text && tkfont_)
    XSetFont(display, lgc, Tk_FontId(tkfont_));

  Tk_FontMetrics metrics;
  Tk_GetFontMetrics(tkfont_, &metrics);
  int width = Tk_TextWidth(tkfont_, text, strlen(text));
  int height = metrics.ascent-metrics.descent;

  Vector vv = center * mx;
  Matrix mm = Translate(-vv) *
    Translate(-width/2., height/2.) *
    Rotate(ang) *
    Translate(vv);
  Vector vvv = vv * mm;

  if (ang>-.01 && ang<+.01)
    Tk_DrawChars(display, drawable, lgc, tkfont_, text, strlen(text), 
		 vvv[0], vvv[1]);
  else
    DrawRotString(display, drawable, lgc, tkfont_, ang, vvv, text, 
		  (Widget*)parent);
}

void Text::ps(int mode, int tt)
{
  // always render regardless of tt
  renderPS(mode);
}

void Text::renderPS(int mode)
{
  renderPSGC(mode);

  double ang = rotate ? calcAngle() : 0;
  if (text && psfont_) {
    Tcl_DString psdstr;
    Tcl_DStringInit(&psdstr);
    int psSize = Tk_PostscriptFontName(psfont_, &psdstr) *
      parent->getDisplayRatio();

    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(psfont_, &metrics);
    int width = Tk_TextWidth(psfont_, text, strlen(text));

    Vector cc = center * parent->refToCanvas;
    Matrix m = Translate(-cc) * Rotate(-ang) * 
      Translate(-width/2.,metrics.descent) *
      Rotate(ang) * Translate(cc);

    ostringstream str;
    str << '/' << Tcl_DStringValue(&psdstr) 
	<< " findfont " << psSize << " scalefont setfont" << endl;
      
    Tcl_DStringFree(&psdstr);

    str << "gsave "
        << (cc*m).TkCanvasPs(parent->canvas) << " moveto" << endl
      	<< radToDeg(ang) << " rotate " 
	<< '(' << psQuote(text) << ')' 
	<< " show grestore" << endl << ends;
    
    Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);
  }
}

#ifdef _MACOSX
void Text::macosx(int tt)
{
  // always render regardless of tt
  renderMACOSX();
}

void Text::renderMACOSX()
{
  renderMACOSXGC();

  double ang = rotate ? calcAngle() : 0;
  if (text && psfont_) {
    Tcl_DString psdstr;
    Tcl_DStringInit(&psdstr);
    int psSize = Tk_PostscriptFontName(psfont_, &psdstr);
    macosxFont(Tcl_DStringValue(&psdstr), psSize);
    Tcl_DStringFree(&psdstr);

    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(psfont_, &metrics);
    int width = Tk_TextWidth(psfont_, text, strlen(text));

    Vector cc = center * parent->refToCanvas;
    Matrix m = Translate(-cc) * Rotate(-ang) * 
      Translate(-width/2.,metrics.descent) *
      Rotate(ang) * Translate(cc);

    macosxDrawText(cc*m, ang, text);
  }
}
#endif

#ifdef _WIN32
void Text::win32(int tt)
{
  // always render regardless of tt
  renderWIN32();
}

void Text::renderWIN32()
{
  renderWIN32GC();

  double ang = calcAngle();
  if (text && psfont_) {
    Tcl_DString psdstr;
    Tcl_DStringInit(&psdstr);
    int psSize = Tk_PostscriptFontName(psfont_, &psdstr);
    win32Font(Tcl_DStringValue(&psdstr), psSize);
    Tcl_DStringFree(&psdstr);

    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(psfont_, &metrics);
    int width = Tk_TextWidth(psfont_, text, strlen(text));

    Vector cc = center * parent->refToCanvas;
    Matrix m = Translate(-cc) * Rotate(-ang) * 
      Translate(-width/2.,metrics.descent) *
      Rotate(ang) * Translate(cc);

    win32DrawText(cc*m, ang, text);
  }
}
#endif

void Text::updateBBox()
{
  Vector cc = center * parent->refToCanvas;
  bbox = BBox(cc);

  if (text && tkfont_) {
    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(tkfont_, &metrics);
    int width = Tk_TextWidth(tkfont_, text, strlen(text));

    // generate handles
    Vector s(width, metrics.linespace);
    double ang = rotate ? calcAngle() : 0;
    Matrix mm = Rotate(ang) * Translate(cc);

    handle[0] = Vector(-s[0],-s[1])/2 * mm;
    handle[1] = Vector( s[0],-s[1])/2 * mm;
    handle[2] = Vector( s[0], s[1])/2 * mm;
    handle[3] = Vector(-s[0], s[1])/2 * mm;

    // bbox
    for (int i=0; i<4; i++)
      bbox.bound(handle[i]);
  }
  else
    bbox.expand(5);

  // make room for handles
  bbox.expand(3);
  allBBox = bbox;
}

void Text::setRotate(int rot)
{
  rotate = rot ? 1 : 0;
  updateBBox();
}

// list

void Text::list(ostream& str, CoordSystem sys, SkyFrame sky, 
		SkyFormat format, int conj, int strip)
{
  if (!strip) {
    FitsImage* ptr = parent->findFits(sys,center);
    listPre(str, sys, sky, ptr, strip, 1);

    switch (sys) {
    case IMAGE:
    case PHYSICAL:
    case DETECTOR:
    case AMPLIFIER:
      {
	Vector v = ptr->mapFromRef(center,sys);
	str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ')';
      }
      break;
    default:
      if (ptr->hasWCS(sys)) {
	if (ptr->hasWCSEqu(sys)) {
	  switch (format) {
	  case DEGREES:
	    {
	      Vector v = ptr->mapFromRef(center,sys,sky);
	      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] 
		  << ')';
	    }
	    break;
	  case SEXAGESIMAL:
	    {
	      char buf[64];
	      ptr->mapFromRef(center,sys,sky,format,buf,64);
	      char ra[16];
	      char dec[16];

	      string x(buf);
	      istringstream wcs(x);
	      wcs >> ra >> dec;
	      str << type << '(' << ra << ',' << dec << ')';
	    }
	    break;
	  }
	}
	else {
	  Vector v = ptr->mapFromRef(center,sys);
	  str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ')';
	}
      }
    }

    if (conj)
      str << " ||";

    if (angle != 0)
      str << " textangle=" << radToDeg(parent->mapAngleFromRef(angle,sys,sky));
    if (!rotate)
      str << " textrotate=" << 0;
    listProperties(str, 0);
  }
}

void Text::listXML(ostream& str, CoordSystem sys, SkyFrame sky, 
		   SkyFormat format)
{
  FitsImage* ptr = parent->findFits(sys,center);

  XMLRowInit();
  XMLRow(XMLSHAPE,type);

  XMLRowCenter(ptr,sys,sky,format);
  XMLRowAng(sys,sky);
  XMLRow(XMLPARAM,rotate);

  XMLRowProps(ptr,sys);
  XMLRowEnd(str);
}

void Text::listSAOtng(ostream& str, CoordSystem sys, SkyFrame sky,
		      SkyFormat format, int strip)
{
  FitsImage* ptr = parent->findFits();
  if (properties&INCLUDE)
    str << '+';
  else
    str << '-';

  switch (sys) {
  case IMAGE:
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    {
      Vector v = ptr->mapFromRef(center,IMAGE);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] 
	  << ", \"" << text << "\")";
    }
    break;
  default:
    if (ptr->hasWCSEqu(sys)) {
      switch (format) {
      case DEGREES:
	{
	  Vector v = ptr->mapFromRef(center,sys,sky);
	  str << type << '(' << setprecision(8) << v[0] << ',' << v[1] <<", \""
	      << text << "\")";
	}
	break;
      case SEXAGESIMAL:
	{
	  char buf[64];
	  ptr->mapFromRef(center,sys,sky,format,buf,64);
	  char ra[16];
	  char dec[16];
	  string x(buf);
	  istringstream wcs(x);
	  wcs >> ra >> dec;
	  str << type << '(' << ra << ',' << dec
	      << ", \"" << text << "\")";
	}
	break;
      }
    }
  }

  listSAOtngPost(str, strip);
}




