// Copyright (C) 1999-2010
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "gridbase.h"
#include "attribute.h"
#include "widget.h"
#include "rotstr.h"
#include "util.h"

GridBase* astGridPtr = NULL;

extern "C" {
  #include "ast.h"
  #include "grf.h"
}

GridBase::GridBase(Widget* p) : parent(p)
{
  renderMode = X11;
  line = new Attribute(parent);
  text = new Attribute(parent);

  gc_ = NULL;
  pixmap_ = 0;

  mode_ = Widget::RGB;
}

GridBase::GridBase(Widget* p, const char* o) : parent(p)
{
  option_ = dupstr(o);

  renderMode = X11;
  line = new Attribute(parent);
  text = new Attribute(parent);

  gc_ = NULL;
  pixmap_ = 0;

  mode_ = Widget::RGB;
}

GridBase::~GridBase()
{
  if (option_)
    delete [] option_;

  if (line)
    delete line;
  if (text)
    delete text;
}

// Generic

int GridBase::gAttr(int which, double value, double* old, int prim)
{
  Attribute* attr;
  switch (prim) {
  case GRF__TEXT:
    attr = text;

    switch (which) {
    case GRF__STYLE:
      break;
    case GRF__WIDTH:
      break;
    case GRF__SIZE:
      if (old)
	*old = attr->size();
      if (value != AST__BAD) 
	attr->setSize(value);
      break;
    case GRF__FONT:
      if (old)
	*old = attr->font();
      if (value != AST__BAD) 
	attr->setFont(value);
      break;
    case GRF__COLOUR:
      if (old)
	*old = attr->colour();
      if (value != AST__BAD) 
	attr->setColour(value);
      break;
    }

    break;
  case GRF__LINE:
    attr = line;

    switch (which) {
    case GRF__STYLE:
      if (old)
	*old = attr->style();
      if (value != AST__BAD) 
	attr->setStyle(value);
      break;
    case GRF__WIDTH:
      if (old)
	*old = attr->width();
      if (value != AST__BAD) 
	attr->setWidth(value);
      break;
    case GRF__SIZE:
      break;
    case GRF__FONT:
      break;
    case GRF__COLOUR:
      if (old)
	*old = attr->colour();
      if (value != AST__BAD) 
	attr->setColour(value);
      break;
    }

    break;
  }

  return 1;
}

int GridBase::gLine(int n, float* x, float* y)
{
  switch (renderMode) {
  case X11:
    x11Line(n,x,y);
    break;
  case PS:
    psLine(n,x,y);
    break;
#ifdef _MACOSX
  case MACOSX:
    macosxLine(n,x,y);
    break;
#endif
#ifdef _WIN32
  case GWIN32:
    win32Line(n,x,y);
    break;
#endif
  }

  return 1;
}

int GridBase::gQch(float* chv, float* chh)
{
  Tk_Font font =NULL;
  switch (renderMode) {
  case X11:
    font = text->tkfont();
    break;
  case PS:
    font = text->psfont();
    break;
#ifdef _MACOSX
  case MACOSX:
    font = text->psfont();
    break;
#endif
#ifdef _WIN32
  case GWIN32:
    font = text->psfont();
    break;
#endif
  }

  if (font) {
    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(font, &metrics);

    *chv = (float)metrics.linespace;
    *chh = (float)metrics.linespace;
    return 1;
  }
  else {
    *chv = *chh = 0;
    return 0;
  }
}

int GridBase::gText(const char* txt, float x, float y, const char* just, 
		    float upx, float upy)
{
  if (!(txt && txt[0] && just && just[0] && just[1]))
    return 0;

  switch (renderMode) {
  case X11:
    return x11Text(txt,x,y,just,upx,upy);
  case PS:
    return psText(txt,x,y,just,upx,upy);
#ifdef _MACOSX
  case MACOSX:
    return macosxText(txt,x,y,just,upx,upy);
#endif
#ifdef _WIN32
  case GWIN32:
    return win32Text(txt,x,y,just,upx,upy);
#endif
  }

  return 0;
}

int GridBase::gTxExt(const char* txt, float x, float y, const char* just,
		 float upx, float upy, float* xb, float* yb)
{
  if (!(txt && txt[0] && just)) {
    xb[0] = xb[1] = xb[2] = xb[3] = 0;
    yb[0] = yb[1] = yb[2] = yb[3] = 0;

    return 0;
  }

  Tk_Font font;

  switch (renderMode) {
  case X11:
    font = text->tkfont();
    break;
  case PS:
    font = text->psfont();
    break;
#ifdef _MACOSX
  case MACOSX:
    font = text->psfont();
    break;
#endif
#ifdef _WIN32
  case GWIN32:
    font = text->psfont();
    break;
#endif
  }

  Vector vv = Vector(x,y);
  double angle = calcTextAngle(just, upx, upy);
  Vector cc = vv * calcTextPos(vv, angle, txt, just, font);

  Tk_FontMetrics metrics;
  Tk_GetFontMetrics(font, &metrics);
  int width = Tk_TextWidth(font, txt, strlen(txt));

  BBox nn = BBox(Vector(0,0), Vector(width,metrics.linespace));
  BBox bb = nn * Rotate(angle) * Translate(cc);

  xb[0] = bb.ll[0];
  yb[0] = bb.ll[1];

  xb[1] = bb.ur[0];
  yb[1] = bb.ll[1];

  xb[2] = bb.ur[0];
  yb[2] = bb.ur[1];

  xb[3] = bb.ll[0];
  yb[3] = bb.ur[1];

  return 1;
}

int GridBase::gScales(float* alpha, float* beta)
{
  /*
  *alpha = float(XDisplayWidthMM(parent->getDisplay(),parent->screenNumber)) / 
    float(XDisplayWidth(parent->getDisplay(),parent->screenNumber));

  *beta = float(XDisplayHeightMM(parent->getDisplay(),parent->screenNumber)) / 
    float(XDisplayHeight(parent->getDisplay(),parent->screenNumber));
  */

  return 1;
}

int GridBase::gCap(int cap, int value)
{
  switch (cap) {
  case GRF__SCALES:
    return 0;
  case GRF__MJUST:
    return 1;
  case GRF__ESC:
    return 0;
  }
}

// X11 Render functions

int GridBase::x11Line(int n, float* x, float* y)
{
  if (n<2 || !x || !y)
    return 1;

  XSetForeground(parent->getDisplay(), gc_, line->color());
  int w = (int)line->width();
  if (w<1)
    w = 1;
  switch (line->style()) {
  case Attribute::SOLID:
    XSetLineAttributes(parent->getDisplay(), gc_, w, 
		       LineSolid, CapButt, JoinMiter);
    break;
  case Attribute::DASH:
    XSetLineAttributes(parent->getDisplay(), gc_, w, 
		       LineOnOffDash, CapButt, JoinMiter);
    char dlist[] = {8,3};
    XSetDashes(parent->getDisplay(), gc_, 0, dlist, 2);
    break;
  }

  for (int i=0; i<n-1; i++) {
    Vector s = Vector(x[i],y[i]).round();
    Vector e = Vector(x[i+1],y[i+1]).round();
    XDrawLine(parent->getDisplay(), pixmap_, gc_, 
	      (int)s[0],(int)s[1],(int)e[0],(int)e[1]);
  }

  return 1;
}

int GridBase::x11Text(const char* txt, float x, float y, 
		   const char* just, float upx, float upy)
{
  XSetFont(parent->getDisplay(), gc_, Tk_FontId(text->tkfont()));
  XSetForeground(parent->getDisplay(), gc_, text->color());

  Vector vv = Vector(x,y);
  double angle = calcTextAngle(just, upx, upy);
  Vector cc = vv * calcTextPos(vv, angle, txt, just, text->tkfont());

  if ((angle>-.01 && angle<+.01))
    Tk_DrawChars(parent->getDisplay(), pixmap_, gc_, text->tkfont(), 
		 txt, strlen(txt), (int)cc[0], (int)cc[1]);
  else
    DrawRotString(parent->getDisplay(), pixmap_, gc_, text->tkfont(), 
		  angle, cc, txt, parent);
  return 1;
}

// PS Render functions

int GridBase::psLine(int n, float* x, float* y)
{
  if (n<2 || !x || !y)
    return 1;

  psColor(line);
  {
    ostringstream str;
    str << line->width() << " setlinewidth" << endl << ends;
    Tcl_AppendResult(parent->getInterp(), str.str().c_str(), NULL);
  }
  {
    ostringstream str;
    switch (line->style()) {
    case Attribute::SOLID:
      str << "[] 0 setdash" << endl << ends;
      break;
    case Attribute::DASH:
      str << "[8 3] 0 setdash" << endl << ends;
      break;
    }
    Tcl_AppendResult(parent->getInterp(), str.str().c_str(), NULL);
  }

  for (int i=0; i<n; i++) {
    Vector v = Vector(x[i],y[i]) * matrix_;

    ostringstream str;
    if (i == 0) {
      str << "newpath " << endl;
      str << v.TkCanvasPs(parent->getCanvas()) << " moveto" << endl << ends;
    }
    else
      str << v.TkCanvasPs(parent->getCanvas()) << " lineto" << endl << ends;

    Tcl_AppendResult(parent->getInterp(), str.str().c_str(), NULL);
  }

  ostringstream str;
  str << "stroke" << endl << ends;
  Tcl_AppendResult(parent->getInterp(), str.str().c_str(), NULL);

  return 1;
}

int GridBase::psText(const char* txt, float x, float y, 
		const char* just, float upx, float upy)
{
  Tcl_DString psdstr;
  Tcl_DStringInit(&psdstr);
  int psSize = Tk_PostscriptFontName(text->psfont(), &psdstr) * 
    parent->getDisplayRatio();

  Vector vv = Vector(x,y) * matrix_;
  double angle = calcTextAngle(just, upx, upy);
  Vector cc = vv * calcTextPos(vv, angle, txt, just, text->psfont());
    
  ostringstream str;
  str << '/' << Tcl_DStringValue(&psdstr) 
      << " findfont " << psSize << " scalefont setfont" << endl;
  Tcl_DStringFree(&psdstr);

  psColor(text);

  str << "gsave " 
      << cc.TkCanvasPs(parent->getCanvas()) << " moveto" << endl
      << radToDeg(angle) << " rotate " 
      << '(' << psQuote(txt) << ')' << " show"
      << " grestore" << endl << ends;
  Tcl_AppendResult(parent->getInterp(), str.str().c_str(), NULL);

  return 1;
}

void GridBase::psColor(Attribute* attr)
{
  ostringstream str;
  switch ((Widget::PSColorSpace)mode_) {
  case Widget::BW:
  case Widget::GRAY:
    psColorGray(parent->getXColor(attr->colorName()), str);
    str << " setgray";
    break;
  case Widget::RGB:
    psColorRGB(parent->getXColor(attr->colorName()), str);
    str << " setrgbcolor";
    break;
  case Widget::CMYK:
    psColorCMYK(parent->getXColor(attr->colorName()), str);
    str << " setcmykcolor";
    break;
  }
  str << endl << ends;
  Tcl_AppendResult(parent->getInterp(), str.str().c_str(), NULL);
}

#ifdef _MACOSX
int GridBase::macosxLine(int n, float* x, float* y)
{
  if (n<2 || !x || !y)
    return 1;

  macosxColor(parent->getXColor(line->colorName()));
  macosxWidth(line->width());
  switch (line->style()) {
  case Attribute::SOLID:
    macosxDash(NULL,0);
    break;
  case Attribute::DASH:
    float dlist[] = {8,3};
    macosxDash(dlist,2);
    break;
  }

  Vector v[n];
  for (int i=0; i<n; i++)
    v[i] = Vector(x[i],y[i]) * matrix_;

  macosxDrawLines(v, n);
  
  return 1;
}

int GridBase::macosxText(const char* txt, float x, float y, 
			 const char* just, float upx, float upy)
{
  // change this later
  Tcl_DString psdstr;
  Tcl_DStringInit(&psdstr);
  int psSize = Tk_PostscriptFontName(text->psfont(), &psdstr);

  macosxFont(Tcl_DStringValue(&psdstr),psSize); 
  Tcl_DStringFree(&psdstr);

  Vector vv = Vector(x,y) * matrix_;
  double angle = calcTextAngle(just, upx, upy);
  Vector cc = vv * calcTextPos(vv, angle, txt, just, text->tkfont());
    
  macosxColor(parent->getXColor(text->colorName()));
  macosxDrawText(cc, angle, txt);

  return 1;
}
#endif

#ifdef _WIN32
int GridBase::win32Line(int n, float* x, float* y)
{
  if (n<2 || !x || !y)
    return 1;

  win32Color(parent->getXColor(line->colorName()));
  win32Width(line->width());
  switch (line->style()) {
  case Attribute::SOLID:
    win32Dash(NULL,0);
    break;
  case Attribute::DASH:
    float dlist[] = {8,3};
    win32Dash(dlist,2);
    break;
  }

  Vector v[n];
  for (int i=0; i<n; i++)
    v[i] = Vector(x[i],y[i]) * matrix_;

  win32DrawLines(v, n);
  
  return 1;
}

int GridBase::win32Text(const char* txt, float x, float y, 
			 const char* just, float upx, float upy)
{
  // change this later
  Tcl_DString psdstr;
  Tcl_DStringInit(&psdstr);
  int psSize = Tk_PostscriptFontName(text->psfont(), &psdstr);

  win32Font(Tcl_DStringValue(&psdstr),psSize); 
  Tcl_DStringFree(&psdstr);

  Vector vv = Vector(x,y) * matrix_;
  double angle = calcTextAngle(just, upx, upy);
  Vector cc = vv * calcTextPos(vv, angle, txt, just, text->tkfont());
    
  win32Color(parent->getXColor(text->colorName()));
  win32DrawText(vv, angle, txt);

  return 1;
}
#endif

// Support

double GridBase::calcTextAngle(const char* just, float upx, float upy)
{
  // this is kludge fix to get around an issue with AST
  if (just[0] == 'T' && upy<0)
    upy = fabs(upy);

  double a = Vector(upx,upy).angle();

  // our angle is 90 off from ast's, and the other direction
  double r = -(a - M_PI/2);

  // normalize
  if (r>0)
    while (r>2*M_PI)
      r -= 2*M_PI;
  else 
    while (r<0)
      r += 2*M_PI;

  // and flip if upside down
  // we only want to do this if the number is really, really upside down
  // you might have the case of a curved vertical line
  if (r > M_PI/2 && r <= 3*M_PI/2)
    r -= M_PI;

  return r;
}

Matrix GridBase::calcTextPos(const Vector& vv, double angle, const char* txt, 
			     const char* just, Tk_Font font)
{
  Tk_FontMetrics metrics;
  Tk_GetFontMetrics(font, &metrics);
  int width = Tk_TextWidth(font, txt, strlen(txt));

  Matrix m1,m2;
  switch (just[0]) {
  case 'T':
    break;
  case 'C':
    m1 = Translate(0,metrics.linespace/2);
    break;
  case 'B':
    m1 = Translate(0,metrics.ascent);
    break;
  case 'M':
    m1 = Translate(0,metrics.linespace);
    break;
  }

  switch (just[1]) {
  case 'L':
    break;
  case 'C':
    m2 = Translate(-width/2.,0);
    break;
  case 'R':
    m2 = Translate(-width,0);
    break;
  }

  return Translate(-vv) *
    Rotate(-angle) *
    m1 * m2 *
    Rotate(angle) *
    Translate(vv);
}

int GridBase::matrixMap(void* fs, Matrix& mx, const char* str)
{
  AstFrameSet* frameSet = (AstFrameSet*)fs;

  double ss[] = {mx.matrix(0,0),mx.matrix(1,0),mx.matrix(0,1),mx.matrix(1,1)};
  double tt[] = {mx.matrix(2,0),mx.matrix(2,1)};

  AstMatrixMap* mm;
  if (!(mm= astMatrixMap(2, 2, 0, ss, "")))
    return 0;

  AstShiftMap* sm;
  if (!(sm = astShiftMap(2, tt, "")))
    return 0;

  AstCmpMap* cmp;
  if (!(cmp = astCmpMap(mm, sm, 1, "")))
    return 0;

  astAddFrame(frameSet, AST__CURRENT, cmp, astFrame(2, str));
}

