// Copyright (C) 1999-2010
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "rotstr.h"
#include "util.h"
#include "widget.h"

#include <tk.h>
#include <X11/Xutil.h>

#ifndef _MACOSX

void DrawRotString(Display* display, Drawable drawable, GC gc, Tk_Font font, 
		    double angle, const Vector& v, const char* text, 
		    Widget* parent)
{
  // this is a very crude "green screen" method.
  // since the text is rendered with an anti-alias method, the text color
  // will vary from the initial color to the mask color.
  // so to avoid a 'dark' ring around the text, we calculate a mask value, 
  // based on the average of the pixels directly under the text

  int depth = parent->getDepth();

  Tk_FontMetrics metrics;
  Tk_GetFontMetrics(font, &metrics);
  int width = Tk_TextWidth(font, text, strlen(text));
  int height = metrics.linespace;

  // calculate bounding box
  int srcw = width+4;
  int srch = height+2;
  Vector corner = (Vector(srcw,srch)-Vector(1,1))/2;
  BBox bb;
  bb.bound(-corner * Rotate(angle));
  bb.bound( corner * Rotate(angle));
  bb.bound(Vector( corner[0],-corner[1]) * Rotate(angle));
  bb.bound(Vector(-corner[0], corner[1]) * Rotate(angle));
  Vector dstsize = bb.size();
  int dstw = (int)dstsize[0];
  int dsth = (int)dstsize[1];

  // dst pixmap
  Pixmap dstpmap = Tk_GetPixmap(display, drawable, dstw, dsth, depth);

  // calculate dst location
  int ww = width;
  int hh = metrics.ascent-metrics.descent;
  Matrix mm = Translate(ww/2.,-hh/2.) * Rotate(angle) * Translate(v);
  BBox vv(v,v);
  vv.bound(-corner * mm);
  vv.bound( corner * mm);
  vv.bound(Vector( corner[0],-corner[1]) * mm);
  vv.bound(Vector(-corner[0], corner[1]) * mm);
  Vector tt = vv.ll;

  // get original image data into dst pixmap
  XCopyArea(display, drawable, dstpmap, gc, (int)tt[0], (int)tt[1], 
	    dstw, dsth, 0, 0);

  XImage* dstxmap = XGetImage(display, dstpmap, 0, 0, dstw, dsth,
			      AllPlanes, ZPixmap);

  // calc mask
  union {
    unsigned long ll;
    unsigned char cc[4];
  } mask;

  {
    int cnt =0;
    unsigned char* dst = (unsigned char*)dstxmap->data;
    int dstl = dstxmap->bytes_per_line;
    int dstb = dstxmap->bits_per_pixel/8;

    unsigned long mm[4];
    for (int kk=0; kk<3; kk++) 
      mm[kk] =0;

    for (int jj=0; jj<dsth; jj++) {
      // the line may be padded at the end
      unsigned char* dptr = dst + jj*dstl;

      for (int ii=0; ii<dstw; ii++, dptr+=dstb) {
	for (int kk=0; kk<dstb; kk++)
	  mm[kk] += *(dptr+kk);
	cnt++;
      }
    }

    for (int kk=0; kk<3; kk++) {
      if (mm[kk])
	mm[kk] /= cnt;
      mask.cc[kk] = mm[kk];
    }
  }

  // src pixmap
  GC ggc = XCreateGC(display, drawable, 0L, NULL);
  XSetForeground(display, ggc, mask.ll);

  Pixmap srcpmap = Tk_GetPixmap(display, drawable, srcw, srch, depth);
  XFillRectangle(display, srcpmap, ggc, 0, 0, srcw, srch);

  // render text
  XCopyGC(display, gc, GCForeground|GCFont, ggc);
  Tk_DrawChars(display, srcpmap, ggc, font, text, strlen(text), 
	       2, metrics.ascent);
  XFreeGC(display, ggc);
  XImage* srcxmap = XGetImage(display, srcpmap, 0, 0, srcw, srch,
			      AllPlanes, ZPixmap);

  // rotate text from srcxmap to dstxmap
  {
    unsigned char* src = (unsigned char*)srcxmap->data;
    int srcl = srcxmap->bytes_per_line;
    int srcb = dstxmap->bits_per_pixel/8;
    unsigned char* dst = (unsigned char*)dstxmap->data;
    int dstl = dstxmap->bytes_per_line;
    int dstb = dstxmap->bits_per_pixel/8;
    Matrix m = 
      Translate(-Vector(srcw,srch)/2) *
      Rotate(angle) *
      Translate(Vector(dstw,dsth)/2);
    double* mm = (m.invert()).mm();

    for (int jj=0; jj<dsth; jj++) {
      // the line may be padded at the end
      unsigned char* dptr = dst + jj*dstl;

      for (int ii=0; ii<dstw; ii++, dptr+=dstb) {
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	if (xx >= 0 && xx < srcw && yy >= 0 && yy < srch) {
	  unsigned char* sptr = src + ((int)yy)*srcl + ((int)xx)*srcb;
	  switch (dstb) {
	  case 1:
	    if (*sptr != mask.cc[0])
	      memcpy(dptr,sptr,srcb);
	    break;
	  case 2:
	    if (*sptr != mask.cc[0] || 
		*(sptr+1) != mask.cc[1])
	      memcpy(dptr,sptr,srcb);
	    break;
	  case 3:
	  case 4:
	    if (*sptr != mask.cc[0] || 
		*(sptr+1) != mask.cc[1] || 
		*(sptr+2) != mask.cc[2])
	      memcpy(dptr,sptr,srcb);
	    break;
	  }
	}
      }
    }
  }

  // put rotated text into image
  XPutImage(display, dstpmap, gc, dstxmap, 0, 0, 0, 0, dstw, dsth);
  XCopyArea(display, dstpmap, drawable, gc, 0, 0, dstw, dsth, 
	    (int)tt[0], (int)tt[1]);

  // clean up
  Tk_FreePixmap(display, srcpmap);
  Tk_FreePixmap(display, dstpmap);
  XDestroyImage(srcxmap);
  XDestroyImage(dstxmap);
}

#endif
